/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012, 2013, 2014 INRIA.
**
** Contributors :
**
** francois.broquedis@imag.fr
** thierry.gautier@inrialpes.fr
**
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "libiomp.h"
#include <string.h>

#if KMP_ARCH_X86
# define KMP_SIZE_T_MAX (0xFFFFFFFF)
#else
# define KMP_SIZE_T_MAX (0xFFFFFFFFFFFFFFFF)
#endif


// Round up a size to a power of two specified by val
// Used to insert padding between structures co-allocated using a single malloc() call
static size_t
__kmp_round_up_to_val( size_t size, size_t val ) {
    if ( size & ( val - 1 ) ) {
        size &= ~ ( val - 1 );
        if ( size <= KMP_SIZE_T_MAX - val ) {
            size += val;    // Round up if there is no overflow.
        }; // if
    }; // if
    return size;
} // __kmp_round_up_to_va


/*
*/
kmp_task_t *__kmpc_omp_task_alloc(ident_t *loc_ref,
                                  kmp_int32 gtid,
                                  kmp_int32 flags,
                                  size_t sizeof_kmp_task_t,
                                  size_t sizeof_shareds,
                                  kmp_routine_entry_t task_entry) 
{
    TRACE_ENTER();
    kaapi_thread_t* thread = kaapi_self_thread();

    size_t shareds_offset = sizeof( kmp_taskdata_t ) + sizeof_kmp_task_t;
    shareds_offset = __kmp_round_up_to_val( shareds_offset, sizeof( void * ));

    size_t size_task = shareds_offset + sizeof_shareds;
    kmp_taskdata_t *new_taskdata = kaapi_data_push(thread, size_task);
    kmp_task_t     *new_task = (kmp_task_t*)(new_taskdata+1);
    new_taskdata->size    = (uint32_t)size_task;
#if KOMP_USE_ICV
    new_taskdata->icvs    = 0;
#endif
    new_task->shareds = (void*)(new_task+1);
    new_task->routine = task_entry;
    new_task->part_id = 0;

    kaapi_assert_debug( ( ((uintptr_t)new_taskdata) & (sizeof(double)-1) ) == 0 );
    kaapi_assert_debug( ( ((uintptr_t)new_task) & (sizeof(double)-1) ) == 0 );
    kaapi_assert_debug( ( ((uintptr_t)new_task->shareds) & (sizeof(double)-1) ) == 0 );

    TRACE_LEAVE();

    return new_task;
}


/*
*/
static void __kmpc_wrapper_task( kaapi_task_t* task, kaapi_thread_t* thread )
{
  kmp_task_t* kmptask = (kmp_task_t*)task->arg;
  kmp_taskdata_t *kmptaskdata = ((kmp_taskdata_t*)kmptask) - 1;
  kaapi_context_t*    context = kaapi_thread2context(thread);
#if KOMP_USE_ICV
  komp_dataenv_icv_t* save_icvs = context->komp_icvs;
  context->komp_icvs = kmptaskdata->icvs;
#endif
  kmptask->routine( (kmp_int32)context->proc->kid, kmptask );
#if KOMP_USE_ICV
  context->komp_icvs = save_icvs;
#endif
}

/*
*/
kmp_int32 __kmpc_omp_task(ident_t *loc_ref,
                          kmp_int32 gtid,
                          kmp_task_t *kmptask)
{
  TRACE_ENTER();
  kaapi_processor_t* kproc = kaapi_self_processor();
  kaapi_context_t* context = kproc->context;

#if KOMP_USE_ICV
  kmp_taskdata_t *kmptaskdata = ((kmp_taskdata_t*)kmptask) - 1;
  komp_dataenv_icv_t* env = komp_get_dataenv_icv( kproc );
  kmptaskdata->icvs       = komp_newdataenv_icvs( kproc, env );
#endif

  kaapi_task_push( kaapi_context2thread(context),
                  __kmpc_wrapper_task,
                  kmptask);
  
  TRACE_LEAVE();
  return 0;
}


/*
*/
kmp_int32 __kmpc_omp_taskwait(ident_t *loc_ref, kmp_int32 gtid)
{
  TRACE_ENTER();
  
  kaapi_sched_sync(kaapi_self_thread());
  
  TRACE_LEAVE();

  return 0;
}


/* format definition for any Intel OMP Kaapi's task */
static size_t
kaapi_iomp_task_format_get_size(const struct kaapi_format_t* fmt, const void* sp)
{
  kmp_taskdata_t *task = ((kmp_taskdata_t *)sp)-1;
  return task->size;
}

static void
kaapi_iomp_task_format_task_copy(const struct kaapi_format_t* fmt, void* sp_dest, const void* sp_src)
{
  kmp_taskdata_t *task_src = ((kmp_taskdata_t *)sp_src)-1;
  memcpy( sp_dest, task_src, task_src->size );
}

static
unsigned kaapi_iomp_task_format_get_count_params(const struct kaapi_format_t* fmt, const void* sp)
{
  return 0;
}

static
kaapi_access_mode_t kaapi_iomp_task_format_get_mode_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  return KAAPI_ACCESS_MODE_V;
}

static
void* kaapi_iomp_task_format_get_data_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  void* retval = 0;
  abort();
  return retval;
}

static
kaapi_access_t* kaapi_iomp_task_format_get_access_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  return 0;
}

static
void kaapi_iomp_task_format_set_access_param(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const kaapi_access_t* a
)
{
  abort();
}

static
const struct kaapi_format_t* kaapi_iomp_task_format_get_fmt_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  abort();
  return 0;
}

static
void kaapi_iomp_task_format_get_view_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp, kaapi_memory_view_t* view
)
{
  abort();
}

static
void kaapi_iomp_task_format_set_view_param(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const kaapi_memory_view_t* view
)
{
  abort();
}

static
void kaapi_iomp_task_format_reducor(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const void* v
)
{
  abort();
}

static
void kaapi_iomp_task_format_redinit(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp, void* v
)
{
  abort();
}

static kaapi_adaptivetask_splitter_t	kaapi_iomp_task_fnc_get_splitter(
  const struct kaapi_format_t* fmt, const void* sp
)
{
  return 0;
}


static struct kaapi_format_t* kaapi_iomp_task_format = 0;

/*
*/
__attribute__((constructor))
static void kmpc_kaapi_task_format_constructor(void)
{
  if (kaapi_iomp_task_format !=0) return;
  TRACE_ENTER();
  kaapi_iomp_task_format = kaapi_format_allocate();
  kaapi_format_taskregister_func(
                                 kaapi_iomp_task_format,
                                 __kmpc_wrapper_task,
                                 "kaapi_iomp_task_format",
                                 0,
                                 kaapi_iomp_task_format_get_size,
                                 kaapi_iomp_task_format_task_copy,
                                 kaapi_iomp_task_format_get_count_params,
                                 kaapi_iomp_task_format_get_mode_param,
                                 kaapi_iomp_task_format_get_data_param,
                                 kaapi_iomp_task_format_get_access_param,
                                 kaapi_iomp_task_format_set_access_param,
                                 kaapi_iomp_task_format_get_fmt_param,
                                 kaapi_iomp_task_format_get_view_param,
                                 kaapi_iomp_task_format_set_view_param,
                                 kaapi_iomp_task_format_reducor,
                                 kaapi_iomp_task_format_redinit,
                                 kaapi_iomp_task_fnc_get_splitter
                                 );
    TRACE_LEAVE();
}

void __kmpc_omp_task_begin_if0(ident_t *loc_ref, kmp_int32 gtid, kmp_task_t * task)
{
    task->routine(kaapi_self_kid(), task);
}

void __kmpc_omp_task_complete_if0(ident_t *loc_ref, kmp_int32 gtid, kmp_task_t *task)
{
    /* Nothing to do here. */
}

