/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** francois.broquedis@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include <errno.h>

#include "libkomp_impl.h"
#include "kaapi_util.h"

// DEBUG
#include "kaapi_foreach_impl.h"

#define KOMP_DECLARE_FMT(name) \
  struct komp_format_t* komp_##name##_format;

#define KOMP_INIT_FMT(name) \
  komp_##name##_format = (struct komp_format_t*)kaapi_##name##_format;

KOMP_DECLARE_FMT(schar)
KOMP_DECLARE_FMT(char)
KOMP_DECLARE_FMT(shrt)
KOMP_DECLARE_FMT(int)
KOMP_DECLARE_FMT(long)
KOMP_DECLARE_FMT(llong)
KOMP_DECLARE_FMT(int8)
KOMP_DECLARE_FMT(int16)
KOMP_DECLARE_FMT(int32)
KOMP_DECLARE_FMT(int64)
KOMP_DECLARE_FMT(uchar)
KOMP_DECLARE_FMT(ushrt)
KOMP_DECLARE_FMT(uint)
KOMP_DECLARE_FMT(ulong)
KOMP_DECLARE_FMT(ullong)
KOMP_DECLARE_FMT(uint8)
KOMP_DECLARE_FMT(uint16)
KOMP_DECLARE_FMT(uint32)
KOMP_DECLARE_FMT(uint64)
KOMP_DECLARE_FMT(flt)
KOMP_DECLARE_FMT(dbl)
KOMP_DECLARE_FMT(ldbl)
KOMP_DECLARE_FMT(voidp)


static komp_device_icv_t komp_global_device_icv = {
  .def_sched         = komp_sched_static,
  .stacksize         = 0,  /* means system default */
  .waitpolicy        = komp_wait_policy_active, /* no choice in this version ! */
  .max_active_levels = 256,
  .cancel            = 0
};

komp_device_icv_t* komp_get_global_device_icv(void)
{
  return &komp_global_device_icv;
}

komp_dataenv_icv_t komp_global_dataenv_icv = {
  .dyn           = 0,
  .nest          = 0,
  .h_nthreads    = 0, /* runtime selection */
  .nthreads      = 0, /* runtime selection */
  .run_sched     = komp_sched_static,
  .run_sched_modifier = 0,
  .h_bind        = komp_proc_bind_false,
  .binds         = 0,
  .thread_limit  = 8192, /* sens on cpu */
  .active_levels = 0,
  .levels        = 0,
  .task_group    = 0,
  .prev_icvs     = 0,
  .team          = 0,
  .preferred_ld  = 0,
  .prio          = 0,
  .flag          = 0
};


static bool
komp_parse_stacksize (const char *name, unsigned long *pvalue)
{
  char* env = getenv(name);
  if (env ==0) return false;
  if (kaapi_parse_size( &env, pvalue) && (*env == 0))
    return true;

  fprintf (stderr, "[libKOMP] Wrong stacksize value '%s' for '%s'!\n", getenv(name), name);
  return false;
}


static bool
komp_parse_unsigned_int( char* name, unsigned int* pvalue )
{
  char* env = getenv(name);
  if (env ==0) return false;
  if (kaapi_parse_unsigned_int( &env, pvalue) && (*env == 0))
    return true;

  fprintf (stderr, "[libKOMP] Wrong value '%s' for '%s'!\n", getenv(name), name);
  return false;
}


static bool
komp_parse_int( char* name, int* pvalue )
{
  char* env = getenv(name);
  if (env ==0) return false;
  if (kaapi_parse_int( &env, pvalue) && (*env == 0))
    return true;

  fprintf (stderr, "[libKOMP] Wrong value for '%s'!\n", name);
  return false;
}


static bool
komp_parse_list_unsigned_int (char* name, unsigned int **pvalue)
{
  char* env = getenv(name);
  unsigned int count;
  if (env ==0) return false;
  if (kaapi_parse_list_unsigned_int( &env, &count, pvalue) && (*env == 0))
    return true;

  fprintf (stderr, "[libKOMP] Wrong value for '%s'!\n", name);
  return false;
}


static bool
komp_parse_list_procbind (char* name, komp_proc_bind_t **pvalue)
{
  char* env = getenv(name);
  unsigned int count;
  if (env ==0) return false; 
  if (kaapi_parse_list_procbind( &env, &count, (kaapi_procbind_t**)pvalue) && (*env == 0))
  {
    /* alway one more place free in the list */
    (*pvalue)[count] = komp_proc_bind_end_list;
    return true;
  }
  *pvalue = 0;
  fprintf (stderr, "[libKOMP] Wrong value for '%s'!\n", name);
  return false;
}


static const char* komp_unparse_procbind(komp_proc_bind_t value)
{
  switch (value) {
    case komp_proc_bind_false: return "false";
    case komp_proc_bind_true: return "true";
    case komp_proc_bind_master: return "master";
    case komp_proc_bind_close: return "close";
    case komp_proc_bind_spread: return "spread";

    default:
      fprintf(stderr,"[libKOMP] bad value: %i for procbind code\n", value);
      exit(1);
  }
}

char komp_unparse_procbind_list_buffer[4096];
static const char* komp_unparse_procbind_list(komp_dataenv_icv_t* icvs)
{
  komp_proc_bind_t* bindproc_list;
  char* pos;
  int cnt;

  pos = &komp_unparse_procbind_list_buffer[0];
  komp_unparse_procbind_list_buffer[0] = 0;
  cnt = sprintf(pos,"%s", komp_unparse_procbind(icvs->h_bind));
  kaapi_assert(cnt >0);
  pos += cnt;


  bindproc_list = icvs->binds;
  while ((bindproc_list !=0) && (*bindproc_list !=0))
  {
    cnt = sprintf(pos,",%s",komp_unparse_procbind(*bindproc_list));
    pos += cnt;
    kaapi_assert(cnt >0);
    ++bindproc_list;
  }
  return komp_unparse_procbind_list_buffer;
}


static bool
komp_parse_bool( char* name, int* pvalue )
{
  char* env = getenv(name);
  if (env ==0) return false; 
  if (kaapi_parse_bool( &env, pvalue) && (*env == 0))
    return true;

  fprintf (stderr, "[libKOMP] Wrong value for '%s'! Should be set to one of (true|false).\n", name);
  return false;
}


/* Parse the active|passive environment variable. */
static bool
parse_waitpolicy (char* env, bool* pvalue)
{
  if (env == NULL)
    return false;

  if (strcasecmp (env, "passive") == 0)
  {
    *pvalue = true;
    return true;
  }
  if (strcasecmp (env, "active") == 0)
  {
    *pvalue = false;
    return true;
  }
  return false;
}

static bool
komp_parse_waitpolicy( const char* name, bool* pvalue )
{
  char* env = getenv(name);
  if (env ==0) return false; 
  if (!parse_waitpolicy( env, pvalue))
  {
    fprintf (stderr, "[libKOMP] Wrong value for '%s'! Should be set to one of (passive | active).\n",
                   name);
    return false;
  }
  return true;
}

static bool
komp_parse_perfcounter( const char* name, kaapi_display_perf_value_t* pvalue )
{
  char* env = getenv(name);
  if (env ==0) return false; 
  if (kaapi_parse_perfcounter( &env, pvalue) && (*env ==0))
    return true;

  fprintf (stderr, "[libKOMP] Wrong value for '%s'! Should be set to one of (no | full | resume | gplot).\n",
                 name);
  return false;
}


static bool
komp_parse_stealprotocol( const char* name, kaapi_rtparam_t* rt_param )
{
  char* env = getenv(name);
  if (env ==0) return false; 
  if (kaapi_parse_stealprotocol( &env, rt_param) && (*env ==0))
    return true;

  fprintf (stderr, "[libKOMP] Wrong value for '%s'! Should be set to one of (ccsync | noaggr).\n",
                   name);
  return false;
}

static inline const char *
komp_env_defined (char* name)
{
  if (getenv(name) != 0)
    return "user-defined";
  return "libKOMP default";
}

static const char* komp_nthreads2str( unsigned int h_nthread, unsigned int* nthreads)
{
  static char buffer[1024];
  char* data = buffer;
  int first = 1;
  if (h_nthread ==0) return "<empty>";
  data += snprintf(data, 1024-(data-buffer), "%u", h_nthread );
  if (nthreads ==0) return buffer;
  do {
    if (first)
    {
      data += snprintf(data, 1024-(data-buffer), "%u", *nthreads);
      first = 0;
    }
    else
      data += snprintf(data, 1024-(data-buffer), ",%u", *nthreads);
    ++nthreads;
  } while (*nthreads != 0);
  return buffer;
}

static const char* komp_unparse_schedule(komp_sched_t sched, unsigned long run_sched_modifier)
{
  static char buffer[1024];
  char* data = buffer;
  switch (sched) {
    case komp_sched_auto:
      data += sprintf(data, "auto");
      break;
    case komp_sched_dynamic:
      data += sprintf(data, "auto");
      break;
    case komp_sched_static:
      data += sprintf(data, "auto");
      break;
    case komp_sched_guided:
      data += sprintf(data, "auto");
      break;
    case komp_sched_adaptive:
      data += sprintf(data, "auto");
      break;
    case komp_sched_steal:
      data += sprintf(data, "auto");
      break;
  }
  if (run_sched_modifier !=0)
  {
    switch (sched) {
      case komp_sched_auto:
      case komp_sched_dynamic:
      case komp_sched_static:
      case komp_sched_guided:
      case komp_sched_adaptive:
      case komp_sched_steal:
        printf(data,",%lu",run_sched_modifier);
        break;
    }
  }
  return buffer;
}



/* Display runtime-related information as a banner printed before
   running the OpenMP application. */
static void
komp_display_env (int mode)
{
  if (!mode) return;

  fprintf(stdout, "OPENMP DISPLAY ENVIRONMENT BEGIN\n");
  fprintf (stdout, "  OMP_NUM_THREADS\t=\t%s\t(%s)\n",
	   komp_nthreads2str(komp_global_dataenv_icv.h_nthreads, komp_global_dataenv_icv.nthreads),
     komp_env_defined("OMP_NUM_THREADS")
  );
  fprintf (stdout, "  OMP_THREAD_LIMIT\t=\t%u\t(%s)\n",
	   komp_global_dataenv_icv.thread_limit,
     komp_env_defined("OMP_THREAD_LIMIT")
  );
  fprintf (stdout, "  OMP_STACKSIZE\t\t=\t%i\t(%s)\n",
	   (int)komp_global_device_icv.stacksize,
     komp_env_defined("OMP_STACKSIZE")
  );
  fprintf (stdout, "  OMP_MAX_ACTIVE_LEVELS\t=\t%i\t(%s)\n",
	   (int)komp_global_device_icv.max_active_levels,
     komp_env_defined("OMP_MAX_ACTIVE_LEVELS")
  );
  fprintf (stdout, "  OMP_NESTED\t\t=\t%s\t(%s)\n",
	   (komp_global_dataenv_icv.nest ? "true" : "false"),
     komp_env_defined("OMP_NESTED")
  );
  fprintf (stdout, "  OMP_DYNAMIC\t\t=\t%s\t(%s)\n",
	   (komp_global_dataenv_icv.dyn ? "true" : "false"),
     komp_env_defined("OMP_DYNAMIC")
  );
  fprintf (stdout, "  OMP_PROC_BIND\t\t=\t%s\t(%s)\n",
	   komp_unparse_procbind_list(&komp_global_dataenv_icv),
     komp_env_defined("OMP_PROC_BIND")
  );
  fprintf (stdout, "  OMP_PLACES\t\t=\t%s\t(%s)\n",
	   kaapi_unparse_places( (unsigned int)kaapi_default_param.places.len, kaapi_default_param.places.first ),
     komp_env_defined("OMP_PLACES")
  );
  fprintf (stdout, "  OMP_SCHEDULE\t\t=\t%s\t(%s)\n",
	   komp_unparse_schedule(komp_global_dataenv_icv.run_sched,
                        komp_global_dataenv_icv.run_sched_modifier),
     komp_env_defined("OMP_SCHEDULE")
  );
  fprintf (stdout, "  OMP_CANCELLATION\t=\t%s\t(%s)\n",
	   (komp_global_device_icv.cancel ? "true" : "false"),
     komp_env_defined("OMP_NESTED")
  );
  fprintf (stdout, "  KOMP_DISPLAY_PERF\t=\t%s\t(%s)\n",
	   (   kaapi_default_param.display_perfcounter == KAAPI_NO_DISPLAY_PERF ? "no"
       : kaapi_default_param.display_perfcounter == KAAPI_DISPLAY_PERF_FULL ? "full"
       : kaapi_default_param.display_perfcounter == KAAPI_DISPLAY_PERF_RESUME ? "resume"
       : kaapi_default_param.display_perfcounter == KAAPI_DISPLAY_PERF_GPLOT ? "gplot"
       : "bad value" ),
     komp_env_defined("KOMP_DISPLAY_PERF")
  );
  fprintf (stdout, "  KOMP_STEAL_PROTOCOL\t=\t%s\t(%s)\n",
	   (   kaapi_default_param.emitsteal == kaapi_sched_noaggr_emitsteal ? "noaggr"
#if defined(KAAPI_USE_FLAT_CC_SYNC)
       : kaapi_default_param.emitsteal == kaapi_sched_ccsync_emitsteal ? "ccsync"
#endif
       : "bad value"
     ),
     komp_env_defined("KOMP_STEAL_PROTOCOL")
  );

  fprintf(stdout, "OPENMP DISPLAY ENVIRONMENT END\n");
}

/* Parse the OMP_DIPLAY_ENV environment variable.
   This feature comes from OpenMP 4 RC2.

   As described in section 4.12, OMP_DISPLAY_ENV can be set to one of
   these values: 
   true | false | verbose

   and display runtime-related information accordingly (ICVs, version
   number, ...).

   Note that, as opposed to the current version of the 4RC2 specs,
   lowercase characters have been chosen to fit the style of other
   OpenMP variables like OMP_NESTED.
 */
static bool
komp_parse_display_env (void)
{
  char *env = getenv ("OMP_DISPLAY_ENV");
  if (env == NULL)
    return false;

  int value;
  if (kaapi_parse_display_env(&env, &value) && (*env == 0))
  {
    if (value != 0)
      komp_display_env (value);
    return true;
  }

  fprintf (stderr, "[libKOMP] Wrong value for OMP_DISPLAY_ENV (%s)! "
                     "Should be set to one of (true | false | verbose).\n", env);
  return false;
}


static bool
komp_parse_schedule (char* name, komp_sched_t* prun_sched, unsigned long* prun_sched_modifier)
{
  char* env = getenv (name);
  unsigned long run_sched_modifier = 0;
  if (env == NULL)
    return false;

  kaapi_foreach_attr_policy_t kaapi_run_sched;
  if (kaapi_parse_schedule(&env, &kaapi_run_sched, &run_sched_modifier) && (*env ==0))
  {
    *prun_sched_modifier = run_sched_modifier;
    *prun_sched = (kaapi_run_sched == KAAPI_FOREACH_SCHED_AUTO ?
                      komp_sched_auto :
                      (komp_sched_t)kaapi_run_sched
    );
    return true;
  }
  fprintf (stderr, "[libKOMP] Wrong value for %s! "
               "Should be set to one of (static|dynamic|auto|guided|adaptive|steal)[,chunksize]\n", name);
  return false;
}


/* to free dynamic memory */
unsigned int* l_alloc1 = 0;

/* guard against multiple initialization */
static kaapi_atomic_t komp_isinit = { 0 };

/*
*/
void komp_init (void)
{
  if (KAAPI_ATOMIC_INCR(&komp_isinit) !=1) 
    return;

  /* */
  kaapi_mt_hwdetect();

  /* KOMP_CPU_AFFINITY has higher priority with GOMP_CPU_AFFINITY */
  if (0 != getenv("KOMP_CPU_AFFINITY"))
    kaapi_default_param.cpuset_str = getenv("KOMP_CPU_AFFINITY");

  /* OMP_NUM_THREADS. */
  komp_parse_list_unsigned_int("OMP_NUM_THREADS", &komp_global_dataenv_icv.nthreads);
  kaapi_default_param.cpucount_str = getenv("OMP_NUM_THREADS");

  l_alloc1 = komp_global_dataenv_icv.nthreads;

  /* Turn GOMP_STACKSIZE and OMP_STACKSIZE to control the pthread stack size */
  if (komp_parse_stacksize ("OMP_STACKSIZE", &komp_global_device_icv.stacksize)
      || komp_parse_stacksize ("GOMP_STACKSIZE", &komp_global_device_icv.stacksize))
  {
    kaapi_default_param.threadstacksize = komp_global_device_icv.stacksize;
  }
  else {
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_getstacksize( &attr, &komp_global_device_icv.stacksize );
    kaapi_default_param.threadstacksize = komp_global_device_icv.stacksize;
  }

  /* Parse OMP_MAX_ACTIVE_LEVELS. */
  komp_parse_int ("OMP_MAX_ACTIVE_LEVELS", &komp_global_device_icv.max_active_levels);

  /* Parse OMP_NESTED. */
  komp_parse_bool("OMP_NESTED", &komp_global_dataenv_icv.nest);

  /* Parse OMP_DYNAMIC. */
  komp_parse_bool("OMP_DYNAMIC", &komp_global_dataenv_icv.dyn);

  /* Parse OMP_DYNAMIC. */
  komp_parse_bool("OMP_CANCELLATION", &komp_global_device_icv.cancel);

  /* Parse OMP_PROC_BIND */
  komp_parse_list_procbind("OMP_PROC_BIND", &komp_global_dataenv_icv.binds);
  if (komp_global_dataenv_icv.binds ==0)
    komp_global_dataenv_icv.h_bind = komp_proc_bind_false;
  else
  {
    komp_global_dataenv_icv.h_bind = *komp_global_dataenv_icv.binds;
    ++komp_global_dataenv_icv.binds;
    if (*komp_global_dataenv_icv.binds == komp_proc_bind_end_list)
      komp_global_dataenv_icv.binds = 0;
  }

  /* FIXEME : currently KAAPI should be used with OMP_PLACES defined ... 
  */
  if (0 == getenv("OMP_PLACES"))
  {
    kaapi_mt_hwdetect();
    int nthreads = ( komp_global_dataenv_icv.nthreads == 0 ?
        kaapi_default_param.syscpucount
      : *komp_global_dataenv_icv.nthreads );
    kaapi_assert( nthreads >0 );
    char tmp[32];
    snprintf(tmp, 32, "threads(%i)", nthreads);
    setenv("OMP_PLACES",tmp, 1);
  }

  /* OMP_PLACES has lower priority than KOMP_CPU_AFFINITY or GOMP_CPU_AFFINITY */
  if ((0 != getenv("OMP_PLACES")) && (kaapi_default_param.cpuset_str ==0))
  {
    kaapi_default_param.cpuset_str = getenv("OMP_PLACES");
    if (getenv("OMP_PROC_BIND")==0) komp_global_dataenv_icv.h_bind = komp_proc_bind_true;
  }

  bool activeon = true;
  komp_parse_waitpolicy("OMP_WAIT_POLICY", &activeon );
  if (activeon)
    komp_global_device_icv.waitpolicy = komp_wait_policy_active;
  else
    komp_global_device_icv.waitpolicy = komp_wait_policy_passive;
  
  /* OMP_THREAD_LIMIT: unsigned int */
  komp_parse_unsigned_int("OMP_THREAD_LIMIT", &komp_global_dataenv_icv.thread_limit);

  /* Parse OMP_SCHEDULE */
  komp_parse_schedule ("OMP_SCHEDULE",
    &komp_global_dataenv_icv.run_sched,
    &komp_global_dataenv_icv.run_sched_modifier);

  /* Parse KOMP env variables */
  komp_parse_perfcounter("KOMP_DISPLAY_PERF", &kaapi_default_param.display_perfcounter);

  /* Parse KOMP env variables */
  komp_parse_stealprotocol("KOMP_STEAL_PROTOCOL", &kaapi_default_param);

  KOMP_INIT_FMT(schar)
  KOMP_INIT_FMT(char)
  KOMP_INIT_FMT(shrt)
  KOMP_INIT_FMT(int)
  KOMP_INIT_FMT(long)
  KOMP_INIT_FMT(llong)
  KOMP_INIT_FMT(int8)
  KOMP_INIT_FMT(int16)
  KOMP_INIT_FMT(int32)
  KOMP_INIT_FMT(int64)
  KOMP_INIT_FMT(uchar)
  KOMP_INIT_FMT(ushrt)
  KOMP_INIT_FMT(uint)
  KOMP_INIT_FMT(ulong)
  KOMP_INIT_FMT(ullong)
  KOMP_INIT_FMT(uint8)
  KOMP_INIT_FMT(uint16)
  KOMP_INIT_FMT(uint32)
  KOMP_INIT_FMT(uint64)
  KOMP_INIT_FMT(flt)
  KOMP_INIT_FMT(dbl)
  KOMP_INIT_FMT(ldbl)
  KOMP_INIT_FMT(voidp)
  

  /* init kaapi without parsing KAAPI_XXX environement variables */
  kaapi_mt_init(KAAPI_START_ONLY_MAIN);

  if (komp_global_dataenv_icv.nthreads ==0)
    komp_global_dataenv_icv.h_nthreads = kaapi_default_param.cpucount;
  else
  {
    komp_global_dataenv_icv.h_nthreads = *komp_global_dataenv_icv.nthreads;
    ++komp_global_dataenv_icv.nthreads;
    if (komp_global_dataenv_icv.nthreads[0] ==0)
      komp_global_dataenv_icv.nthreads = 0;
  }

  /* Parse OMP_DISPLAY_ENV and display informations, if any */
  komp_parse_display_env ();
}


void komp_finalize (void)
{
  if (KAAPI_ATOMIC_DECR(&komp_isinit) !=0) 
    return;

  kaapi_mt_finalize ();

  if (l_alloc1 !=0)
  {
    free(l_alloc1);
    l_alloc1 = 0;
  }

#if KAAPI_KOMP_TRACE
  printf("KOMP: time start parallel: %f\n", 1e6*kaapi_komp_start_parallel/kaapi_komp_start_parallel_count); 
  printf("KOMP: time end parallel: %f\n", 1e6*kaapi_komp_end_parallel/kaapi_komp_end_parallel_count); 
#endif
}


/*
*/
struct komp_thread_t* komp_self_thread(void)
{ 
  komp_thread_t* thread = (komp_thread_t*)kaapi_self_thread(); 
  if (thread !=0) return thread;
  kaapi_processor_t* kproc = kaapi_create_orphan_kprocessor();
  return (komp_thread_t*)kaapi_context2thread(kproc->context);
}


/*
*/
struct komp_team_t* komp_self_team(void)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
  if (kproc ==0) return 0;
  return (struct komp_team_t*)kproc->team;
}


/*
*/
komp_dataenv_icv_t* komp_self_icvs(void)
{
#if KOMP_USE_ICV
  kaapi_processor_t* kproc = kaapi_self_processor();
  komp_dataenv_icv_t* icv = komp_get_dataenv_icv( kproc );
#else
  komp_dataenv_icv_t* icv = 0;
#endif
  return icv;
}


/*
*/
komp_dataenv_icv_t* komp_get_dataenv_icv(kaapi_processor_t* proc)
{
#if KOMP_USE_ICV
  struct kaapi_context_t* thread;
  if ((proc ==0) || (proc->context ==0))
    return &komp_global_dataenv_icv;

  thread = proc->context;
  if (thread->komp_icvs == 0)
  {
    thread->komp_icvs = &komp_global_dataenv_icv;
    return &komp_global_dataenv_icv;
  }
  return (komp_dataenv_icv_t*)thread->komp_icvs;
#else
  return &komp_global_dataenv_icv;
#endif
}


/* Used for task creation !
*/
komp_dataenv_icv_t* komp_inittask_dataenv_icvs(
  struct komp_thread_t* th, int copy
)
{
  struct kaapi_thread_t* thread = (kaapi_thread_t*)th;
  komp_dataenv_icv_t* newicvs;
  kaapi_context_t* context = kaapi_thread2context(thread);

  if (!copy)
    return context->komp_icvs;

  if (thread !=0)
    newicvs = (komp_dataenv_icv_t*)kaapi_alloca( thread, sizeof(komp_dataenv_icv_t) );
  else
    kaapi_abort(__LINE__,__FILE__, "Invalid case");

  kaapi_team_t* team = context->proc->team;
  komp_dataenv_icv_t* curr_icvs = newicvs = komp_get_dataenv_icv(context->proc);
  *newicvs = *curr_icvs;
  newicvs->team = (struct komp_team_t*)team;
  newicvs->prev_icvs = curr_icvs;
  return newicvs;
}

/* for task
*/
komp_dataenv_icv_t* komp_swap_dataenv_icvs(
  struct komp_thread_t* th,
  komp_dataenv_icv_t* newicvs
)
{
  struct kaapi_thread_t* thread = (kaapi_thread_t*)th;
  kaapi_context_t* context = kaapi_thread2context(thread);
  komp_dataenv_icv_t* old = (komp_dataenv_icv_t*)context->komp_icvs;
  context->komp_icvs = newicvs;
  return old;
}


/* create a copy of the ICVS
*/
komp_dataenv_icv_t* komp_newdataenv_icvs(
  kaapi_processor_t* kproc,
  komp_dataenv_icv_t* parent_icvs
)
{
  kaapi_context_t* context = (kproc== 0 ? 0 : kproc->context);
  kaapi_team_t* team = (kproc== 0 ? 0 : kproc->team);
  komp_dataenv_icv_t* newicvs;
  if (context !=0)
    newicvs = (komp_dataenv_icv_t*)kaapi_alloca( kaapi_context2thread(context), sizeof(komp_dataenv_icv_t) );
  else
    newicvs = (komp_dataenv_icv_t*)malloc( sizeof(komp_dataenv_icv_t) );

  *newicvs = *parent_icvs;
  newicvs->team = (struct komp_team_t*)team;
  newicvs->prev_icvs = parent_icvs;

  newicvs->levels = 1+parent_icvs->levels;
  if ((team !=0) && (team->count >1))
    newicvs->active_levels = 1+parent_icvs->active_levels;
  else
    newicvs->active_levels = parent_icvs->active_levels;
  return newicvs;
}


/* kaapi_data_push
*/
void* komp_data_push( komp_thread_t* th, unsigned long size )
{
  return kaapi_data_push((kaapi_thread_t*)th, size);
}


/* kaapi_task_push
*/
int komp_task_push(komp_thread_t* th, komp_task_body_t body, void* arg, int ifClause)
{
  kaapi_thread_t* thread = (kaapi_thread_t *)th;
  kaapi_task_t* task = thread->sp;
  if (!_kaapi_has_enough_taskspace(thread,1))
    task = kaapi_thread_slow_pushtask(thread);
  task->body      = (kaapi_task_body_t)body;
  task->state.steal     = (kaapi_task_body_t)body;
  task->arg       = arg;
  task->u.dummy   = 0;
#if !defined(NDEBUG)
  task->next      = 0;
  KAAPI_ATOMIC_WRITE( &task->wc, 0);
#endif
  if (ifClause)
  {
#if KOMP_USE_ICV
    komp_dataenv_icv_t* icvs = komp_get_dataenv_icv( kaapi_thread2context(thread)->proc );
    if (icvs->prio)
    {
      task->u.s.priority = icvs->prio;
      task->u.s.flag |= KAAPI_TASK_FLAG_PRIORITY;
      icvs->prio = 0; /* reset for next task */
    }
    if (icvs->preferred_ld)
    {
//kaapi_assert(icvs->preferred_ld <8+1);
      task->u.s.site     = (uint8_t)icvs->preferred_ld;
      icvs->preferred_ld = 0; /* reset for next task */
    }
    if (icvs->flag & KAAPI_TASK_FLAG_LD_BOUND)
      task->u.s.flag |= KAAPI_TASK_FLAG_LD_BOUND;
    if (icvs->flag & KOMP_FLAG_LLC_AWARE)
      task->u.s.flag |= KAAPI_TASK_FLAG_LLC_AWARE;
    icvs->flag = 0; /* reset for next task */
#endif
    kaapi_writemem_barrier();
    ++thread->sp;
  }
  else
  {
    /* required if dependencies exists with created task and the previous tasks */
    kaapi_sched_sync(thread);
    body((komp_task_t*)task, th);
  }
  return 0;
}


void komp_taskgroup_begin(void)
{
  kaapi_context_t* self_context = kaapi_self_context();
  komp_taskgroup_t* task_group
    = (komp_taskgroup_t*)kaapi_data_push(
            kaapi_context2thread(self_context),
            sizeof(komp_taskgroup_t)
  );
  komp_dataenv_icv_t* icvs = self_context->komp_icvs;
  task_group->prev = icvs->task_group;
  task_group->cancelled = 0;
  //task_group->frame = kaapi_stack_push_frame( &self_context->stack );
  icvs->task_group = task_group;

  /* do notthing more : if cancellation is considered, then the taskgroup should
     be named to all childs tasks between _start and _end.
  */
}


void komp_taskgroup_end (void)
{
  kaapi_context_t* self_context = kaapi_self_context();
  kaapi_sched_sync( kaapi_self_thread() );
  komp_dataenv_icv_t* icvs = self_context->komp_icvs;
  komp_taskgroup_t* tg = icvs->task_group;
  icvs->task_group = tg->prev;
  //kaapi_stack_pop_frame(&self_context->stack, tg->frame );
}


/*
*/
void komp_atomic_initlock( struct komp_mutex_t* l)
{
  l->ptr = malloc( sizeof(kaapi_lock_t) );
  kaapi_atomic_initlock(l->ptr);
}

/*
*/
void komp_atomic_destroylock( struct komp_mutex_t* l)
{
  kaapi_atomic_destroylock(l->ptr);
  free(l->ptr);
}

void komp_atomic_lock( struct komp_mutex_t* l )
{
    kaapi_atomic_lock(l->ptr);
}

void komp_atomic_unlock( struct komp_mutex_t* l )
{
    kaapi_atomic_unlock(l->ptr);
}

/*
*/
void komp_barrier( struct komp_team_t* team )
{
  if (team ==0) return;
  kaapi_team_barrier_wait((kaapi_team_t*)team, kaapi_self_processor(), KAAPI_BARRIER_FLAG_DEFAULT);
}

/*
*/
void komp_sched_sync( struct komp_thread_t* thread)
{
  if (thread ==0) return;
  kaapi_sched_sync((kaapi_thread_t*)thread);
}

/*
*/
int komp_lock ( struct komp_team_t* team )
{ return kaapi_team_lock((kaapi_team_t*)team, 1); }

/*
*/
int komp_unlock ( struct komp_team_t* team )
{ return kaapi_team_unlock((kaapi_team_t*)team, 1); }


/*
*/
int komp_master ( struct komp_team_t* team )
{
  if (team ==0) return 1;
  return kaapi_team_master((kaapi_team_t*)team,0);
}


/* Execute a parallel region with body the body of the region
   - similar interface than in Intel IOMP
*/
typedef struct {
  void (*body)(void*);
  void* arg;
  komp_dataenv_icv_t* parent_icvs;
} trampoline_forkcall_arg_t;

static void trampoline_forkcall_wrapper(void* a)
{
  trampoline_forkcall_arg_t* arg = (trampoline_forkcall_arg_t*)a;

#if KOMP_USE_ICV
  kaapi_processor_t* kproc = kaapi_self_processor();
  kaapi_context_t* context = kproc->context;
  komp_dataenv_icv_t* newicvs = komp_newdataenv_icvs( kproc, arg->parent_icvs );
  komp_update_implicit_icvs(newicvs);
  context->komp_icvs = newicvs;
#endif

  arg->body(arg->arg);

#if KOMP_USE_ICV
  context->komp_icvs = 0;
#endif
}



/*
*/
int komp_begin_parallel(
  unsigned int     num_threads,
  komp_proc_bind_t procbind,
  void (*body)(void*),
  void* data
)
{
  kaapi_processor_t* kproc;
#if KOMP_USE_ICV
  kproc = kaapi_self_processor();
  komp_dataenv_icv_t* icvs = komp_get_dataenv_icv( kproc );
  num_threads = komp_get_thread_parallel_region( num_threads, 1, icvs );
  if (procbind == komp_proc_bind_default)
    procbind = icvs->h_bind;
#endif
  /* what does means true here ? convert to default that has mapping to kaapi */
  if (procbind == komp_proc_bind_true)
    procbind = komp_proc_bind_default;

  kaapi_team_t* team = kaapi_begin_parallel( num_threads, (kaapi_procbind_t)procbind );

  trampoline_forkcall_arg_t* arg = 0;
  if (body)
  {
    arg = (trampoline_forkcall_arg_t*)&team->ws.data[0];
    arg->body = body;
    arg->arg  = data;
#if KOMP_USE_ICV
    arg->parent_icvs = icvs;
#endif
  }

  /* start parallel team, except the current context */
  kaapi_start_parallel(team, body == 0 ? 0 : trampoline_forkcall_wrapper, arg, 0);

  kproc = kaapi_self_processor();
  kaapi_context_t* context = kproc->context;
#if KOMP_USE_ICV
  /* set the ICV to the newly created context */
  context->komp_icvs = komp_newdataenv_icvs( kproc, icvs );
  komp_update_implicit_icvs(context->komp_icvs);
#endif

  if (body)
    body(data);

  return 0;
}


/**
*/
int komp_end_parallel(void)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
  kaapi_context_t* context = kproc->context;

  if (context->ws.count_for_end != context->ws.count_for)
    komp_loop_workfini( (komp_thread_t*)kaapi_context2thread(context) );
  kaapi_assert_debug(context->ws.count_for_end == context->ws.count_for);

  kaapi_end_parallel(kproc->team);

  return 0;
}


/** Keep it until begin/end debug.
    == begin_parallel(num_threads, procbind, body, data); end_parallel();
*/
void komp_fork_call(
  unsigned int num_threads,
  komp_proc_bind_t procbind,
  void (*body)(void*),
  void* data
)
{
  kaapi_processor_t* kproc;
#if KOMP_USE_ICV
  kproc = kaapi_self_processor();
  komp_dataenv_icv_t* icvs = komp_get_dataenv_icv( kproc );
  num_threads = komp_get_thread_parallel_region( num_threads, 1, icvs );
  if (procbind == komp_proc_bind_default)
    procbind = icvs->h_bind;
#endif
  /* what does means true here ? convert to default that has mapping to kaapi */
  if (procbind == komp_proc_bind_true)
    procbind = komp_proc_bind_default;

  kaapi_team_t* team = kaapi_begin_parallel( num_threads, (kaapi_procbind_t)procbind );

  trampoline_forkcall_arg_t* arg = (trampoline_forkcall_arg_t*)&team->ws.data[0];
  arg->body = body;
  arg->arg  = data;
#if KOMP_USE_ICV
  arg->parent_icvs = icvs;
#endif

  /* start parallel team, except the current context */
  kaapi_start_parallel(team, trampoline_forkcall_wrapper, arg, 0 );

  kproc = kaapi_self_processor();
  kaapi_context_t* context = kproc->context;
#if KOMP_USE_ICV
  /* set the ICV to the newly created context */
  context->komp_icvs = komp_newdataenv_icvs( kproc, icvs );
  komp_update_implicit_icvs(context->komp_icvs);
#endif

  body(data);

#if 0
{
  static int counter = 0;
  char filename[128];
  snprintf(filename,128,"/tmp/%s_forkcall_%i.dot",getenv("USER"), ++counter);
  FILE* file = fopen(filename,"w");
  kaapi_assert_debug(file !=0);
  kaapi_context_print_dot( file, context);
  fclose(file);
}
#endif

  if (context->ws.count_for_end != context->ws.count_for)
    komp_loop_workfini( (komp_thread_t*)kaapi_context2thread(context) );
  kaapi_assert_debug(context->ws.count_for_end == context->ws.count_for);

  kaapi_end_parallel(team);
}


/*
*/
int komp_single (struct komp_team_t* team)
{
  if (team ==0) return 1;
  return kaapi_team_single((kaapi_team_t*)team,0);
}


/*
*/
void* komp_single_copy_start( struct komp_team_t* team )
{
  if ((team ==0) || kaapi_team_single((kaapi_team_t*)team, 0))
    return 0;

  kaapi_team_barrier_wait((kaapi_team_t*)team, kaapi_self_processor(), KAAPI_BARRIER_FLAG_DEFAULT);
  return ((kaapi_team_t*)team)->ws.single_data;
}


/*
*/
void komp_single_copy_end( struct komp_team_t* team, void* data)
{
  if (!team) return;
  ((kaapi_team_t*)team)->ws.single_data = data;
  kaapi_team_barrier_wait((kaapi_team_t*)team, kaapi_self_processor(), KAAPI_BARRIER_FLAG_NOSCHEDULE);
  /* unlink threads blocked in komp_single_copy_start */
}



/*
*/
void komp_mem_barrier(void)
{
  kaapi_mem_barrier();
}


/* kaapi_format_allocate
*/
struct komp_format_t* komp_format_allocate(void)
{
  return (struct komp_format_t*) kaapi_format_allocate();
}


/* kaapi_format_structregister
*/
int komp_format_structregister(
  struct komp_format_t*        fmt,
  const char*                  name,
  unsigned long                size,
  void                       (*cstor)( void* ),
  void                       (*dstor)( void* ),
  void                       (*cstorcopy)( void*, const void*),
  void                       (*copy)( void*, const void*),
  void                       (*assign)( void*, const komp_memory_view_t*, const void*, const komp_memory_view_t*)
)
{
  return kaapi_format_structregister( (struct kaapi_format_t*)fmt,
        name,
        size,
        cstor,
        dstor,
        cstorcopy,
        copy,
        (void (*)( void*, const kaapi_memory_view_t*, const void*, const kaapi_memory_view_t*)
) assign
  );
}

/* kaapi_format_taskregister_func
*/
int komp_format_taskregister_func(
  struct komp_format_t*         fmt,
  komp_task_body_t              body,
  komp_task_body_t              bodywh,
  const char*                   name,
  komp_fmt_fnc_get_name         get_name,
  komp_fmt_fnc_get_size         get_size,
  komp_fmt_fnc_task_copy        task_copy,
  komp_fmt_fnc_get_count_params get_count_params,
  komp_fmt_fnc_get_mode_param   get_mode_param,
  komp_fmt_fnc_get_data_param   get_data_param,
  komp_fmt_fnc_get_access_param get_access_param,
  komp_fmt_fnc_set_access_param set_access_param,
  komp_fmt_fnc_get_fmt_param    get_fmt_param,
  komp_fmt_fnc_get_view_param   get_view_param,
  komp_fmt_fnc_set_view_param   set_view_param,
  komp_fmt_fnc_reducor          reducor,
  komp_fmt_fnc_redinit          redinit,
  komp_fmt_fnc_get_splitter	    get_splitter
)
{
  return kaapi_format_taskregister_func(
    (struct kaapi_format_t*)         fmt,
    (kaapi_task_body_t)              body,
                                     name,
    (kaapi_fmt_fnc_get_name)         get_name,
    (kaapi_fmt_fnc_get_size)         get_size,
    (kaapi_fmt_fnc_task_copy)        task_copy,
    (kaapi_fmt_fnc_get_count_params) get_count_params,
    (kaapi_fmt_fnc_get_mode_param)   get_mode_param,
    (kaapi_fmt_fnc_get_data_param)   get_data_param,
    (kaapi_fmt_fnc_get_access_param) get_access_param,
    (kaapi_fmt_fnc_set_access_param) set_access_param,
    (kaapi_fmt_fnc_get_fmt_param)    get_fmt_param,
    (kaapi_fmt_fnc_get_view_param)   get_view_param,
    (kaapi_fmt_fnc_set_view_param)   set_view_param,
    (kaapi_fmt_fnc_reducor)          reducor,
    (kaapi_fmt_fnc_redinit)          redinit,
    (kaapi_fmt_fnc_get_splitter)	    get_splitter
  );
}

/*
*/
int komp_foreach_attr_set_policy(
  komp_foreach_attr_t* attr, 
  komp_foreach_attr_policy_t policy
)
{
  return
    kaapi_foreach_attr_set_policy((kaapi_foreach_attr_t*)attr, (kaapi_foreach_attr_policy_t)policy);
}

/*
*/
int komp_foreach_attr_get_policy(
  const komp_foreach_attr_t* attr,
  komp_foreach_attr_policy_t* policy
)
{
  return
    kaapi_foreach_attr_get_policy((kaapi_foreach_attr_t*)attr, (kaapi_foreach_attr_policy_t*)policy);
}


/*
*/
int komp_foreach_attr_set_grains(
  komp_foreach_attr_t* attr, 
  unsigned long long s_grain,
  unsigned long long p_grain
)
{
  return
    kaapi_foreach_attr_set_grains((kaapi_foreach_attr_t*)attr, s_grain, p_grain);
}


/*
*/
int komp_foreach_attr_get_grains(
  const komp_foreach_attr_t* attr,
  unsigned long long* s_grain,
  unsigned long long* p_grain
)
{
  return
    kaapi_foreach_attr_get_grains((kaapi_foreach_attr_t*)attr, s_grain, p_grain);
}


/*
*/
int komp_foreach_attr_set_localitydomain(
  komp_foreach_attr_t* attr, 
  komp_hws_levelid_t  level,
  int                  strictness
)
{
  return
    kaapi_foreach_attr_set_localitydomain((kaapi_foreach_attr_t*)attr, (kaapi_hws_levelid_t)level, strictness);
}


/*
*/
int komp_foreach_attr_get_localitydomain(
  const komp_foreach_attr_t* attr,
  komp_hws_levelid_t*       level,
  int*                       strictness
)
{
  return
    kaapi_foreach_attr_get_localitydomain((kaapi_foreach_attr_t*)attr, (kaapi_hws_levelid_t*)level, strictness);
}


/*
*/
int komp_foreach_attr_set_distribution(
  komp_foreach_attr_t* attr, 
  komp_hws_levelid_t   level,
  int                  strict
)
{
  return kaapi_foreach_attr_set_distribution(
        (kaapi_foreach_attr_t*)attr,
        (kaapi_hws_levelid_t)level,
        strict );
}

/*
*/
int komp_foreach_attr_get_distribution(
  const komp_foreach_attr_t* attr,
  komp_hws_levelid_t*        level,
  int*                       strict
)
{
  return kaapi_foreach_attr_get_distribution(
        (kaapi_foreach_attr_t*)attr,
        (kaapi_hws_levelid_t*)level,
        strict );
}


/*
*/
int komp_foreach(
  unsigned long long         first,
  unsigned long long         last,
  unsigned long long         incr,
  int                        sync,
  const komp_foreach_attr_t* attr,
  komp_foreach_body_t body,
  void* arg
)
{
  if (last <= first)
    return 0;

  /* sync / or not */
  
  /* is_format true if called from kaapif_foreach_with_format */
  kaapi_processor_t* const kproc = kaapi_self_processor();
  if ((kproc ==0) || (kproc->team ==0))
  {
    body(first, last, arg);
    return 0;
  }
  komp_thread_t* const self_thread = (komp_thread_t*)kaapi_context2thread(kproc->context);

  KAAPI_EVENT_KPROC_PUSH0(kproc, 0, KAAPI_EVT_FOREACH_BEG );

  if (0== komp_loop_workinit_ull( self_thread, sync, attr, 1, first, last, 1, body, arg))
    return -1;

  while (komp_loop_worknext_ull(self_thread, &first, &last) != 0)
    body( first, last, arg);

  komp_loop_workfini( self_thread );
  KAAPI_EVENT_KPROC_PUSH0(kproc, 0, KAAPI_EVT_FOREACH_END );

  return 0;
}


/* loop init 
*/
int komp_loop_workinit(
  komp_thread_t* thread,
  int sync,
  const komp_foreach_attr_t* attr,
  long long start,
  long long end,
  long long incr,
  komp_foreach_body_t body,
  void* arg
)
{
  unsigned long long iter_size;
  kaapi_context_t* context = kaapi_thread2context(thread);

  if (incr >0)
    iter_size = (end - start + incr-1)/incr;
  else
    iter_size = (start - end - incr-1)/-incr;

  void* retval = kaapi_foreach_workinit( context, sync,
                    (const kaapi_foreach_attr_t*)attr,
                    iter_size,
                    body, arg
  );

  if (retval !=0) 
  {
    context->ws.for_data.first   = (unsigned long long)start;
    context->ws.for_data.last    = (unsigned long long)end;
    context->ws.for_data.incr    = (unsigned long long)incr;
    ++context->ws.count_for;
    return 1;
  }
  return 0;
}

/*
*/
int komp_loop_workinit_ull(
  komp_thread_t*           thread,
  int                        sync,
  const komp_foreach_attr_t* attr,
  int                        up,
  unsigned long long         start,
  unsigned long long         end,
  unsigned long long         incr,
  komp_foreach_body_t        body,
  void*                      arg
)
{
  unsigned long long iter_size;
  kaapi_context_t* context = kaapi_thread2context(thread);

  kaapi_assert_debug( !((up && (start > end)) || (!up && (start < end))) );

  if (up)
    iter_size = (end - start + incr-1)/incr;
  else
    iter_size = (start - end - incr-1)/-incr;

  void* retval = kaapi_foreach_workinit( context, sync,
                    (const kaapi_foreach_attr_t*)attr,
                    iter_size,
                    body, arg
  );
  if (retval !=0) 
  {
    context->ws.for_data.first   = start;
    context->ws.for_data.last    = end;
    context->ws.for_data.incr    = incr;
    ++context->ws.count_for;
    return 1;
  }
  return 0;
}


/* return !=0 value if success */
int komp_loop_worknext(
  komp_thread_t* thread,
  long long *istart,
  long long *iend
)
{
  unsigned long long ristart;
  unsigned long long riend;
  kaapi_context_t* context = kaapi_thread2context(thread);
  kaapi_foreach_work_t* lwork = context->ws.for_data.rt_ctxt;

  /* pop next range and start execution (on return...) */
  if ( (lwork !=0) &&
       (0  == kaapi_foreach_worknext(
          lwork,
          &ristart,
          &riend
          )
        )
      )
  {
    context->ws.for_data.curr_start = (context->ws.for_data.first + context->ws.for_data.incr * ristart);
    context->ws.for_data.curr_end   = (context->ws.for_data.first + context->ws.for_data.incr * riend);
    *istart = (long long)context->ws.for_data.curr_start;
    *iend   = (long long)context->ws.for_data.curr_end;
    return 1;
  }
  return 0;
}

/* return !=0 value if success */
int komp_loop_worknext_ull(
  komp_thread_t* thread,
  unsigned long long *istart,
  unsigned long long *iend
)
{
  kaapi_context_t* context = kaapi_thread2context(thread);
  kaapi_foreach_work_t* lwork = context->ws.for_data.rt_ctxt;

  /* pop next range and start execution (on return...) */
  if ( (lwork !=0) &&
       (0  == kaapi_foreach_worknext( lwork, istart, iend) )
     )
  {
    context->ws.for_data.curr_start = (context->ws.for_data.first + context->ws.for_data.incr * *istart);
    context->ws.for_data.curr_end   = (context->ws.for_data.first + context->ws.for_data.incr * *iend);
    *istart = context->ws.for_data.curr_start;
    *iend   = context->ws.for_data.curr_end;
    return 1;
  }
  return 0;
}


/* return !=0 value if success */
int komp_loop_workfini(
  komp_thread_t* thread
)
{
  /* pop next range and start execution (on return...) */
  kaapi_context_t* context = kaapi_thread2context(thread);
  kaapi_foreach_work_t* lwork = context->ws.for_data.rt_ctxt;
  if ( (lwork !=0) &&
       (0  == kaapi_foreach_workfini( lwork )) )
  {
    return 1;
  }
  return 0;
}


/*
*/
int komp_loop_policy_start (
  komp_thread_t* thread,
  const komp_foreach_attr_t* attr,
  long long start,
  long long end,
  long long incr,
  long long *istart,
  long long *iend
)
{
  if (0 == komp_loop_workinit( thread, 1, attr, start, end, incr, 0, 0))
    return 0;

  if ((istart !=0) && (iend !=0))
    return komp_loop_worknext( thread, istart, iend );
  return 1;
}


/*
*/
int komp_loop_policy_start_ull (
  komp_thread_t* thread,
  const komp_foreach_attr_t* attr,
  int up,
  unsigned long long start,
  unsigned long long end,
  unsigned long long incr,
  unsigned long long *istart,
  unsigned long long *iend
)
{
  if (0 == komp_loop_workinit_ull( thread, 1, attr, up, start, end, incr, 0, 0))
    return 0;

  if ((istart !=0) && (iend !=0))
    return komp_loop_worknext_ull( thread, istart, iend );
  return 1;
}


/*
*/
void komp_ordered_start (void)
{
  kaapi_processor_t* kproc = kaapi_self_processor();

  kaapi_context_t* thread = kproc->context;
  kaapi_team_t* team = kproc->team;

  if (team ==0)
    return;

  kaapi_workqueue_index_t ordered_it;
  kaapi_workqueue_index_t curr_ordered_index = KAAPI_ATOMIC_READ(&team->ws.ordered_index);
  
  /* my iteration bounds */
  kaapi_workqueue_index_t it_first = thread->ws.for_data.curr_start;
  kaapi_workqueue_index_t it_last  = thread->ws.for_data.curr_end;

do_test:
  ordered_it = thread->ws.for_data.first + thread->ws.for_data.incr * curr_ordered_index;

  if (thread->ws.for_data.incr >0)
  {
    if ((ordered_it >= it_first) && (ordered_it < it_last))
      goto return_value;
  }
  else
  {
    if ((ordered_it <= it_first) && (ordered_it > it_last))
      goto return_value;
  }

  /* wait until current_ordered_index change */
  while (curr_ordered_index == KAAPI_ATOMIC_READ(&team->ws.ordered_index))
    kaapi_slowdown_cpu();

  /* read it again */
  curr_ordered_index = KAAPI_ATOMIC_READ(&team->ws.ordered_index);
  goto do_test;
  
return_value:
  //printf("Out: Ordered: %llu\n", curr_ordered_index);
  return;
}


/*
*/
void komp_ordered_end (void)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
  kaapi_team_t* team = kproc->team;
  if (kproc ==0)
    return;

  /* assume atomic */
  kaapi_assert_debug(
      __kaapi_isaligned( &team->ws.ordered_index, sizeof(team->ws.ordered_index)
      )
  )
  //printf("Out: Ordered_end: %llu\n", KAAPI_ATOMIC_READ(&team->ws.ordered_index));
  kaapi_writemem_barrier();
  KAAPI_ATOMIC_INCR(&team->ws.ordered_index);
}

