/*
** @file quark_impl.h
**
** QUARK: Dynamic scheduler functions
**
** PLASMA is a software package provided by Univ. of Tennessee,
** Univ. of California Berkeley and Univ. of Colorado Denver
**
** xkaapi
** 
**
** Copyright 2011 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/

/* for debug only */
#include <string.h>         /* strcmp, for debug: strdup */

#include <stdarg.h>
#include <stdio.h>

#include "kaapi_impl.h"

#define STATIC 1  /* to force DFG computation */
//#define TRACE 1
//#define LOG_ACCESS 1


#include "quark_impl.h"
#include "quark_unpack_args.h"

#if defined(STATIC)
#include "kaapi_dbg.h"
#endif


static kaapi_atomic_t kaapi_quark_isinit = {0};


/* format for the kaapi' quark task */
static struct kaapi_format_t* kaapi_quark_task_format = 0;

/* dump dot representation */
static int quark_dump_dot = 0;

/* State fo Sequence */
#define QUARK_SEQUENCE_INIT          0x1
#define QUARK_SEQUENCE_FREE          0x2
#define QUARK_SEQUENCE_TASKLIST_INIT 0x4

XKaapi_Quark* default_Quark = 0;
static int num_threads_quark = 0;


static void kaapi_quark_task_format_constructor(void);

/*
*/
static char get_mode( kaapi_access_mode_t mode)
{
  switch (KAAPI_ACCESS_GET_MODE(mode)) {
  case KAAPI_ACCESS_MODE_V:
    return 'V';
  case KAAPI_ACCESS_MODE_R:
    return 'R';
  case KAAPI_ACCESS_MODE_W:
    return 'W';
  case KAAPI_ACCESS_MODE_RW:
    return 'X';
  case KAAPI_ACCESS_MODE_CW:
    return 'C';
  case KAAPI_ACCESS_MODE_SCRATCH:
    return 'S';
  default:
    return '?';
  }
}

/**
*/
Quark *QUARK_Setup(int num_threads)
{
  if (KAAPI_ATOMIC_INCR(&kaapi_quark_isinit) > 1)
  {
    /* already initialized */
    return &default_Quark[kaapi_self_kid()];
  }
  char* e = getenv("QUARK_DOT_DAG_ENABLE");
  if (e !=0)
  {
    if (strcmp(e, "0") ==0)
      quark_dump_dot = 0;
    else 
      quark_dump_dot = 1;
  }

  kaapi_init(KAAPI_START_ALL_THREADS, 0, 0);
  kaapi_quark_task_format_constructor();

  num_threads= kaapi_get_concurrency();
  default_Quark= (XKaapi_Quark*)malloc( num_threads * sizeof(XKaapi_Quark) );
  num_threads_quark = num_threads;
  memset( default_Quark, 0, num_threads*sizeof(default_Quark) );

  kaapi_processor_id_t selfkid = kaapi_self_kid();
  default_Quark[selfkid].thread   = kaapi_self_thread();
  default_Quark[selfkid].sequence = 0;
  default_Quark[selfkid].task     = 0;

  //PLASMA_SIZE = kaapi_get_concurrency();

  return &default_Quark[selfkid];
}


/* Setup scheduler data structures, spawn worker threads, start the workers working  */
Quark *QUARK_New(int num_threads)
{
#if defined(TRACE)
printf("IN %s\n", __PRETTY_FUNCTION__);  
  fflush(stdout);
#endif
  kaapi_processor_id_t selfkid = kaapi_self_kid();
  default_Quark[selfkid].thread = kaapi_self_thread();
#if defined(TRACE)
printf("OUT %s\n", __PRETTY_FUNCTION__);  
  fflush(stdout);
#endif
  return &default_Quark[selfkid];
}

/* trampoline to call Quark function */
static void kaapi_wrapper_quark_function( kaapi_task_t* ktask, kaapi_thread_t* thread )
{
  kaapi_quark_task_t* task  = kaapi_task_getargst(ktask,kaapi_quark_task_t);
  kaapi_processor_id_t selfkid = kaapi_self_kid();
  XKaapi_Quark* myquark    = &default_Quark[selfkid];
  myquark->thread          = kaapi_self_thread();
  myquark->task            = task;
  task->function( myquark );
  /* may be some thing here to release temporary */
}

/* Returns a pointer to the list of arguments, used when unpacking the
   arguments; Returna a pointer to icl_list_t, so icl_list.h will need
   bo included if you use this function */
void *QUARK_Args_List(Quark *quark)
{
  kaapi_processor_id_t selfkid = kaapi_self_kid();
  kaapi_assert_debug( quark == &default_Quark[selfkid] );
  return default_Quark[selfkid].task;
}


/* Return a pointer to an argument.  The variable last_arg should be
   NULL on the first call, then each subsequent call will used
   last_arg to get the the next argument. */
void *QUARK_Args_Pop( void *args_list, void **last_arg)
{
  kaapi_quark_task_t* taskarg = (kaapi_quark_task_t*)args_list;
  quark_one_param_t* arg2pop  = (quark_one_param_t*)*last_arg;
  void* retval;

  if (arg2pop ==0) 
    arg2pop = taskarg->param;

  if (arg2pop->addr.mode & KAAPI_ACCESS_MODE_T)
  {
    /* allocate a scratch zone: freeed after task execution */
    if (arg2pop->addr.data ==0)
    {
      arg2pop->addr.data = kaapi_gettemporary_data(
          (unsigned int) (arg2pop-taskarg->param),
          arg2pop->size
      );
    }
    retval = &arg2pop->addr.data;
  }
  else
  {
    if ((arg2pop->addr.mode & KAAPI_ACCESS_MODE_V) ==0)
      /* because, make memcpy on return (to copy a pointer !)*/
      retval = &arg2pop->addr.data;
    else
    {
      if (arg2pop->size <= sizeof(void*))
        retval = &arg2pop->addr.data;
      else
        retval = arg2pop->addr.data;
    }
  }
  *last_arg = arg2pop+1;
  return retval;
}


/*************************************************************/
/* Add a task, called by the master process (thread_rank 0)  */
unsigned long long QUARK_Insert_Task(
  XKaapi_Quark * quark, void (*function) (Quark *), Quark_Task_Flags *task_flags, ...
)
{
  va_list varg_list;
  int arg_size;
  kaapi_thread_t* thread = kaapi_self_thread();

  kaapi_quark_task_t* arg = kaapi_alloca( 
      thread, 
      sizeof(kaapi_quark_task_t) /* + nparam*sizeof(quark_one_param_t)*/ 
  );
  arg->scratchbit = 0;
  arg->nparam     = 0;
  arg->function   = function;

  /* For each argument */
  int nparam = 0;

#if defined(LOG_ACCESS)
{
  printf("Quark Task: %p\n", arg );
  if (task_flags->task_priority)
    printf("Priority info on task: %i\n", task_flags->task_priority );
}
#endif

  va_start(varg_list, task_flags);
  while( (arg_size = va_arg(varg_list, int)) != 0) 
  {
      void *arg_ptr  = va_arg(varg_list, void*);
      int arg_flags  = va_arg(varg_list, int);

      kaapi_access_init(&arg->param[nparam].addr, 0);
      arg->param[nparam].size         = arg_size;
      quark_direction_t arg_direction = (quark_direction_t) (arg_flags & QUARK_DIRECTION_BITMASK);
      switch ( arg_direction ) 
      {
        case VALUE:
          arg->param[nparam].addr.mode      = KAAPI_ACCESS_MODE_V;
          if (arg_size <= sizeof(void*)) /* copy the value in the pointer */
            arg->param[nparam].addr.data = *(void**)arg_ptr;
          else
          {
            arg->param[nparam].addr.data = kaapi_alloca(thread, arg_size);
            memcpy(arg->param[nparam].addr.data, arg_ptr, arg_size);
          }
          
        break;

        case NODEP: /* but keep pointer */
          arg->param[nparam].addr.data = arg_ptr;
          arg->param[nparam].addr.mode = KAAPI_ACCESS_MODE_V;
        break;

        case INPUT:
          arg->param[nparam].addr.data = arg_ptr;
          arg->param[nparam].addr.mode = KAAPI_ACCESS_MODE_R;
        break;
        
        case OUTPUT:
          arg->param[nparam].addr.data = arg_ptr;
          if ((arg_flags & GATHERV) || (arg_flags & ACCUMULATOR))
          {
            arg->param[nparam].addr.mode = KAAPI_ACCESS_MODE_ICW;
            printf("ACCUMULATOR: Not yet implemented\n");
            abort();
          }
          else
            arg->param[nparam].addr.mode = KAAPI_ACCESS_MODE_W;
        break;

        case INOUT:
          arg->param[nparam].addr.data = arg_ptr;
          if ((arg_flags & GATHERV) || (arg_flags & ACCUMULATOR))
            //arg->param[nparam].addr.mode = KAAPI_ACCESS_MODE_ICW;
            arg->param[nparam].addr.mode = KAAPI_ACCESS_MODE_RW;
          else
            arg->param[nparam].addr.mode = KAAPI_ACCESS_MODE_RW;
        break;

        case SCRATCH:
          arg->param[nparam].addr.data = arg_ptr;
          arg->param[nparam].addr.mode = KAAPI_ACCESS_MODE_SCRATCH; /* new extension for Kaapi */
          arg->scratchbit |= (1 << nparam);
        break;

        default:
          printf("Unknown access mode: %i\n", (int)arg_direction );
          abort();
      }
#if defined(LOG_ACCESS)
{
  printf(",  @: %p size:%i, mode:%c", (void*)arg_ptr, arg_size, get_mode(arg->param[nparam].addr.mode));
  if (arg_flags & LOCALITY) 
  {
    printf(" | ");
    printf("LOCALITY");
  }
  if (arg_flags & ACCUMULATOR) 
  {
    printf(" | ");
    printf("ACCUM");
  }
  if (arg_flags & GATHERV) 
  {
    printf(" | ");
    printf("GATHERV");
  }
  if (arg_flags & QUARK_REGION_0) 
  {
    printf(" | ");
    printf("REGION_0");
  }
  if (arg_flags & QUARK_REGION_1) 
  {
    printf(" | ");
    printf("REGION_1");
  }
  if (arg_flags & QUARK_REGION_2) 
  {
    printf(" | ");
    printf("REGION_2");
  }
  if (arg_flags & QUARK_REGION_3) 
  {
    printf(" | ");
    printf("REGION_3");
  }
  if (arg_flags & QUARK_REGION_4) 
  {
    printf(" | ");
    printf("REGION_4");
  }
  if (arg_flags & QUARK_REGION_5) 
  {
    printf(" | ");
    printf("REGION_5");
  }
  if (arg_flags & QUARK_REGION_6) 
  {
    printf(" | ");
    printf("REGION_6");
  }
  if (arg_flags & QUARK_REGION_7) 
  {
    printf(" | ");
    printf("REGION_7");
  }
}
#endif
      ++nparam;
      kaapi_assert( nparam <= MAX_PARAMQUARK );
  }
  va_end(varg_list);


  arg->nparam        = nparam;

  int prio = 0;
  if (task_flags->task_priority)
    prio = task_flags->task_priority;

#if defined(LOG_ACCESS)
{
  printf("<<\n");
  printf("KAAPI Task: %p <=> QUARK Task %p,   #param:%i\n", thread->sp, arg, nparam );
  fflush(stdout);
}
#endif

  /* No function: use default QUARK format */
  kaapi_assert( arg != 0 );
  kaapi_task_push_withprio(thread, kaapi_wrapper_quark_function, (void*)arg, prio);

  return 0;
}


/* Main work loop, called externally by everyone but the master
 * (master manages this internally to the insert_task and waitall
 * routines). Each worker thread can call work_main_loop( quark,
 * thread_rank), where thread rank is 1...NUMTHREADS ) */
void QUARK_Worker_Loop(Quark *quark, int thread_rank)
{
  kaapi_abort(__LINE__, __FILE__, "[XKaapi/Quark: not implemented]");
}

/* Finish work and return.  Workers do not exit */
void QUARK_Barrier(Quark * quark)
{
}

/* Just wait for current tasks to complete, the scheduler and
 * strutures remain as is... should allow for repeated use of the
 * scheduler.  The workers return from their loops.*/
void QUARK_Waitall(Quark * quark)
{
  if (quark->sequence !=0) {
    QUARK_Sequence_Wait(quark, quark->sequence);
#if 0
    kaapi_memory_synchronize();
#endif
  }
  else {
    kaapi_sched_sync( kaapi_self_thread() );
  }
}

/* Delete scheduler, shutdown threads, finish everything, free structures */
void QUARK_Delete(Quark * quark)
{
  /* TODO not used */
#if 0
  kaapi_memory_synchronize();
#endif
  kaapi_finalize();
}

/* Free scheduling data structures */
void QUARK_Free(Quark * quark)
{
#if 0
  kaapi_memory_synchronize();
#endif
  if (KAAPI_ATOMIC_DECR(&kaapi_quark_isinit)  == 0)
  {
    kaapi_finalize();
  }
}

/* Cancel a specific task */
int QUARK_Cancel_Task(Quark *quark, unsigned long long taskid)
{
#if (SIZEOF_VOIDP==4)
  kaapi_task_t* task = (kaapi_task_t*)(uintptr_t)taskid;
#else
  kaapi_task_t* task = (kaapi_task_t*)taskid;
#endif
  kaapi_task_body_t body = 0;
  KAAPI_ATOMIC_EXCHANGE(body, task->state.steal );

  return 0;
}


/* Returns the rank of a thread in a parallel task */
int QUARK_Get_RankInTask(Quark *quark)
{ return kaapi_self_kid(); }


/* to debug quark */
extern void *kaapi_memcpy(void *dest, const void *src, size_t n)
{
  return memcpy( dest, src, n );
}

/* Utility function returning rank of the current thread */
int QUARK_Thread_Rank(Quark *quark)
{ return kaapi_self_kid(); }

/* Packed task interface */
/* Create a task data structure to hold arguments */
Quark_Task *QUARK_Task_Init(Quark * quark, void (*function) (Quark *), Quark_Task_Flags *task_flags )
{
  kaapi_assert(0);
  return 0;
}

/* Add (or pack) the arguments into a task data structure (make sure of the correct order) */
void QUARK_Task_Pack_Arg( Quark *quark, Quark_Task *task, int arg_size, void *arg_ptr, int arg_flags )
{
  kaapi_assert(0);
  return;
}

/* Insert the packed task data strucure into the scheduler for execution */
unsigned long long QUARK_Insert_Task_Packed(Quark * quark, Quark_Task *task )
{
  kaapi_assert(0);
  return 0;
}

/* Unsupported function for debugging purposes; execute task AT ONCE */
unsigned long long QUARK_Execute_Task(Quark * quark, void (*function) (Quark *), Quark_Task_Flags *task_flags, ...)
{
  kaapi_assert(0);
  return 0;
}

/* Unsupported function for debugging purposes; execute task AT ONCE */
unsigned long long QUARK_Execute_Task_Packed( Quark *quark, Quark_Task *task )
{
  kaapi_assert(0);
  return 0;
}

/* Get the label (if any) associated with the current task; used for printing and debugging  */
char *QUARK_Get_Task_Label(Quark *quark)
{
  kaapi_assert(0);
  return 0;
}


/* Method for setting task flags */
Quark_Task_Flags *QUARK_Task_Flag_Set( Quark_Task_Flags *task_flags, int flag, intptr_t val )
{
  switch (flag)
  {
    case TASK_PRIORITY:
        task_flags->task_priority = (int)val;
        break;
    case TASK_LOCK_TO_THREAD:
        task_flags->task_lock_to_thread = (int)val;
        break;
    case TASK_LABEL:
        task_flags->task_label = (char *)val;
        break;
    case TASK_COLOR:
        task_flags->task_color = (char *)val;
        break;
    case TASK_SEQUENCE:
        task_flags->task_sequence = (Quark_Sequence *)val;
        break;
    case TASK_THREAD_COUNT:
        task_flags->task_thread_count = (int)val;
        break;
    case THREAD_SET_TO_MANUAL_SCHEDULING:
        task_flags->thread_set_to_manual_scheduling = (int)val;
        break;
  }
  return task_flags;
}

static Quark_Sequence* last_qs_free = 0;

/* Create a seqeuence structure, to hold sequences of tasks */
Quark_Sequence *QUARK_Sequence_Create( Quark *quark )
{
  /* activate static schedule */
  kaapi_quark_sequence_t* qs;
  if (last_qs_free !=0)
  {
    qs = last_qs_free;
    last_qs_free = 0;
    qs->state_init &= ~QUARK_SEQUENCE_FREE;
  } 
  else 
  {
    qs = (kaapi_quark_sequence_t*)malloc( sizeof(kaapi_quark_sequence_t) );
    qs->state_init = 0;
  }

  kaapi_context_t* context = kaapi_self_context();

#if defined(STATIC)
  kaapi_frame_t* frame = context->stack.unlink;
  frame->flag |= KAAPI_FRAME_FLAG_NOTSTEAL;
#endif

  /* push new frame for task */
  //kaapi_thread_save( &context->stack.thread, &qs->save_thread );
  qs->save_frame = kaapi_stack_push_frame( &context->stack);

  quark->sequence = qs;
  qs->state_init |= QUARK_SEQUENCE_INIT;
  return qs;
}

/* Called by worker, cancel any pending tasks, and mark sequence so that it does not accept any more tasks */
int QUARK_Sequence_Cancel( Quark *quark, Quark_Sequence *sequence )
{
  if ( quark==NULL || sequence==NULL ) 
    return QUARK_ERR;
  return QUARK_SUCCESS;

  kaapi_assert(0);
  return 0;
}

/* Destroy a sequence structure, cancelling any pending tasks */
Quark_Sequence *QUARK_Sequence_Destroy( Quark *quark, Quark_Sequence *sequence )
{
//printf("%s\n", __PRETTY_FUNCTION__);  
//  QUARK_Sequence_Wait( quark, sequence );
#if defined(TRACE)
printf("IN %s\n", __PRETTY_FUNCTION__);  fflush(stdout);
#endif

  if (last_qs_free ==0)
  {
    last_qs_free = sequence;
  }
  else {
    free(sequence);
  }
  quark->sequence = 0;

#if defined(TRACE)
printf("OUT %s\n", __PRETTY_FUNCTION__);  fflush(stdout);
#endif
  return 0;
}

/* Wait for a sequence of tasks to complete */
int QUARK_Sequence_Wait( Quark *quark, Quark_Sequence *sequence )
{
#if defined(TRACE)
printf("IN %s\n", __PRETTY_FUNCTION__);  fflush(stdout);
#endif
//  kaapi_assert_debug( quark->thread == kaapi_self_context() );

  kaapi_context_t* context = kaapi_self_context();

#if defined(STATIC)
  kaapi_frame_t* frame = context->stack.unlink;
  if (quark_dump_dot)
  {
    static uint32_t counter = 0;
    char filename[128]; 
    if (getenv("USER") !=0)
      sprintf(filename,"graph.stack.%s.%i.dot", getenv("USER"), counter++ );
    else
      sprintf(filename,"graph.%i.dot",counter++);
    FILE* filedot = fopen(filename, "w");
    kaapi_frame_print_dot( filedot,  &context->stack, frame );
    fclose(filedot);
  }

  if ((sequence->state_init & QUARK_SEQUENCE_TASKLIST_INIT) == 0)
  {
    //??
  }

  /* currently: that all, do not compute other things */
  kaapi_frame_computereadylist(&context->stack,frame);
#if 0
  printf(">>>>>>>>>>---\n\n");
  kaapi_frame_tasklist_print( stdout, frame->rdlist );
  printf("-------------\n\n");
#endif
  
  /* force serialisation of previous write with the next write to stealable flag */
  kaapi_writemem_barrier();

  /* reset stealable flag on frame */
  KAAPI_ATOMIC_ANDPTR(&frame->flag, ~KAAPI_FRAME_FLAG_NOTSTEAL);
//  frame->flag &= ~KAAPI_FRAME_FLAG_NOTSTEAL;
#endif

  /* real execution of tasks: here */
  kaapi_sched_sync(&context->stack.thread);

  /* restore thread */
  //kaapi_atomic_lock( &context->proc->lock );
  //kaapi_thread_restore( kaapi_self_thread(), &quark->sequence->save_thread );
  //kaapi_atomic_unlock( &context->proc->lock );
  kaapi_stack_pop_frame(&context->stack, quark->sequence->save_frame);

  sequence->state_init &= ~QUARK_SEQUENCE_INIT;
  sequence->state_init |= QUARK_SEQUENCE_FREE;

  return 1;
}

/* Get the sequence information associated the current task/worker, this was provided when the tasks was created */
Quark_Sequence *QUARK_Get_Sequence(Quark *quark)
{
printf("%s\n", __PRETTY_FUNCTION__);  
  fflush(stdout);
  return quark->sequence;
}

/* Get the priority associated the current task/worker */
int QUARK_Get_Priority(Quark *quark)
{
  return 0;
}

/* Get information associated the current task and worker thread;
 * Callable from within a task, since it works on the currently
 * executing task */
intptr_t QUARK_Task_Flag_Get( Quark *quark, int flag )
{
  return 0;
}

/* Enable and disable DAG generation via API.  Only makes sense after
 * a sync, such as QUARK_Barrier. */
void QUARK_DOT_DAG_Enable( Quark *quark, int boolean_value )
{
  return;
}

/* format definition for any QUARK task */
static size_t
kaapi_quark_task_format_get_size(const struct kaapi_format_t* fmt, const void* sp)
{
  return sizeof(kaapi_quark_task_t);
}

static void
kaapi_quark_task_format_task_copy(const struct kaapi_format_t* fmt, void* sp_dest, const void* sp_src)
{
  memcpy( sp_dest, sp_src, sizeof(kaapi_quark_task_t) );
}

static
unsigned int kaapi_quark_task_format_get_count_params(const struct kaapi_format_t* fmt, const void* sp)
{
  kaapi_quark_task_t* arg = (kaapi_quark_task_t*)sp;
  return arg->nparam;
}

static
kaapi_access_mode_t kaapi_quark_task_format_get_mode_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  kaapi_quark_task_t* arg = (kaapi_quark_task_t*)sp;
  return arg->param[i].addr.mode;
}


static
void* kaapi_quark_task_format_get_data_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  kaapi_quark_task_t* arg = (kaapi_quark_task_t*)sp;
  void* retval = 0;
  kaapi_assert_debug ((arg->param[i].addr.mode & KAAPI_ACCESS_MODE_V));
  retval = arg->param[i].addr.data;
  return retval;
}

static
kaapi_access_t* kaapi_quark_task_format_get_access_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  kaapi_quark_task_t* arg = (kaapi_quark_task_t*)sp;
  kaapi_access_t* retval = 0;
  kaapi_assert_debug ((arg->param[i].addr.mode & KAAPI_ACCESS_MODE_V) == 0);
  retval = &arg->param[i].addr;
  return retval;
}

static
void kaapi_quark_task_format_set_access_param(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const kaapi_access_t* a
)
{
  kaapi_quark_task_t* arg = (kaapi_quark_task_t*)sp;
  kaapi_assert_debug ((arg->param[i].addr.mode & KAAPI_ACCESS_MODE_V) == 0);
  arg->param[i].addr = *a;
}

static
const struct kaapi_format_t* kaapi_quark_task_format_get_fmt_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp
)
{
  kaapi_quark_task_t* arg __attribute__((unused)) = (kaapi_quark_task_t*)sp;
  return kaapi_char_format;
}

static void kaapi_quark_task_format_get_view_param(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp,
  kaapi_memory_view_t* view
)
{
  kaapi_quark_task_t* arg = (kaapi_quark_task_t*)sp;
  kaapi_memory_view_make1d( view, 0, arg->param[i].size, 1);
}

static
void kaapi_quark_task_format_set_view_param(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const kaapi_memory_view_t* view
)
{
  kaapi_quark_task_t* arg __attribute__((unused))= (kaapi_quark_task_t*)sp;
  kaapi_abort(__LINE__, __FILE__, "invalid case");
}

static
void kaapi_quark_task_format_reducor(
  const struct kaapi_format_t* fmt, unsigned int i, void* sp, const void* v
)
{
  kaapi_quark_task_t* arg __attribute__((unused))= (kaapi_quark_task_t*)sp;
}

static
void kaapi_quark_task_format_redinit(
  const struct kaapi_format_t* fmt, unsigned int i, const void* sp, void* v
)
{
  kaapi_quark_task_t* arg __attribute__((unused))= (kaapi_quark_task_t*)sp;
}


static void kaapi_quark_task_format_constructor(void)
{
  if (kaapi_quark_task_format !=0) return;
  kaapi_quark_task_format = kaapi_format_allocate();
  kaapi_format_taskregister_func(
                   kaapi_quark_task_format,
                   (kaapi_task_body_t)kaapi_wrapper_quark_function,
                   "kaapi_quark_task_format",
                   0, /* get name */
                   kaapi_quark_task_format_get_size,
                   kaapi_quark_task_format_task_copy,
                   kaapi_quark_task_format_get_count_params,
                   kaapi_quark_task_format_get_mode_param,
                   kaapi_quark_task_format_get_data_param,
                   kaapi_quark_task_format_get_access_param,
                   kaapi_quark_task_format_set_access_param,
                   kaapi_quark_task_format_get_fmt_param,
                   kaapi_quark_task_format_get_view_param,
                   kaapi_quark_task_format_set_view_param,
                   kaapi_quark_task_format_reducor,
                   kaapi_quark_task_format_redinit,
                   0 /*splitter */
                  );
}


#if defined(KAAPI_USE_KBLAS)

void kaapi_quark_task_format_register(
  const char* name, kaapi_quark_task_body_t plasma,
  kaapi_quark_task_body_t wrapper,
  kaapi_quark_task_body_t wrapper_wh,
  kaapi_quark_task_body_t body,
  kaapi_quark_task_body_t wrapper_cuda,
  kaapi_quark_task_body_t wrapper_cuda_wh,
  kaapi_quark_task_body_t body_cuda
)
{
  kaapi_format_t* fmt = kaapi_format_allocate();
  
#if 0
  fprintf(stdout, "%d:%s: name=%s plasma=%p wrapper=%p wrapper_wh=%p body=%p\n", kaapi_self_kid(), __FUNCTION__,
          name, (void*)plasma, (void*)wrapper, (void*)wrapper_wh, (void*)body );
  fflush(stdout);
#endif
 
  kaapi_format_taskregister_func(
                                 fmt,
                                 (kaapi_task_body_t)wrapper,
                                 (kaapi_task_body_t)wrapper_wh,
                                 name,
                                 0, /* get name */
                                 kaapi_quark_task_format_get_size,
                                 kaapi_quark_task_format_task_copy,
                                 kaapi_quark_task_format_get_count_params,
                                 kaapi_quark_task_format_get_mode_param,
                                 kaapi_quark_task_format_get_access_param,
                                 kaapi_quark_task_format_set_access_param,
                                 kaapi_quark_task_format_get_fmt_param,
                                 kaapi_quark_task_format_get_view_param,
                                 kaapi_quark_task_format_set_view_param,
                                 kaapi_quark_task_format_reducor,
                                 kaapi_quark_task_format_redinit,
                                 0 /* splitter */
                                 );
  
  kaapi_quark_task_body_register(plasma, wrapper, body, KAAPI_PROC_TYPE_CPU);
  
#if defined(KAAPI_USE_CUDA)
  if(wrapper_cuda == 0 || wrapper_cuda_wh == 0 || body_cuda == 0)
    return;
  
  kaapi_format_taskregister_body
  (
   fmt,
   (kaapi_task_body_t)wrapper_cuda,
   (kaapi_task_body_t)wrapper_cuda_wh,
   KAAPI_PROC_TYPE_CUDA
   );
  
  /* No need to re-writer the function wrapper */
  kaapi_quark_task_body_register(plasma, 0, body_cuda, KAAPI_PROC_TYPE_CUDA);
#endif /* KAAPI_USE_CUDA */
}

#endif /* KAAPI_USE_KBLAS */
