/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** francois.broquedis@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include <errno.h>

#include "kaapi_impl.h"
#include "kaapi_util.h"


/*
 */
static inline int is_range_delim(const int c)
{
  return (c == ':') || (c == '-');
}

/*
 */
static inline int is_delim_stride(const int c)
{
  return (c == ':');
}

/* end of list
 */
static inline int is_eol(const int c)
{
  /* end of list */
  return (c == ',') || (c == 0);
}

/*
*/
static void eat_space(char** str)
{
  while ((*str !=0) && isspace ((int)**str))
    ++*str;
}



/* Parse the interger with unit that represents a size.
   Return true if one was successfully parsed.
*/
bool kaapi_parse_size (
    char **str,
    unsigned long *pvalue
)
{
  char *end;
  unsigned long value, shift = 10;
  
  if (*str == 0)
    return false;
  
  eat_space(str);
  if (*str == '\0')
    goto invalid;

  errno = 0;
  value = strtoul (*str, &end, 10);
  if (errno || value == ULONG_MAX)
    goto invalid;

  *str = end;
  eat_space(str);

  if (isalpha((int)**str))
  {
    switch (tolower ((int) **str))
    {
      case 'b':
        shift = 0;
        break;
      case 'k':
        shift = 10;
        break;
      case 'm':
        shift = 20;
        break;
      case 'g':
        shift = 30;
        break;
      default:
        goto invalid;
    }
    ++*str;
    eat_space(str);
  }
  
  if (((value << shift) >> shift) != value)
    goto invalid;
  
  *pvalue = value << shift;
  return true;
  
invalid:
  return false;
}


/* Parse an unsigned long environment varible.  Return true if one was
   present and it was successfully parsed.  */
bool
kaapi_parse_unsigned_long(
    char** str,
    unsigned long *pvalue
)
{
  char *end;
  unsigned long value;

  if (*str == 0)
    return false;

  eat_space(str);
  if (**str == '\0')
    goto invalid;

  errno = 0;
  value = strtoul (*str, &end, 10);
  if (errno || value == ULONG_MAX)
    goto invalid;

  *pvalue = (unsigned long)value;
  *str = end;
  return true;

 invalid:
  return false;
}


/* Parse an unsigned int environment varible.  Return true if one was
   present and it was successfully parsed.  */
bool
kaapi_parse_unsigned_int(
    char** str,
    unsigned int *pvalue
)
{
  unsigned long value = 0;
  if (!kaapi_parse_unsigned_long(str, &value))
    return false;
  if (value > UINT_MAX)
     return false;
  *pvalue = (unsigned int)value;
  return true;
}


/* Parse an unsigned long environment varible.  Return true if one was
   present and it was successfully parsed.  */
bool
kaapi_parse_int (char** str, int *pvalue)
{
  unsigned long value = 0;
  if (!kaapi_parse_unsigned_long(str, &value))
    return false;
  if ((value > UINT_MAX) || ((long)value < INT_MIN))
     return false;
  *pvalue = (int)value;
  return true;
}


/* Parse an unsigned long environment varible.  Return true if one was
   present and it was successfully parsed.  */
bool kaapi_parse_list_unsigned_int (char** str, unsigned int* count, unsigned int **pvalue)
{
  unsigned int value;
  unsigned int* list = 0;
  int size = 0;

  if (*str == NULL)
    return false;
  *count = 0;

redo:
  if (!kaapi_parse_unsigned_int(str, &value))
    goto invalid;

  if (*count+1 > size)
  {
    size = 2*size;
    if (size < *count+2) size = (*count+2)*2;
    list = (unsigned int*)realloc( list, size*sizeof(unsigned int));
  }
  list[*count] = (unsigned int)value;
  ++*count;

  eat_space(str);
  if (!is_eol((int)**str))
    goto invalid;

  if (**str == ',')
  {
    ++*str;
    goto redo;
  }

  *pvalue = list;
  return true;

invalid:
  if (list !=0) free(list);
  return false;
}


/* parse PROC_BIND value: false,true,master,close,spread and
   return the numerical value komp_proc_bind_t
*/
bool kaapi_parse_procbind (char** str, kaapi_procbind_t* procbind )
{
  if (**str == 0)
    return false;

  eat_space(str);

  if (strncasecmp(*str,"false", 5) ==0)
  {
    *procbind = KAAPI_PROCBIND_FALSE;
    *str += 5;
  }
  else if (strncasecmp(*str,"true", 4) ==0)
  {
    *procbind = KAAPI_PROCBIND_TRUE;
    *str += 4;
  }
  else if (strncasecmp(*str,"master", 6) ==0)
  {
    *procbind = KAAPI_PROCBIND_MASTER;
    *str += 6;
  }
  else if (strncasecmp(*str,"close", 5) ==0)
  {
    *procbind = KAAPI_PROCBIND_CLOSE;
    *str += 5;
  }
  else if (strncasecmp(*str,"spread", 6) ==0)
  {
    *procbind = KAAPI_PROCBIND_SPREAD;
    *str += 6;
  }
  else
    return false;
  return true;
}

/* Parse a list of kaapi_procbind_t.  Return true if one was
   present and it was successfully parsed.  
*/
bool
kaapi_parse_list_procbind (char** str, unsigned int* count, kaapi_procbind_t **pvalue)
{
  char *token, *context;
  const char *sep = ",";
  kaapi_procbind_t value;
  kaapi_procbind_t* list = 0;
  int size = 0;

  if (**str == 0)
    return false;

  eat_space(str);

  for (token = strtok_r(*str, sep, &context);
       token;
       token = strtok_r(NULL, sep, &context))
  {
    *str = token;
    if (!kaapi_parse_procbind(str, &value))
      goto invalid;

    if (*count+2 > size)
    {
      size = 2*size;
      if (size < *count+2) size = (*count+2)*2;
      list = (kaapi_procbind_t*)realloc( list, size*sizeof(kaapi_procbind_t));
    }
    list[*count] = value;
    ++*count;
  }

  return true;

invalid:
  if (list !=0) free(list);
  return false;
}


/* Parse the Bool environment variable. 
*/
bool
kaapi_parse_bool (char** str, int* pvalue)
{
  if (*str == 0)
    return false;

  eat_space(str);

  if (strncasecmp (*str, "true", 4) == 0)
  {
    *pvalue = true;
    *str += 4;
  }
  else if (strncasecmp (*str, "false", 5) == 0)
  {
    *pvalue = false;
    *str += 5;
  }
  else
    return false;
  return true;
}


/* Parse the full|resume|gplot, true=active */
bool
kaapi_parse_perfcounter (char** str, kaapi_display_perf_value_t* pvalue)
{
  if (*str == 0)
    return false;

  eat_space(str);

  if (strncasecmp (*str, "no", 2) == 0)
  {
    *pvalue = KAAPI_NO_DISPLAY_PERF;
    *str += 2;
  }
  else if (strncasecmp (*str, "0", 1) == 0)
  {
    *pvalue = KAAPI_NO_DISPLAY_PERF;
    *str += 1;
  }
  else if (strncasecmp (*str, "full", 4) == 0)
  {
    *pvalue = KAAPI_DISPLAY_PERF_FULL;
    *str += 4;
  }
  else if (strncasecmp (*str, "1", 1) == 0)
  {
    *pvalue = KAAPI_DISPLAY_PERF_FULL;
    *str += 1;
  }
  else if (strncasecmp (*str, "resume", 6) == 0)
  {
    *pvalue = KAAPI_DISPLAY_PERF_RESUME;
    *str += 6;
  }
  else if (strncasecmp (*str, "gplot", 5) == 0)
  {
    *pvalue = KAAPI_DISPLAY_PERF_GPLOT;
    *str += 5;
  }
  else
    return false;
  return true;
}


/* Parse the ccsync|noaggr.
   Initialize the steal protocol in the runtime parameter rt_param + initialization and 
   dstor function.
*/
bool
kaapi_parse_stealprotocol (char** str, kaapi_rtparam_t* rt_param)
{
  if (*str == NULL)
    return false;

  eat_space(str);

#if defined(KAAPI_USE_FLAT_CC_SYNC)
  if (strncasecmp (*str, "ccsync", 6) == 0)
  {
    rt_param->emitsteal          = kaapi_sched_ccsync_emitsteal;
    rt_param->emitsteal_initctxt = kaapi_sched_ccsync_emitsteal_init;
    rt_param->emitsteal_dstorctxt= kaapi_sched_ccsync_emitsteal_dstor;
    *str += 6;
    return true;
  }
#endif
  if (strncasecmp (*str, "noaggr", 6) == 0)
  {
    rt_param->emitsteal          = kaapi_sched_noaggr_emitsteal;
    rt_param->emitsteal_initctxt = kaapi_sched_noaggr_emitsteal_init;
    rt_param->emitsteal_dstorctxt= kaapi_sched_noaggr_emitsteal_dstor;
    *str += 6;
    return true;
  }
  return false;
}


/* Parse the DIPLAY_ENV environment variable.
   This feature comes from OpenMP 4 RC2.

   As described in section 4.12, OMP_DISPLAY_ENV can be set to one of
   these values: 
   * true    -> *pvalue = 1;
   * false   -> *pvalue = 0;
   * verbose -> *pvalue = 2;
 */
bool
kaapi_parse_display_env (char** str, int* pvalue)
{
  if (*str == NULL)
    return false;

  eat_space(str);

  if (strncasecmp (*str, "true", 4) == 0)
  {
    *pvalue = 1;
    *str += 4;
  }
  else if (strncasecmp (*str, "verbose", 7) == 0)
  {
    *pvalue = 2;
    *str += 7;
  }
  else if (strncasecmp (*str, "false", 5) != 0)
  {
    *pvalue = 0;
    *str += 5;
  }
  else
    return false;
  return true;
}


bool
kaapi_parse_schedule (char** str, kaapi_foreach_attr_policy_t* prun_sched, unsigned long* prun_sched_modifier)
{
  int size;
  unsigned long run_sched_modifier = 0;
  if (*str == NULL)
    return false;

  eat_space(str);
  char* policy = strsep(str, ",");
  if (*str == 0)
  { /* no sep */
    run_sched_modifier = 0;
    *str = policy;
  }
  else if (!kaapi_parse_unsigned_long( str, &run_sched_modifier ))
    return false;

  if (strncasecmp (policy, "static",6) == 0)
  { *prun_sched = KAAPI_FOREACH_SCHED_STATIC; size = 6; }
  else if (strncasecmp (policy, "dynamic", 7) == 0)
  { *prun_sched = KAAPI_FOREACH_SCHED_DYNAMIC; size = 7; }
  else if (strncasecmp (policy, "guided",6) == 0)
  { *prun_sched = KAAPI_FOREACH_SCHED_GUIDED; size = 6; }
  else if (strncasecmp (policy, "adaptive",8) == 0)
  { *prun_sched = KAAPI_FOREACH_SCHED_ADAPTIVE; size = 8; }
  else if (strncasecmp (policy, "steal",5) == 0)
  { *prun_sched = KAAPI_FOREACH_SCHED_STEAL; size = 5; }
  else if (strncasecmp (policy, "auto",4) == 0)
  { *prun_sched = KAAPI_FOREACH_SCHED_AUTO; size = 4; }
  else
    return false;
  *prun_sched_modifier = run_sched_modifier;
  if (policy == *str)
    *str += size;
  return true;
}



/** Parse :
    [!] [ low ] [ -|: [ high ] [: stride] ]

    low -> num | <empty>
    high -> num | <empty>
    if empty detected for low, high then it set value to (unsigned int)-1.
    If stride pointer is not null and no stride is specified, then stride is set to 1.
    If stride pointer is null and a stride is specified, then it is an error
    If low '-' high then, low <= high.
    Else if format is low ':' high, then this is the OpenMP format with low:count.
 */
bool kaapi_parse_range(
    char** str,
    unsigned int* index_low,
    unsigned int* index_high,
    int* stride
)
{
  char range_delim = ' ';
  if (isdigit(**str))
  {
    if (!kaapi_parse_unsigned_int(str, index_low))
      return false;
  }
  else
  {
    if (!is_range_delim(**str)) /* - or : */
      return false;
    *index_low = (unsigned int)-1;
  }

  *index_high = *index_low;
  if (stride !=0)
    *stride = 1;

  if (is_range_delim(**str)) /* - or : */
  {
    range_delim = **str;
    ++*str;

    /* 1 token look ahead */
    if (is_eol(**str))
      *index_high = (unsigned int)-1;
    else
    {
      if (!kaapi_parse_unsigned_int(str, index_high))
        return false;

      if ((stride ==0) && is_delim_stride(**str))
        return false;

      if (is_delim_stride(**str))
      {
        ++*str;
        if (is_eol(**str))
          *stride = 1;
        if (!kaapi_parse_int(str, stride))
          return false;
      }
    }
  }

  /* test if empty set (index_low-index_high:stride)
     or *stride ==0.
  */
  if (range_delim == ':')  
  {
    if (*index_high == -1) *index_high = 1;
    if (*index_high < 1) return false;
    *index_high = *index_low + (*index_high -1) * (stride == 0 ? 1 : *stride);
  }

  if ((*index_low > *index_high) || ((stride !=0) && (*stride ==0)))
    return false;

  return true;
}


/* range or list of number
*/
bool kaapi_parse_oneplace(
    char** str,
    unsigned int* cpu_count,
    kaapi_place_t* place
)
{
  if (*str ==0)
    return false;

  *cpu_count = 0;

  /* parse list */
  int negate;
  unsigned int low;
  unsigned int high;
  int stride;

#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
  kaapi_place_t set;
  kaapi_place_t negateset;
#endif

#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
  CPU_ZERO( place );
  CPU_ZERO( &set );
  CPU_ZERO( &negateset );
#endif
  while (1)
  {
    eat_space(str);
    if (**str == '!')
    {
      negate = 1;
      ++*str;
    }
    else negate = 0;

    /* numa(i), socket(id), core(id) thread(id) extensions to OMP_PLACES
    */
    if (isalpha(**str))
    {
      kaapi_hws_levelid_t hid = KAAPI_HWS_LEVELID_MAX;
      unsigned id;

      if (strncasecmp (*str, "thread", 6) == 0)
      {
        *str += 6;
        hid = KAAPI_HWS_LEVELID_THREAD;
      }
      else if (strncasecmp (*str, "core", 4) == 0)
      {
        *str += 4;
        hid = KAAPI_HWS_LEVELID_CORE;
      }
      else if (strncasecmp (*str, "socket", 6) == 0)
      {
        *str += 6;
        hid = KAAPI_HWS_LEVELID_SOCKET;
      }
      else if (strncasecmp (*str, "numa", 4) == 0)
      {
        *str += 4;
        hid = KAAPI_HWS_LEVELID_NUMA;
      }
      if (hid == KAAPI_HWS_LEVELID_MAX)
        /* not correct */
        return false;

      eat_space(str);
      if (**str != '(') return false;
      ++*str;
      if (!kaapi_parse_unsigned_int(str, &id))
        return false;
      eat_space(str);
      if (**str != ')') return false;
      ++*str;

      /* fill set and negate set */
      if (kaapi_default_param.affinity_set[hid].count < id)
         kaapi_abort(__LINE__, __FILE__, "*** Error while parsing places: invalid ressource index.");
#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
      if (!negate)
        CPU_OR(&set, &set, &kaapi_default_param.affinity_set[hid].nis[id].cpuset);
      else
        CPU_OR(&negateset, &negateset, &kaapi_default_param.affinity_set[hid].nis[id].cpuset);
#endif
    }
    else
    {
      if (!kaapi_parse_range(str, &low, &high, &stride))
        return false;

      eat_space(str);
      /* fill set and negate set */
      for (unsigned int i=low; i<=high; i+= stride)
      {
#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
         if (!negate) CPU_SET(i, &set);
         else CPU_SET(i, &negateset);
#endif
      }
    }
    if (**str != ',')
      break;
    ++*str;
  } /* place loop */

  /* merge two sets together: finalset = set & ~negateset */
#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
  CPU_XOR( place, &set, &negateset );
  CPU_AND( place, place, &set );
  /* */
  unsigned int cpu = CPU_COUNT(place);
  *cpu_count = cpu;
#endif

  return true;
}


#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
static void CPU_SHIFT( kaapi_place_t* shifted, unsigned int shift )
{
   __cpu_mask old = 0;
   for (unsigned int i=0; i<__CPU_SETSIZE / __NCPUBITS; ++i)
   {
     __cpu_mask nextold = shifted->__bits[i];
     shifted->__bits[i] = (shifted->__bits[i] << shift) | (old >> (sizeof(__cpu_mask)*8 - shift));
     old = nextold;
   }
}
#endif


/* Grammar for parsing places as in OpenMP.
[ expr ] is for optional expression.
numas is for taking places from numa nodes that does not correspond
to socket on some machines.
places := threads [ ( num ) ] | cores [ ( num ) ] | sockets [ ( num ) ] | numas [ ( num ) ]
          | places_list;
place_list := { place } [ : num [ : stride ] ] [, place_list ]
place := range , [ place ]
range := [!] [num] [ : [num] [ : num ] ]
*/
bool kaapi_parse_places(
    char** str,
    unsigned int* count,
    unsigned int* cpu_count,
    kaapi_place_t** places
)
{
  kaapi_hws_levelid_t hid = KAAPI_HWS_LEVELID_MAX;
  if (*str ==0)
    return false;

  *places = 0;
  *count = 0;
  *cpu_count = 0;

  if (strncasecmp (*str, "threads", 7) == 0)
  {
    *str += 7;
    hid = KAAPI_HWS_LEVELID_THREAD;
  }
  else if (strncasecmp (*str, "cores", 5) == 0)
  {
    *str += 5;
    hid = KAAPI_HWS_LEVELID_CORE;
  }
  else if (strncasecmp (*str, "sockets", 7) == 0)
  {
    *str += 7;
    hid = KAAPI_HWS_LEVELID_SOCKET;
  }
  else if (strncasecmp (*str, "numas", 5) == 0)
  {
    *str += 5;
    hid = KAAPI_HWS_LEVELID_NUMA;
  }
  if (hid != KAAPI_HWS_LEVELID_MAX)
  {
    kaapi_mt_hwdetect(); /* required ! */

    unsigned int num;
    /* parse remaing '(' num ')' */
    eat_space(str);
    if (**str == '(')
    {
      ++*str;
      if (!kaapi_parse_unsigned_int(str, &num) || (**str != ')'))
        return false;
      ++*str;
    }
    else
      num = kaapi_default_param.affinity_set[hid].count;

    *places = malloc( sizeof(kaapi_place_t)*num );
    if (kaapi_default_param.affinity_set[hid].count < num)
       kaapi_abort(__LINE__, __FILE__, "*** Error while parsing places: Too many ressources in places.");

    for (unsigned int i=0; i<num; ++i)
    {
#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
      memcpy( &(*places)[i], &kaapi_default_param.affinity_set[hid].nis[i].cpuset, sizeof(cpu_set_t));
      *cpu_count += CPU_COUNT(&kaapi_default_param.affinity_set[hid].nis[i].cpuset);
#endif
    }

    *count = num;
    return true;
  }

  /* parse list */
  kaapi_place_t oneplace;

  while (1)
  {
    eat_space(str);
    if (**str != '{') return false;
    ++*str;
    unsigned int cpu;
    if (!kaapi_parse_oneplace(str, &cpu, &oneplace)) return false;
    eat_space(str);
    if (**str != '}') return false;
    ++*str;

    /* */
    if (cpu >0)
    {
      *places = realloc(*places, sizeof(kaapi_place_t)*(*count +1));
      memcpy(&(*places)[*count], &oneplace, sizeof(kaapi_place_t));
      ++*count;
      *cpu_count += cpu;
    }

    eat_space(str);
    if (**str == ':')
    {
      unsigned int nrep;
      unsigned int shift;
      ++*str;
      if (!kaapi_parse_unsigned_int(str, &nrep))
        return false;
      eat_space(str);
      if (**str != ':') return false;
      ++*str;
      if (!kaapi_parse_unsigned_int(str, &shift))
        return false;
      eat_space(str);

      /* */
      kaapi_place_t shifted = oneplace;
      memcpy(&shifted, &oneplace, sizeof(kaapi_place_t));
      for (unsigned int i=0; i<nrep-1; ++i)
      {
#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
        CPU_SHIFT( &shifted, shift );
        if (cpu >0)
        {
          *places = realloc(*places, sizeof(kaapi_place_t)*(*count +1));
          memcpy(&(*places)[*count], &shifted, sizeof(kaapi_place_t));
          ++*count;
          *cpu_count += cpu;
        }
#endif
      }
    }
    if (**str == 0)
      break;
    if (**str != ',') return false;

    ++*str;
  }
  return true;
}


/*
*/
char* kaapi_unparse_places_r(
    char* buffer, unsigned int buffer_size,
    unsigned int count,
    kaapi_place_t* places
)
{
  size_t size = 0;
  int s;
  char* pos = &buffer[0];
  memset(buffer, 0, buffer_size);

  for (unsigned int i=0; i<count; ++i)
  {
    s = snprintf(pos, buffer_size-size, "{");
    if (s<0) return buffer;
    size += s; pos += s;
#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
    kaapi_place_t* place = places+i;
    unsigned long j, max = 8 * sizeof(cpu_set_t), len;
    bool notfirst = false;
    for (j = 0, len = 0; j < max; j++)
      if (CPU_ISSET(j, place))
      {
        if (len == 0)
        {
          if (notfirst)
          {
            s = snprintf(pos, buffer_size-size, ",");
            if (s<0) return buffer;
            size += s; pos += s;
          }
          notfirst = true;
          s = snprintf (pos, buffer_size-size, "%lu", j);
          if (s<0) return buffer;
          size += s; pos += s;
        }
        ++len;
      }
      else
      {
        if (len > 1)
        {
          s = snprintf (pos, buffer_size-size, ":%lu", len);
          if (s<0) return buffer;
          size += s; pos += s;
        }
        len = 0;
      }

    if (len > 1)
    {
      s = snprintf (pos, buffer_size-size, ":%lu", len);
      if (s<0) return buffer;
      size += s; pos += s;
    }
#endif
    if (i < count-1)
      s = snprintf(pos, buffer_size-size, "},");
    else
      s = snprintf(pos, buffer_size-size, "}");
    if (s<0) return buffer;
    size += s; pos += s;
  }
  return buffer;
}



/*
*/
char* kaapi_unparse_places(
    unsigned int count,
    kaapi_place_t* places
)
{
  #define MAX_SIZE 4096
  static char buffer[MAX_SIZE];
  return kaapi_unparse_places_r(buffer, MAX_SIZE, count, places);
}
