/*
** kaapi_task_standard.c
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "kaapi_impl.h"

/*
*/
int kaapi_task_push_withstate(kaapi_thread_t* thread, kaapi_task_body_t body, void* arg, uint8_t state)
{
  kaapi_task_t* task = thread->sp;
  if (!_kaapi_has_enough_taskspace(thread,1))
    task = kaapi_thread_slow_pushtask(thread);
  task->body      = body;
  if (state == KAAPI_TASK_STATE_EXEC)
    task->state.steal     = 0;
  else if (state == KAAPI_TASK_STATE_STEAL)
    task->state.steal     = kaapi_slowexec_body;
  task->arg       = arg;
  task->u.dummy   = 0;
#if !defined(NDEBUG) && defined(KAAPI_DFG_ON_STEAL)
  task->next      = 0;
  KAAPI_ATOMIC_WRITE( &task->wc, 0);
#endif
  kaapi_writemem_barrier();
  ++thread->sp;
  return 0;
}


/*
*/
void kaapi_nop_body(
  kaapi_task_t*   task,
  kaapi_thread_t* thread
)
{
#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_processor_t* kproc = ((kaapi_context_t*)thread)->proc;
  if (kproc->perf !=0)
    ++KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_TASKS);
#endif
}


/*
*/
void kaapi_taskmain_body( 
  kaapi_task_t* task,
  kaapi_thread_t* thread __attribute__((unused))
)

{
  kaapi_taskmain_arg_t* arg = (kaapi_taskmain_arg_t*)kaapi_task_getargs(task);
  arg->mainentry( arg->argc, arg->argv );
#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_processor_t* kproc = ((kaapi_context_t*)thread)->proc;
  if (kproc->perf !=0)
  {
    ++KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_TASKS);
  }
#endif
}


/**
*/
void kaapi_aftersteal_body( kaapi_task_t* task, kaapi_thread_t* thread )
{
  unsigned int          i;
  size_t                count_params;
  const kaapi_format_t* fmt;   /* format of the task */
  const kaapi_format_t* fmt_param;
  kaapi_access_t*       access_param;
  void*                 taskarg = kaapi_task_getargs(task);
  
  /* the task has been stolen: the body contains the original task body */
  fmt = kaapi_format_resolvebybody( task->body );
  kaapi_assert_debug( fmt !=0 );

  count_params = kaapi_format_get_count_params(fmt, task );

  /* for each access parameter, we have:
     - data that points to the original effective parameter
     - version that points to the data that should have been used during the execution of the stolen task
     If the access is a W (or CW) access then, version contains the newly produced data.
     The purpose here is to merge the version into the original data
        - if W mode -> copy + free of the data
        - if CW mode -> accumulation (data is the left side of the accumulation, version the right side)
  */
  for (i=0; i<count_params; ++i)
  {
    kaapi_access_mode_t mode = kaapi_format_get_mode_param(fmt, i, taskarg);
    kaapi_access_mode_t m = KAAPI_ACCESS_GET_MODE( mode ); /* forget extra flags */
    if (m & KAAPI_ACCESS_MODE_V)
      continue;

    if (KAAPI_ACCESS_IS_ONLYWRITE(m) || KAAPI_ACCESS_IS_CUMULWRITE(m) )
    {
      fmt_param    = kaapi_format_get_fmt_param(fmt, i, taskarg);
      access_param = kaapi_format_get_access_param(fmt, i, taskarg);

      /* if m == W || CW and data == version it means that data used by W was not copied due 
         to non WAR dependency or no CW access 
      */
      if (access_param->data != access_param->version )
      {
        /* add an assign + dstor function will avoid 2 calls to function, especially for basic types which do not
           required to be dstor.
        */
        if (!KAAPI_ACCESS_IS_CUMULWRITE(m))
        {
          kaapi_memory_view_t view_dest;
          kaapi_format_get_view_param(fmt, i, taskarg, &view_dest);
          kaapi_memory_view_t view_src;
          
          /* in case of WAR resolution, view_src is == to reallocation of view_src */
          view_src = view_dest;
          kaapi_memory_view_reallocated( &view_src );

          /* here we assume that if no assignment, then this is a pod type, so we use memcpy version for cpu2cpu
             If version is produced remotely... version should store the location...
          */
          if (fmt_param->assign ==0)
          {
            kaapi_pointer_t     ptr_dest, ptr_src;
            kaapi_void2pointer(&ptr_dest, access_param->data);
            kaapi_void2pointer(&ptr_src, access_param->version);
            kaapi_memory_copy(ptr_dest, &view_dest, ptr_src, &view_src);
          }
          else
            (*fmt_param->assign)( access_param->data, &view_dest, access_param->version, &view_src );
        }
        else {
          /* this is a CW mode : do reduction except if it is INPLACE */
          if ((mode & KAAPI_ACCESS_MODE_IP) ==0)
            kaapi_format_reduce_param( fmt, i, taskarg, access_param->version );
        }
        if ((mode & KAAPI_ACCESS_MODE_IP) ==0)
        {
          if (fmt_param->dstor !=0) (*fmt_param->dstor) ( access_param->version );
          free(access_param->version);
        }
      }
      access_param->version = 0;
    }
  }
#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_processor_t* kproc = ((kaapi_context_t*)thread)->proc;
  if (kproc->perf !=0)
  {
    ++KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_TASKS);
  }
#endif
}





/**
*/
void kaapi_taskwrite_body( 
  kaapi_task_t*   task,
  kaapi_thread_t* thread
)
{
  unsigned int i;
  size_t count_params;

  const kaapi_format_t* fmt;
  void*                 orig_task_args;
  kaapi_access_t*       access_param;

  kaapi_access_mode_t   mode;
  kaapi_access_mode_t   mode_param;
  const kaapi_format_t* fmt_param;
  unsigned int          war_param;     /* */
  unsigned int          cw_param;     /* */

  void*                 copy_task_args;
  void*                 copy_data_param;
  kaapi_access_t*       copy_access_param;

  kaapi_tasksteal_arg_t* arg = ((kaapi_taskwrite_arg_t*)kaapi_task_getargs(task))->stealarg;
  orig_task_args             = kaapi_task_getargs(arg->origin_task);
  copy_task_args             = arg->copy_task_args;
  war_param                  = arg->war_param;
  cw_param                   = arg->cw_param;

  if (arg->origin_task->u.s.flag & KAAPI_TASK_FLAG_DFGOK)
    kaapi_abort(__LINE__,__FILE__, "Invalid case");

  if (copy_task_args !=0)
  {
    /* for each parameter of the copy of the theft' task on mode:
       - V: we destroy the data
       - R,RW: do nothing
       - W,CW: set in field ->version of the original task args the field ->data of the copy args.
    */
    fmt         = arg->origin_task->fmt;
    kaapi_assert_debug(fmt !=0);

    count_params = kaapi_format_get_count_params( fmt, orig_task_args );
    for (i=0; i<count_params; ++i)
    {
      mode            = kaapi_format_get_mode_param(fmt, i, copy_task_args); 
      mode_param      = KAAPI_ACCESS_GET_MODE( mode ); 
      if (mode_param & KAAPI_ACCESS_MODE_V)
      {
        fmt_param       = kaapi_format_get_fmt_param(fmt, i, orig_task_args);
        copy_data_param = (void*)kaapi_format_get_data_param(fmt, i, copy_task_args);
        (*fmt_param->dstor)(copy_data_param);
        continue;
      }

      if ( KAAPI_ACCESS_IS_ONLYWRITE(mode_param) 
       || (KAAPI_ACCESS_IS_CUMULWRITE(mode_param) && ((mode & KAAPI_ACCESS_MODE_IP) ==0)))
      {
        access_param      = kaapi_format_get_access_param(fmt, i, orig_task_args);
        copy_access_param = kaapi_format_get_access_param(fmt, i, copy_task_args);

        /* write the value as the version */
        kaapi_assert_debug( access_param->version == 0);
        access_param->version = copy_access_param->data;
        kaapi_format_set_access_param(fmt, i, orig_task_args, access_param );
      }
      else if (KAAPI_ACCESS_IS_STACK(mode_param))
      { /* never merge result here */
        copy_access_param = kaapi_format_get_access_param(fmt, i, copy_task_args);
//Temporary delete free(copy_access_param.data);
        copy_access_param->data = 0;

        /* suppress war_param bit to avoid pushing a merge task */
        war_param &= ~(1<<i);
#if 0
kaapi_memory_view_t view = kaapi_format_get_view_param(fmt, i, copy_task_args);
printf("Delete temporary stack object: size=%i\n",(int)kaapi_memory_view_size(&view) );
#endif
      }
    }
  }

#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_processor_t* kproc = ((kaapi_context_t*)thread)->proc;
  if (kproc->perf !=0)
  {
    ++KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_TASKS);
  }
#endif

  /* order write to the memory before changing the origin task' state */
  kaapi_writemem_barrier();

  /* signal the original task : mark it as executed, the old returned body should have steal flag */
  if ((war_param ==0) && (cw_param ==0))
    kaapi_task_markterm( arg->origin_task );
  else 
    kaapi_task_markaftersteal( arg->origin_task );

#if 0
KAAPI_ATOMIC_PRINT(
  fprintf(stdout,"%i::end steal", kaapi_self_kid());
  kaapi_task_print(stdout, arg->origin_task );
)
#endif
}


/* copy task argument 
*/
static void* __kaapi_cpy_taskarg(
  kaapi_thread_t* thread,
  const kaapi_format_t*  fmt,
  kaapi_tasksteal_arg_t* arg,
  unsigned int war_param,
  unsigned int cw_param
)
{
  /* the original task arguments */
  void*                  data_param;
  kaapi_access_t*        access_param;

  kaapi_access_mode_t    mode;
  kaapi_access_mode_t    mode_param;
  const kaapi_format_t*  fmt_param;

  void*                  copy_task_args;
  void*                  copy_data_param;
  kaapi_access_t*        copy_access_param;
  void* orig_task_args  = kaapi_task_getargs(arg->origin_task);
  size_t count_params    = kaapi_format_get_count_params(fmt, orig_task_args);
  unsigned int i;

  size_t task_size     = kaapi_format_get_size( fmt, orig_task_args );
  copy_task_args       = kaapi_data_push( thread, (uint32_t)task_size);
  arg->copy_task_args  = copy_task_args;

  kaapi_format_task_copy( fmt, copy_task_args, orig_task_args );

  for (i=0; i<count_params; ++i)
  {
    mode            = kaapi_format_get_mode_param(fmt, i, orig_task_args); 
    mode_param      = KAAPI_ACCESS_GET_MODE( mode ); 
    fmt_param       = kaapi_format_get_fmt_param(fmt, i, orig_task_args);
    
    if (mode_param & KAAPI_ACCESS_MODE_V)
    {
      data_param      = kaapi_format_get_data_param(fmt, i, orig_task_args);
      copy_data_param = kaapi_format_get_data_param(fmt, i, copy_task_args);
      (*fmt_param->cstorcopy)(copy_data_param, data_param);
      continue;
    }

    /* initialize all parameters ... */
    access_param              = kaapi_format_get_access_param(fmt, i, orig_task_args);
    copy_access_param         = kaapi_format_get_access_param(fmt, i, copy_task_args);
    copy_access_param->data    = access_param->data;
    copy_access_param->version = 0; /*access_param->version; not required... */

    KAAPI_DEBUG_INST(access_param->version = 0;)
    
    if (KAAPI_ACCESS_IS_STACK(mode_param))
    {
      copy_access_param->data = 0;
      kaapi_memory_view_t view;
      kaapi_format_get_view_param(fmt, i, orig_task_args, &view);
      if (kaapi_memory_view_size(&view) < 1024 )
      { /* try to do stack allocation */
        copy_access_param->data = kaapi_data_push(thread, (int)kaapi_memory_view_size(&view));
      }
      if (copy_access_param->data ==0)
      {
#if defined(KAAPI_DEBUG)
        copy_access_param->data   = calloc(1, kaapi_memory_view_size(&view));
#else
        copy_access_param->data   = malloc(kaapi_memory_view_size(&view));
#endif
      }
      kaapi_assert_debug( copy_access_param->data != 0 );
      
      if (fmt_param->cstor !=0) 
        (*fmt_param->cstor)(copy_access_param->data);

      kaapi_memory_view_t view_dest = view;
      kaapi_memory_view_reallocated( &view_dest );
      (*fmt_param->assign)(copy_access_param->data, &view_dest, access_param->data, &view );
      
      /* set new data to copy with new view */
      kaapi_format_set_access_param(fmt, i, copy_task_args, copy_access_param );
      kaapi_format_set_view_param( fmt, i, copy_task_args, &view_dest );
    }
    /* allocate new data for the war or cw param or stack data, else points to the original data
       if mode is CW and Inplace flag is set, then do not copy data.
    */
    else if ( KAAPI_ACCESS_IS_ONLYWRITE(mode_param) 
          || (KAAPI_ACCESS_IS_CUMULWRITE(mode_param) && ((mode & KAAPI_ACCESS_MODE_IP) ==0)) 
    )
    {
      if (((war_param & (1<<i)) !=0) || ((cw_param & (1<<i)) !=0))
      { 
        kaapi_memory_view_t view;
        kaapi_format_get_view_param(fmt, i, orig_task_args, &view);
#if defined(KAAPI_DEBUG)
        copy_access_param->data   = calloc(1, kaapi_memory_view_size(&view));
#else
        copy_access_param->data   = malloc(kaapi_memory_view_size(&view));
#endif
        if (fmt_param->cstor !=0) 
          (*fmt_param->cstor)(copy_access_param->data);
        if (KAAPI_ACCESS_IS_STACK(mode_param))
        {
          kaapi_memory_view_t view_dest = view;
          kaapi_memory_view_reallocated( &view_dest );
          (*fmt_param->assign)(copy_access_param->data, &view_dest, access_param->data, &view );
        }
        else
          /* if cw: init with neutral element with respect to the reduction */
          if ((cw_param & (1<<i)) !=0)
            kaapi_format_redinit_neutral(fmt, i, copy_task_args, copy_access_param->data );
        
        /* set new data to copy with new view */
        kaapi_format_set_access_param(fmt, i, copy_task_args, copy_access_param );
        kaapi_memory_view_reallocated( &view );
        kaapi_format_set_view_param( fmt, i, copy_task_args, &view );
      }
    }
  }
  return copy_task_args;
}


/**
*/
void kaapi_tasksteal_body( kaapi_task_t* task, kaapi_thread_t* thread  )
{
  kaapi_taskwrite_arg_t* taskwrite;
  kaapi_tasksteal_arg_t* arg;
  kaapi_task_body_t      body;          /* format of the stolen task */
  const kaapi_format_t*  fmt;
  unsigned int           war_param;     /* */
  unsigned int           cw_param;      /* */

  void*                  orig_task_args;
  void*                  copy_task_args;

#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_processor_t* kproc = kaapi_self_processor();
#endif

  /* get information of the task to execute */
  arg = (kaapi_tasksteal_arg_t*)kaapi_task_getargs(task);

  /* format of the original stolen task, resolved it if not yet set */  
  body            = arg->origin_task->body;
  fmt             = arg->origin_task->fmt;
  /* the original task arguments */
  orig_task_args  = kaapi_task_getargs(arg->origin_task);

  /* DFG & successors = activate tasks locally
     Startup if stolen task is DFG ok, it to initialize the readlist for the current frame.
  */
  if (arg->origin_task->u.s.flag & KAAPI_TASK_FLAG_DFGOK)
  {
    kaapi_stack_t* stack = (kaapi_stack_t*)thread;
    kaapi_frame_t* frame = stack->unlink;
    kaapi_frame_wrdlist_t* rdlist;

    if (fmt ==0)
      fmt = _kaapi_task_getformat(arg->origin_task);
    kaapi_assert_debug( fmt != 0);
    kaapi_assert_debug( body == fmt->entrypoint[KAAPI_PROC_TYPE_HOST] );

    kaapi_assert_debug( (frame->flag & KAAPI_FRAME_FLAG_MASK_STATE) == KAAPI_FRAME_FLAG_DFG_VOID);
    kaapi_assert_debug( frame->rdlist == 0);

    frame->rdlist = rdlist = (kaapi_frame_wrdlist_t*)kaapi_alloca( thread, sizeof(kaapi_frame_wrdlist_t) + sizeof(kaapi_frame_dfgctxt_t));
    frame->flag |= KAAPI_FRAME_FLAG_STACKALLOC;
    kaapi_frame_rdlist_init(rdlist, (kaapi_frame_dfgctxt_t*)(rdlist+1));
    rdlist->master = arg->origin_frame_rd;

    /* task pushed not realy used & execute : execution is through ready list, but thread could
       not be empty.
    */
    if (arg->origin_task->u.s.flag & KAAPI_TASK_FLAG_COPY_ON_STEAL)
    {
      copy_task_args = __kaapi_cpy_taskarg(
        thread,
        fmt,
        arg,
        0U,
        0U
      );
    }
    else
      copy_task_args = orig_task_args;

    /* push the task with not stolen flag */
    kaapi_task_push_withflags(
        thread,
        body,
        copy_task_args,
        arg->origin_task->u.s.flag | KAAPI_TASK_FLAG_UNSTEALABLE | KAAPI_TASK_FLAG_DFGOK
    );
    stack->pc->fmt         = fmt;
    stack->pc->state.frame = arg->origin_frame_rd;
    rdlist->starttask      = stack->pc;
    frame->flag           |= KAAPI_FRAME_FLAG_DFG_OK;

    return;
  }

  /**/
  war_param = arg->war_param;
  cw_param  = arg->cw_param;

  /* not a bound task */
  arg->copy_task_args = 0;

  if (!war_param && !cw_param)
  {
    /* Execute the orinal body function with the original args.
    */
    KAAPI_EVENT_KPROC_PUSH1(kproc, kproc->thread, KAAPI_EVT_TASK_BEG, (fmt ? fmt->fmtid :0) );
    body( arg->origin_task, thread );
    KAAPI_EVENT_KPROC_PUSH0(kproc, kproc->thread, KAAPI_EVT_TASK_END);
  }
  else /* it exists at least one w parameter with war dependency or a cw_param: recopies the arguments */
  {
    copy_task_args = __kaapi_cpy_taskarg(
      thread,
      fmt,
      arg,
      war_param,
      cw_param
    );

    /* call directly the stolen body function */
    kaapi_task_t newtask;
    newtask.body  = body;
    newtask.state.steal = 0;
    newtask.arg   = copy_task_args;

    KAAPI_EVENT_KPROC_PUSH1(kproc, kproc->thread, KAAPI_EVT_TASK_BEG, (fmt ? fmt->fmtid :0) );
    body( &newtask, thread);
    KAAPI_EVENT_KPROC_PUSH0(kproc, kproc->thread, KAAPI_EVT_TASK_END);
  }

  /* push task that will be executed after all created tasks spawned
     by the user task 'body'
     - this task is unstealable because it does not has format
  */
  taskwrite = (kaapi_taskwrite_arg_t*)kaapi_data_push(thread, sizeof(kaapi_taskwrite_arg_t));
  taskwrite->stealarg = arg;
  kaapi_task_push( thread, kaapi_taskwrite_body, taskwrite);
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    ++KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_TASKS);
  }
#endif
}
