/*
** kaapi_init.c
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@gmail.com 
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "kaapi_impl.h"
#include "kaapi_util.h"
#include <strings.h>
#include <stdio.h>

#if defined(KAAPI_DEBUG)
#  include <unistd.h>
#  include <sys/time.h>
#  include <signal.h>
#endif

#if defined(KAAPI_USE_PERFCOUNTER)
#include <stdio.h>
#include <signal.h>
#endif
#if defined(__linux__)
#include <unistd.h>
#endif
#if defined(__APPLE__)
#include <sys/sysctl.h>
#endif

#if defined(KAAPI_DEBUG)
#include <float.h>
double kaapi_debug_threshold_task_duration = DBL_MAX;
#endif


/** The initial team
*/
kaapi_team_t* kaapi_global_team = 0;

extern kaapi_team_t* kaapi_cached_global_team;

/** The default parameters for Kaapi
*/
kaapi_rtparam_t kaapi_default_param = {
   .startuptime = 0,
#if defined(KAAPI_USE_CUPTI)
   .cudastartuptime = 0,
#endif
#if defined(KAAPI_USE_CUDA)
   .cudawindowsize = 1,
   .cudapeertopeer = 0,
#endif
   .stackblocsize  = KAAPI_STACKBLOCSIZE, /* the kaapi STACK for tasks and args */
   .threadstacksize= 0,  /* the pthread stack size */
   .syscpucount    = 0,
   .cpucount       = 0,
   .cpuset_str     = 0,
   .cpucount_str   = 0,
   .gpucount       = 0,
   .alarmperiod    = 0,

   /* selection of the victim processor */
   .wsselect           = &kaapi_sched_select_victim_rand,
   .wslocalsteal       = &kaapi_sched_select_localsteal_hierarchy,
#if defined(KAAPI_USE_NUMA)
   .wspush             = &kaapi_sched_select_queue_localnuma,
#else
   .wspush             = &kaapi_sched_select_queue_local,
#endif

   /*init in priority CC_SYNC, bitmap and no aggr */
#if defined(KAAPI_USE_FLAT_CC_SYNC)
   .emitsteal          = kaapi_sched_ccsync_emitsteal,
   .emitsteal_initctxt = kaapi_sched_ccsync_emitsteal_init,
   .emitsteal_dstorctxt= kaapi_sched_ccsync_emitsteal_dstor,
#else
   .emitsteal          = kaapi_sched_noaggr_emitsteal,
   .emitsteal_initctxt = kaapi_sched_noaggr_emitsteal_init,
   .emitsteal_dstorctxt= kaapi_sched_noaggr_emitsteal_dstor,
#endif
  
#if defined(KAAPI_USE_PERFMODEL)
   .perfmodel           = 0,
   .perfmodel_calibrate = 0,
   .noremotesteal       = 0,
#endif
   .use_affinity        = 0,
   .affinity_set        = {{0,0,0},  /* threads */
                           {0,0,0},  /* cores */
                           {0,0,0},  /* numas */
                           {0,0,0},  /* sockets */
                           {0,0,0},  /* board */
                           {0,0,0}}, /* machine */
   .procbind            = KAAPI_PROCBIND_DEFAULT,
   .places              = {0,0},
   .ctpriority          = 0,
#if defined(KAAPI_USE_PERFCOUNTER)
   .eventmask           = KAAPI_EVT_MASK_COMPUTE|KAAPI_EVT_MASK_IDLE|KAAPI_EVT_MASK_PERFCOUNTER,
#else
   .eventmask           = 0,
#endif
   .display_perfcounter = KAAPI_NO_DISPLAY_PERF,
   .papi_event_count    = 0,
   .trace_recorder      = 0
};


/* == 0 if lib is not initialized
*/
static kaapi_atomic_t kaapi_count_init = {0};

/*
*/
#if defined(KAAPI_HAVE_COMPILER_TLS_SUPPORT)
__thread kaapi_thread_t*        kaapi_current_thread_key;
__thread kaapi_processor_t*     kaapi_current_processor_key;
#else
pthread_key_t kaapi_current_processor_key;

/**
*/
kaapi_thread_t* kaapi_self_thread(void)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
  return &kproc->context->stack.thread;
}
#endif

/**
*/
unsigned int kaapi_self_kid(void)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
  if (kproc ==0) return -1;
  return (unsigned int)kproc->kid;
}

/**
*/
kaapi_team_t* kaapi_self_team(void)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
  if (kproc ==0) return kaapi_global_team;
  return kproc->team;
}


/**
*/
int kaapi_get_concurrency(void )
{
  kaapi_processor_t* kproc = kaapi_self_processor();
  if (kproc ==0) 
    return 1; // kaapi not initialized
  if (kproc->team ==0)
    return 1; // kproc not in a team
  return kproc->team->count;
}


/**
*/
int kaapi_mt_registerproc(void)
{
  int err = kaapi_mt_hwdetect();
  if ((err !=0) && (err != EALREADY))  return err;

  /* WARNING: here should allocate at lest cpucount+gpucount descriptor
   Currently allocate maximal number of Kprocessor */
  int use_cpu = 0;
  kaapi_default_param.cpucount = 0;

  /* KAAPI_CPUCOUNT */
  if (kaapi_default_param.cpucount_str != 0)
    use_cpu = atoi(kaapi_default_param.cpucount_str);

  /* KAAPI_CPUSET */
  if (kaapi_default_param.cpuset_str != 0)
  {
    unsigned int count = 0;
    unsigned int cpu_count = 0;
    kaapi_place_t* places = 0;
    char* toparse = strdup(kaapi_default_param.cpuset_str);
    char* save = toparse;
    bool err = kaapi_parse_places( &toparse, &count, &cpu_count, &places );
    kaapi_assert(err);
    free(save);

#if 0 //defined(KAAPI_USE_SCHED_AFFINITY)
char buffer[1024];
char buffer2[1024];
    printf("Places: #count=%i, #cpus=%i\n", count, cpu_count );
    for (int i=0; i<count; ++i)
    {
      printf("place='%s', cpuset='%s'\n", kaapi_unparse_places_r(buffer, 256, 1, &places[i]), kaapi_cpuset2string_r(buffer2, 256, 48, &places[i]));
    }
#endif

    /* ok : use affinity */
    kaapi_default_param.use_affinity = 1;

    kaapi_default_param.places.first = places;
    kaapi_default_param.places.len = count;
    if (use_cpu ==0) use_cpu = cpu_count;
  }
  else
    if (use_cpu ==0) use_cpu = kaapi_default_param.syscpucount;

  kaapi_default_param.cpucount = use_cpu;

  //printf("[kaapi] use %i cpus\n", use_cpu );

  return 0;
}


/**
*/
static void _kaapi_init_nodeinfo(  kaapi_nodeinfo_t* ni, kaapi_ldid_t i )
{
  ni->lid = i;
  kaapi_atomic_initlock( &ni->lock );
  kaapi_lfree_clear( &ni->lfree );
  ni->freebloc = 0;
#if defined(__linux__)
  CPU_ZERO(&ni->cpuset);
#endif
}


/**
*/
static kaapi_atomic_t is_hwdetect_called = {0};
int kaapi_mt_hwdetect(void)
{
  if (KAAPI_ATOMIC_INCR(&is_hwdetect_called) >1)
    return EALREADY;

  const char* volatile version __attribute__((unused));

  version = get_kaapi_version();

  /* compute the number of cpu of the system */
#if defined(__linux__)
  kaapi_default_param.syscpucount = sysconf(_SC_NPROCESSORS_CONF);
#elif defined(__APPLE__)
  {
    int mib[2];
    size_t len;
    mib[0] = CTL_HW;
    mib[1] = HW_NCPU;
    len = sizeof(kaapi_default_param.syscpucount);
    sysctl(mib, 2, &kaapi_default_param.syscpucount, &len, 0, 0);
  }
#elif defined(_WIN32)
  SYSTEM_INFO sys_info;
  GetSystemInfo(&sys_info);
  kaapi_default_param.syscpucount = sys_info.dwNumberOfProcessors;
#else
#  warning "Could not compute number of physical cpu of the system. Default value==1"
  kaapi_default_param.syscpucount = 1;
#endif

  kaapi_default_param.use_affinity = 0;

  kaapi_affinityset_t* affset = &kaapi_default_param.affinity_set[0];
  for (int i=0; i<KAAPI_HWS_LEVELID_MAX; ++i)
  {
    affset[i].hid   = i; /* thread, core, socket, numa */
    affset[i].count = 0;
    affset[i].nis = 0;
  }

  /* Assume one MACHINE node */
  affset[KAAPI_HWS_LEVELID_MACHINE].count = 1;
  affset[KAAPI_HWS_LEVELID_MACHINE].nis = calloc( 1, sizeof(kaapi_nodeinfo_t) );
  kaapi_assert(affset[KAAPI_HWS_LEVELID_MACHINE].nis !=0);
  for (int i=0; i<affset[KAAPI_HWS_LEVELID_MACHINE].count; ++i)
  {
    kaapi_nodeinfo_t* ni = &affset[KAAPI_HWS_LEVELID_MACHINE].nis[i];
    _kaapi_init_nodeinfo( ni, i );
#if defined(__linux__) && defined(KAAPI_USE_SCHED_AFFINITY)
    for (int k=0; k<kaapi_default_param.syscpucount; ++k)
      CPU_SET(k, &ni->cpuset);
#endif

  }

  /* detect the topology */
#if defined(__linux__)
  char filename[sizeof("/sys/devices/system/cpu/cpu1024/topology/thread_siblings_list") + 16];
  FILE *file;

  /* 1/ parse numa node informations */
  memcpy(filename,
      "/sys/devices/system/node/online",
      sizeof("/sys/devices/system/node/online"));
  file = fopen(filename,"r");
  if (file !=0)
  {
    char *line = 0;
    char *str = 0;
    size_t len = 0;
    kaapi_assert( getline (&line, &len, file) > 0);
    str = line;

    /* scan num '-' num */
    unsigned int start, end;
    kaapi_assert(kaapi_parse_range(&str, &start, &end, 0));
    fclose( file );
    free(line);

    affset[KAAPI_HWS_LEVELID_NUMA].count = 1+end-start;
    affset[KAAPI_HWS_LEVELID_NUMA].nis = calloc( (1+end-start), sizeof(kaapi_nodeinfo_t) );
    kaapi_assert(affset[KAAPI_HWS_LEVELID_NUMA].nis !=0);
    for (int i=0; i<affset[KAAPI_HWS_LEVELID_NUMA].count; ++i)
    {
      sprintf(filename, "/sys/devices/system/node/node%i/cpulist", i);
      file = fopen(filename,"r");
      kaapi_assert( file != 0 );
      line = 0;
      kaapi_assert( getline (&line, &len, file) > 0);
      str = line;
      fclose(file);

      kaapi_nodeinfo_t* ni = &affset[KAAPI_HWS_LEVELID_NUMA].nis[i];
      _kaapi_init_nodeinfo( ni, i );

      unsigned int cpu;
      kaapi_assert( kaapi_parse_oneplace(&str, &cpu, &ni->cpuset) );
      free(line);
    }
  }
  else
    printf("[kaapi] Cannot access to topology information\n");


  /* 1/ parse thread/core/sockets informations */
  memcpy(filename,
      "/sys/devices/system/cpu/online",
      sizeof("/sys/devices/system/cpu/online"));
  file = fopen(filename,"r");
  if (file !=0)
  {
    char *line = 0;
    char *str = 0;
    size_t len = 0;
    kaapi_assert( getline (&line, &len, file) > 0);
    str = line;
    fclose( file );

    /* scan num '-' num */
    unsigned int start, end;
    kaapi_assert(kaapi_parse_range(&str, &start, &end, 0));
    free(line);

    cpu_set_t core_allcpu, socket_allcpu;
    CPU_ZERO(&socket_allcpu);
    for (int i=start; i<=end; ++i)
      CPU_SET(i, &socket_allcpu);

    //printf("[kaapi] %i cpus online\n",1+end-start);

    /* init thread level : list of PU */
    affset[KAAPI_HWS_LEVELID_THREAD].count = 1+end-start;
    affset[KAAPI_HWS_LEVELID_THREAD].nis = calloc( (1+end-start), sizeof(kaapi_nodeinfo_t) );
    kaapi_assert(affset[KAAPI_HWS_LEVELID_THREAD].nis !=0);
    for (int i=0; i<affset[KAAPI_HWS_LEVELID_THREAD].count; ++i)
    {
      kaapi_nodeinfo_t* ni = &affset[KAAPI_HWS_LEVELID_THREAD].nis[i];
      _kaapi_init_nodeinfo( ni, i );
      CPU_SET(i, &ni->cpuset);
    }

    /* Init core and socket levels */
    affset[KAAPI_HWS_LEVELID_CORE].count = 0;
    affset[KAAPI_HWS_LEVELID_SOCKET].count = 0;
    /* reserve up to end-start places */
    affset[KAAPI_HWS_LEVELID_CORE].nis = calloc( (1+end-start), sizeof(kaapi_nodeinfo_t) );
    kaapi_assert(affset[KAAPI_HWS_LEVELID_CORE].nis !=0);
    affset[KAAPI_HWS_LEVELID_SOCKET].nis = calloc( (1+end-start), sizeof(kaapi_nodeinfo_t) );
    kaapi_assert(affset[KAAPI_HWS_LEVELID_SOCKET].nis !=0);


    memcpy( &core_allcpu, &socket_allcpu, sizeof(cpu_set_t));
    for (int i=0; i<affset[KAAPI_HWS_LEVELID_THREAD].count; ++i)
    {
      for (kaapi_hws_levelid_t hws =KAAPI_HWS_LEVELID_CORE;
           hws != KAAPI_HWS_LEVELID_NUMA;
           hws = KAAPI_HWS_LEVELID_NUMA
      )
      {
        cpu_set_t* allcpu = (hws == KAAPI_HWS_LEVELID_CORE ? &core_allcpu : &socket_allcpu);
        if (CPU_ISSET(i, allcpu))
        {
          sprintf(filename, "/sys/devices/system/cpu/cpu%i/topology/%s_siblings_list",
              i,
              (hws == KAAPI_HWS_LEVELID_CORE ? "thread" : "core" )
          );
          file = fopen(filename,"r");
          kaapi_assert( file != 0 );
          line = 0;
          kaapi_assert( getline (&line, &len, file) > 0);
          str = line;
          fclose( file );

          int alloc_place =0;
          unsigned int cpu;
          kaapi_place_t place;
          kaapi_assert( kaapi_parse_oneplace(&str, &cpu, &place) );
          free(line);
          for (int k=0; k<sizeof(cpu_set_t)*8; ++k)
          {
            if (CPU_ISSET(k, &place) && CPU_ISSET( k, allcpu ))
            {
              CPU_CLR(k, allcpu);
              if (k == i) alloc_place = 1;
            }
          }
          if (alloc_place)
          {
            kaapi_nodeinfo_t* ni = &affset[hws].nis[ affset[hws].count ];
            _kaapi_init_nodeinfo( ni, i );
            memcpy(&ni->cpuset, &place, sizeof(cpu_set_t));
            ++affset[hws].count;
          }
        }
      }
    }
  }

#if 0
  printf("#places 'KAAPI_HWS_LEVELID_THREAD' : %i\n",affset[KAAPI_HWS_LEVELID_THREAD].count);

  printf("#places 'KAAPI_HWS_LEVELID_CORE'   : %i\n",affset[KAAPI_HWS_LEVELID_CORE].count);
#if 0
  for (int i=0; i<affset[KAAPI_HWS_LEVELID_CORE].count; ++i)
    printf("[%i]: '%s'\n", i, kaapi_unparse_places(1,&affset[KAAPI_HWS_LEVELID_CORE].nis[i].cpuset));
#endif

  printf("#places 'KAAPI_HWS_LEVELID_SOCKET' : %i\n",affset[KAAPI_HWS_LEVELID_SOCKET].count);
  for (int i=0; i<affset[KAAPI_HWS_LEVELID_SOCKET].count; ++i)
    printf("[%i]: '%s'\n", i, kaapi_unparse_places(1,&affset[KAAPI_HWS_LEVELID_SOCKET].nis[i].cpuset));

  printf("#places 'KAAPI_HWS_LEVELID_NUMA'   : %i\n",affset[KAAPI_HWS_LEVELID_NUMA].count);
  for (int i=0; i<affset[KAAPI_HWS_LEVELID_NUMA].count; ++i)
    printf("[%i]: '%s'\n", i, kaapi_unparse_places(1,&affset[KAAPI_HWS_LEVELID_NUMA].nis[i].cpuset));
  printf("#places 'KAAPI_HWS_LEVELID_MACHINE': %i\n",affset[KAAPI_HWS_LEVELID_MACHINE].count);
  for (int i=0; i<affset[KAAPI_HWS_LEVELID_MACHINE].count; ++i)
    printf("[%i]: '%s'\n", i, kaapi_unparse_places(1,&affset[KAAPI_HWS_LEVELID_MACHINE].nis[i].cpuset));
#endif

#else // not if defined(__linux__)
  /* Assume one NUMA node */
  affset[KAAPI_HWS_LEVELID_NUMA].count = 1;
  affset[KAAPI_HWS_LEVELID_NUMA].nis = calloc( 1, sizeof(kaapi_nodeinfo_t) );
  kaapi_assert(affset[KAAPI_HWS_LEVELID_NUMA].nis !=0);
  for (int i=0; i<affset[KAAPI_HWS_LEVELID_NUMA].count; ++i)
  {
    kaapi_nodeinfo_t* ni = &affset[KAAPI_HWS_LEVELID_NUMA].nis[i];
    _kaapi_init_nodeinfo( ni, i );
  }
#endif

  return 0;
}



/**
*/
int kaapi_mt_init(int flag)
{
  const char* volatile version __attribute__((unused));
  int                  err;

  if (KAAPI_ATOMIC_INCR(&kaapi_count_init) >1)
    return EALREADY;

  /* format init */
  kaapi_format_init();

  /* set param cpucount, places */
  kaapi_mt_registerproc();

  /* code to move again to kaapi_init */
  const char* str = getenv("KAAPI_WSSELECT");
  kaapi_default_param.wsselect = &kaapi_sched_select_victim_rand;
  kaapi_default_param.wslocalsteal = 0;
  if (str !=0)
  {
    if (strcasecmp(str, "rand") ==0)
      kaapi_default_param.wsselect = &kaapi_sched_select_victim_rand;
    else if (strcasecmp(str, "hws") ==0)
      kaapi_default_param.wsselect = &kaapi_sched_select_victim_hierarchy;
    else if (strcasecmp(str, "numa") ==0)
      kaapi_default_param.wsselect = &kaapi_sched_select_victim_numa;
    else {
      fprintf(stderr, "***Kaapi: bad value for variable KAAPI_WSSELECT: '%s'\n", str);
      return EINVAL;
    }
  }
  str = getenv("KAAPI_WSLOCAL");
  if (str !=0)
  {
    if (strcasecmp(str, "hws") ==0)
      kaapi_default_param.wslocalsteal = &kaapi_sched_select_localsteal_hierarchy;
    else if (strcasecmp(str, "null") ==0)
      kaapi_default_param.wslocalsteal = 0;
    else {
      fprintf(stderr, "***Kaapi: bad value for variable KAAPI_WSLOCAL: '%s'\n", str);
      return EINVAL;
    }
  }
  str = getenv("KAAPI_WSPUSH");
  if (str !=0)
  {
    if (strcasecmp(str, "local") ==0)
      kaapi_default_param.wspush = &kaapi_sched_select_queue_local;
    else if (strcasecmp(str, "numa") ==0)
      kaapi_default_param.wspush = &kaapi_sched_select_queue_localnuma;
    else if (strcasecmp(str, "Wnuma") ==0)
      kaapi_default_param.wspush = &kaapi_sched_select_queue_Wnuma;
    else if (strcasecmp(str, "attr") ==0)
      kaapi_default_param.wspush = &kaapi_sched_select_queue_attr;
    else {
      fprintf(stderr, "***Kaapi: bad value for variable KAAPI_WSPUSH: '%s'\n", str);
      return EINVAL;
    }
  }


  /* initialize the kprocessor key */
#if defined(KAAPI_HAVE_COMPILER_TLS_SUPPORT)
  kaapi_current_thread_key = 0;
#else
  kaapi_assert( 0 == pthread_key_create( &kaapi_current_processor_key, 0 ) );
#endif

  /* init numa allocation functions */
  kaapi_numa_initialize();

   /* start up time */
  kaapi_default_param.startuptime = kaapi_get_elapsedns();

#if defined(KAAPI_USE_PERFCOUNTER)
  /* call prior setconcurrency */
  kaapi_perf_global_init();
#endif

  /* */
  if (flag == KAAPI_START_ALL_THREADS)
  {
    kaapi_global_team = kaapi_team_init( 0,
      (int)kaapi_default_param.cpucount,
      kaapi_default_param.procbind,
      &kaapi_default_param.places,
      KAAPI_TEAM_DEFAULT | KAAPI_TEAM_LDSET
    );
    err = kaapi_team_start( kaapi_global_team );
    kaapi_assert(err ==0);
  }
  else
  {
    kaapi_global_team = 0;
  }
  

#if 0//defined(KAAPI_DEBUG)
  printf("#SYS CPU:%i\n", kaapi_default_param.syscpucount);
  printf("#CPU:%i\n", kaapi_default_param.cpucount);
#endif

#if defined(KAAPI_USE_PERFCOUNTER)
  /* dump output information */
  if (kaapi_default_param.display_perfcounter != KAAPI_NO_DISPLAY_PERF)
  {
    printf("[KAAPI::INIT] use #physical cpu:%u, start time:%15f\n", 
          kaapi_default_param.cpucount,kaapi_get_elapsedtime()
    );
    fflush( stdout );
  }
  if (kaapi_global_team !=0)
  {
    kaapi_context_t* ctxt = kaapi_self_context();
    if (ctxt->proc->perf !=0)
    {
      KAAPI_EVENT_PUSH1(ctxt->proc->perf, ctxt, KAAPI_EVT_TASKMAIN_BEG, 0 );
      kaapi_perf_thread_stopswapstart(ctxt->proc->perf, KAAPI_PERF_USER_STATE );
    }
  }
#endif
  

#if defined(KAAPI_DEBUG)
#if 0
  /* set signal handler print callstack */
  {
    struct sigaction sa;
    sa.sa_handler = (void (*)(int))_kaapi_signal_dump_backtrace;
    sa.sa_flags = SA_RESTART;
    sigemptyset (&sa.sa_mask);
    sigaction(SIGABRT, &sa, NULL);
    sigaction(SIGBUS,  &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGSEGV, &sa, NULL);
    sigaction(SIGFPE, &sa, NULL);
    sigaction(SIGILL, &sa, NULL);
  }
#endif
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kaapi_default_param.trace_recorder !=0)
  {
    /* set signal handler to flush event */
    struct sigaction sa;
    sa.sa_handler = _kaapi_signal_dump_counters;
    sa.sa_flags = SA_RESTART;
    sigemptyset (&sa.sa_mask);
    sigaction(SIGINT,  &sa, NULL);
    sigaction(SIGQUIT, &sa, NULL);
    sigaction(SIGTSTP, &sa, NULL);
    /* other signals: SIGABRT, SIGBUS, SIGTERM, SIGSEGV are handled by backtrace_sighandler */
  }
#endif
  
#elif defined(KAAPI_USE_PERFCOUNTER)
  if (kaapi_default_param.trace_recorder !=0)
  {
    /* set signal handler to flush event */
    struct sigaction sa;
    sa.sa_handler = _kaapi_signal_dump_counters;
    sa.sa_flags = SA_RESTART;
    sigemptyset (&sa.sa_mask);
    sigaction(SIGINT,  &sa, NULL);
    sigaction(SIGQUIT, &sa, NULL);
    sigaction(SIGABRT, &sa, NULL);
    sigaction(SIGBUS,  &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGTSTP, &sa, NULL);
  }
#endif
  
  /* */
#if defined(KAAPI_DEBUG)
  /* set alarm */
  if (kaapi_default_param.alarmperiod >0)
  {
    /* set signal handler on SIGALRM + set alarm */
    struct sigaction sa;
    sa.sa_handler = _kaapi_signal_dump_state;
    sa.sa_flags = SA_RESTART;
    sigemptyset (&sa.sa_mask);
    sigaction(SIGALRM , &sa, NULL);

    /* set periodic alarm */
    alarm( kaapi_default_param.alarmperiod );
  }
#endif
  
  return 0;
}


/**
*/
int kaapi_mt_finalize(void)
{
  int i;
  kaapi_processor_t* kproc = 0;

  if (KAAPI_ATOMIC_DECR(&kaapi_count_init) !=0) 
    return EAGAIN;
  
  if (kaapi_global_team !=0)
  {
    kaapi_assert(kaapi_global_team->count >0);

    kproc = kaapi_self_processor();
    kaapi_assert( kaapi_global_team->all_kprocessors[0] == kproc);

#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
    {
      kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_SCHEDULE_STATE );
      KAAPI_EVENT_KPROC_PUSH0( kproc, kproc->context, KAAPI_EVT_TASKMAIN_END );
    }
#endif

    kaapi_sched_sync( kaapi_context2thread(kproc->context) );
  }

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kaapi_default_param.display_perfcounter != KAAPI_NO_DISPLAY_PERF)
  {  
    printf("[KAAPI::TERM] end time:%15f, delta: %15f(s)\n", kaapi_get_elapsedtime(), 
          (double)(kaapi_get_elapsedns()-kaapi_default_param.startuptime)*1e-9 );
    fflush( stdout );
  }
#endif
  
  if (kaapi_global_team !=0)
  {
#if defined(KAAPI_USE_PERFCOUNTER)
    KAAPI_EVENT_KPROC_PUSH0(kproc, 0, KAAPI_EVT_KPROC_STOP);
#if defined(KAAPI_USE_CUDA)
    if (kaapi_default_param.trace_recorder !=0)
        kaapi_cuda_trace_finalize();
#endif /* KAAPI_USE_CUDA */
#endif

    /* implicit barrier in finalize */
    kaapi_team_finalize( kaapi_global_team );

#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
      kaapi_perf_thread_fini(kproc->perf);
    kaapi_perf_global_fini();
#endif

    /* force destruction */
    kaapi_team_destroy( kaapi_global_team, 1 );

    if (kaapi_global_team == kaapi_cached_global_team)
      kaapi_cached_global_team = 0;
    kaapi_global_team = 0;
  }
  if (kaapi_cached_global_team !=0)
  {
    kaapi_team_destroy( kaapi_cached_global_team, 1 );
    kaapi_cached_global_team = 0;
  }


  /* init numa allocation functions */
  kaapi_numa_finalize();

  /* reclaim format' names allocations */
  kaapi_format_finalize();

  /* reclaim hwdetect allocation */
  kaapi_affinityset_t* affset = &kaapi_default_param.affinity_set[0];
  for (i=0; i<KAAPI_HWS_LEVELID_MAX; ++i)
    free(affset[i].nis);

  free(kaapi_default_param.places.first);
  kaapi_default_param.places.first = 0;
  kaapi_default_param.places.len = 0;
  KAAPI_ATOMIC_WRITE(&is_hwdetect_called,0);

  return 0;
}
