/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** fabien.lementec@gmail.com 
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "kaapi_impl.h"
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <unistd.h>

#if defined(KAAPI_USE_NUMA)
#  include <numaif.h>
#  include <numa.h>
#  ifndef _GNU_SOURCE
#    define _GNU_SOURCE
#  endif
#    include <sched.h>
#endif

#if defined(KAAPI_DEBUG)
#include "kaapi_dbg.h"
#include "kaapi_util.h" /* unparse_places */
#endif


#if defined(__MIC__) || defined(__linux__)
#ifndef _GNU_SOURCE
#  define _GNU_SOURCE         /* See feature_test_macros(7) ??? */
#endif
#include <sched.h>
#include <unistd.h>
#include <stdio.h>
#endif

/* Fwd decl
*/
static int kaapi_team_addthread( kaapi_team_t* team, kaapi_procinfo_t* kpi );
static int kaapi_team_creates_threads(
  kaapi_team_t* team, kaapi_procinfo_t* kpi_start, kaapi_procinfo_t* kpi_end
);

static kaapi_atomic_t kaapi_running_threads = {1};


/*
*/
int kaapi_get_running_threads(void)
{
  return KAAPI_ATOMIC_READ(&kaapi_running_threads);
}


/*
*/
static inline void _kaapi_init_ld(
    kaapi_localitydomain_t* ld,
    kaapi_ldid_t lid,
    int queue_capacity,
    int cpu_capacity )
{
  ld->lid = lid;
  ld->count = 0;
  ld->capacity = cpu_capacity;
  ld->kids = (kaapi_processor_t**)malloc(cpu_capacity*sizeof(kaapi_processor_t*));
  memset(ld->kids, 0, cpu_capacity*sizeof(kaapi_processor_t*) );
  kaapi_atomic_initlock( &ld->lock_r );
  kaapi_atomic_initlock( &ld->lock_up );
  kaapi_queue_init( &ld->queue );
  kaapi_queue_init( &ld->private_queue );
}


static inline void _kaapi_destroy_ld( kaapi_localitydomain_t* ld )
{
  kaapi_atomic_destroylock( &ld->lock_r );
  kaapi_atomic_destroylock( &ld->lock_up );
  kaapi_queue_destroy( &ld->queue );
  kaapi_queue_destroy( &ld->private_queue );
}


/*
*/
static void _kaapi_processor_register_hierarchy( kaapi_processor_t* kproc )
{
  kaapi_team_t* team = kproc->team;
  if (team ==0) 
    return;
  if ((team->flag & KAAPI_TEAM_LDSET) ==0)
    return;
#if defined(KAAPI_USE_NUMA)
  kproc->lds[KAAPI_HWS_LEVELID_MACHINE] = team->localitydomains[KAAPI_HWS_LEVELID_MACHINE].lds[0];
  kaapi_numaid_t numanode = kproc->numanode;

  kaapi_atomic_lock( &kaapi_default_param.affinity_set[KAAPI_HWS_LEVELID_NUMA].nis[numanode].lock );
  kaapi_localitydomain_t* ld = team->localitydomains[KAAPI_HWS_LEVELID_NUMA].lds[numanode];
  if (ld ==0)
  {
    int cpucount;
#if defined(KAAPI_USE_SCHED_AFFINITY)
#if defined(KAAPI_DEBUG)
    if (kproc->place_part.first !=0)
    {
      cpu_set_t cpuset;
      CPU_ZERO( &cpuset );
      CPU_AND( &cpuset, &kaapi_default_param.affinity_set[KAAPI_HWS_LEVELID_NUMA].nis[numanode].cpuset,
                      &kproc->place_part.first[kproc->place_off] );
      kaapi_assert( CPU_COUNT(&cpuset) >0 );
    }
#endif
    cpucount = CPU_COUNT( &kaapi_default_param.affinity_set[KAAPI_HWS_LEVELID_NUMA].nis[numanode].cpuset );
#else
    cpucount = kaapi_default_param.affinity_set[KAAPI_HWS_LEVELID_NUMA].count;
#endif
    ld = (kaapi_localitydomain_t*)malloc(sizeof(kaapi_localitydomain_t));
    _kaapi_init_ld( ld, numanode, 8192, cpucount );
    kaapi_assert_debug(ld->count ==0);
    team->localitydomains[KAAPI_HWS_LEVELID_NUMA].lds[numanode] = ld;
    __sync_or_and_fetch( &team->localitydomains[KAAPI_HWS_LEVELID_NUMA].bitmap, (1<<numanode));
  }
  kproc->lds[KAAPI_HWS_LEVELID_NUMA] = ld;
  kaapi_assert_debug(ld->count >=0);
  int pos = ld->count++;
#if defined(KAAPI_USE_SCHED_AFFINITY)
  if (pos >= ld->capacity)
  {
    ld->capacity *= 2;
    kaapi_assert_debug(pos < ld->capacity);
    ld->kids = (kaapi_processor_t**)realloc(ld->kids, ld->capacity*sizeof(kaapi_processor_t*));
  }
#endif
  ld->kids[pos] = kproc;
  kaapi_atomic_unlock( &kaapi_default_param.affinity_set[KAAPI_HWS_LEVELID_NUMA].nis[numanode].lock );
#endif
}


/* never do menset on the kproc. sp_data is initialized on allocation.
   Assume thread is correctly attach to its cpuset.
*/
static int kaapi_processor_init(
    kaapi_processor_t*   kproc,
    kaapi_team_t*        team,
    kaapi_procinfo_t*    kpi
)
{
  kaapi_context_t* ctxt;
  kaapi_processor_id_t kid = kpi->proc_index;
  kproc->context       = 0;  
  kproc->kid           = kid;
#if defined(KAAPI_USE_SCHED_AFFINITY) && defined(KAAPI_USE_NUMA)
  int rcpu = sched_getcpu();
  if (kproc->place_part.first !=0)
  {
    kaapi_place_t* place = kpi->place_part.first + kpi->place_off;
    kaapi_assert( CPU_ISSET( rcpu, place ) );
  }
  kproc->numanode      = numa_node_of_cpu(rcpu);
#else
  kproc->numanode      = 0;
#endif
  kproc->team          = team;

  kaapi_assert( 0 == kaapi_default_param.emitsteal_initctxt(kproc) );
  kproc->seed = (unsigned int)kproc->kid;
  kproc->fnc_selecarg[0] = 0;
  kproc->fnc_selecarg[1] = 0;
  kproc->fnc_selecarg[2] = 0;
  kproc->fnc_selecarg[3] = 0;
  kproc->proc_type       = kpi->proc_type;
  kproc->place_part      = kpi->place_part;
  kproc->place_off       = kpi->place_off;

  kaapi_atomic_initlock( &kproc->lock );
//  KAAPI_ATOMIC_WRITE(&kproc->stealversion, 0);
  kproc->hasnowork       = 1;

  kaapi_atomic_initlock( &kproc->lock_lsuspend );
  kaapi_list_init( &kproc->lsuspend );

  /* Currently to not share these freelist among other ressources */
  kaapi_nodeinfo_t* ni = kaapi_processor_alloc_data(kproc, sizeof(kaapi_nodeinfo_t));
  kaapi_atomic_initlock( &ni->lock );
  ni->lid = (kaapi_ldid_t)kid;
  ni->freebloc = 0;
  kaapi_lfree_clear( &ni->lfree );

  kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD] = ni;
  /* to do other level ? take system detected levels */

  /* ld for thread level */
  kaapi_localitydomain_t* ld = (kaapi_localitydomain_t*)kaapi_processor_alloc_data(kproc, sizeof(kaapi_localitydomain_t));
  _kaapi_init_ld( ld, (unsigned int)kproc->kid, 8192, 1 );
  ld->count = 1;
  ld->kids[0] = kproc;
  kproc->lds[KAAPI_HWS_LEVELID_THREAD] = ld;
  /* case of orphan task in OpenMP */
  if (team !=0)
    team->localitydomains[KAAPI_HWS_LEVELID_THREAD].lds[kid] = ld;

  /* register processor NUMA level */
  _kaapi_processor_register_hierarchy( kproc );

#if defined(KAAPI_USE_PERFCOUNTER)
  /* prepare to output information */
  if ((kaapi_default_param.display_perfcounter != KAAPI_NO_DISPLAY_PERF) ||
      (kaapi_default_param.trace_recorder == 1))
  {
    kproc->perf = kaapi_processor_alloc_data(kproc, sizeof(kaapi_perfproc_t));
    memset( kproc->perf, 0, sizeof( kaapi_perfproc_t) );
    kproc->perf->start_t[0] = kproc->perf->start_t[1] = -1;
    kproc->perf->curr_perf_regs = kproc->perf->perf_regs[KAAPI_PERF_SCHEDULE_STATE];
    kproc->perf->start_t[KAAPI_PERF_SCHEDULE_STATE] = kaapi_get_elapsedns();
    /* */
    kproc->perf->serial          = 0;
    kproc->perf->eventbuffer     = 0;
  }
#endif

  ctxt = kaapi_context_alloc( kproc );
  kaapi_assert(ctxt !=0);

  /* set new context to the kprocessor */
  kaapi_setcontext(kproc, ctxt);
  
  memset(&kproc->data_specific, 0, KAAPI_KPROC_SPECIFIC_SIZE*sizeof(void*));
  memset(&kproc->size_specific, 0, KAAPI_KPROC_SPECIFIC_SIZE*sizeof(size_t));

  return 0;
}


/*
*/
static int kaapi_processor_destroy(kaapi_processor_t* kproc)
{
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    kaapi_mt_perf_thread_fini( kproc->perf );
  }
#endif

  kaapi_assert_debug( kaapi_list_empty(&kproc->lsuspend) );

  if (kproc->context !=0)
  {
    kaapi_context_t* tmp = kproc->context;
    kaapi_setcontext( kproc, 0);
    kaapi_context_free( kproc, tmp );
  }

  kaapi_context_t* ctxt;
  while ((ctxt = kaapi_lfree_pop(&kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lfree)) !=0)
    kaapi_context_free(kproc, ctxt);

  kaapi_atomic_lock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock );
  while (kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc !=0)
  {
    kaapi_stack_bloc_t* bloc = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
    kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = bloc->next;
#if defined(KAAPI_USE_NUMA)
    numa_free( bloc, kaapi_default_param.stackblocsize );
#else
    free( bloc );
#endif
  }
  kaapi_atomic_unlock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock );

  kaapi_atomic_destroylock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock );
  kaapi_atomic_destroylock( &kproc->lock );
  kaapi_atomic_destroylock( &kproc->lock_lsuspend );

  for (int i=0; i<KAAPI_KPROC_SPECIFIC_SIZE; ++i)
  {
    if (kproc->data_specific[i] !=0)
    {
      free(kproc->data_specific[i]);
    }
    KAAPI_DEBUG_INST(
      kproc->data_specific[i] = 0;
      kproc->size_specific[i] = 0;
    )
  }
  
  kaapi_assert(0 == kaapi_default_param.emitsteal_dstorctxt(kproc) );
  return 0;
}



static inline kaapi_processor_t* kaapi_processor_allocate(void)
{
  size_t size = getpagesize();
  kaapi_assert_debug( sizeof(kaapi_processor_t) < size )
  kaapi_processor_t* kproc;
  kproc = (kaapi_processor_t*) mmap(
    0, size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON, -1, (off_t)0
  );

  if (kproc == (kaapi_processor_t*)-1)
    return 0;
  
  kproc->sp_data = &kproc->stack;
  
//printf("%i: allocate processor: %p\n", (int)pthread_self(), kproc );
  return kproc;
}

static inline void kaapi_processor_free(kaapi_processor_t* kproc)
{
  size_t size = getpagesize();
  munmap(kproc, size);

//printf("%i: deallocate processor: %p\n", (int)pthread_self(), kproc );
}

kaapi_atomic_t count_teamid = {0};



/*
*/
static inline void _kaapi_team_init_ds( kaapi_team_t* team, int count )
{
  team->isterm           = 0;
  team->fnc_select       = kaapi_default_param.wsselect;
  team->fnc_localsteal   = kaapi_default_param.wslocalsteal;
  team->emitsteal        = kaapi_default_param.emitsteal;
  team->reg0             = 0;
  team->barrier.term     = 0;

  //printf("%f::%i \n\nInit TEAM !!! @:%p, term:%i, count:%i\n", kaapi_get_elapsedtime(), kaapi_self_kid(), (void*)&team->barrier, team->barrier.term, count); fflush(stdout);
  team->ws.gwork         = 0;
  KAAPI_ATOMIC_WRITE( &team->ws.single, 0 );
  KAAPI_ATOMIC_WRITE( &team->ws.forcst, 0 );
  KAAPI_ATOMIC_WRITE( &team->ws.ordered_index, 0 );
}


/* Static kaapi_cached_global_team team, the only that is cached
*/
kaapi_team_t* kaapi_cached_global_team = 0;


/*
*/
kaapi_team_t* kaapi_team_init(
  kaapi_processor_t*         kproc,
  int                        count,
  kaapi_procbind_t           procbind,
  const kaapi_places_part_t* places,
  int flag
)
{
  int i;
  kaapi_team_t* team;
  kaapi_thread_t* thread;
  kaapi_thread_register_t reginit0;

  //printf("[%i] In team_init requested cpu:%i, #places:%i\n", (int)kaapi_self_kid(), count, (places ==0 ? -1 : places->len) );
  if (kproc !=0)
  {
    thread = kaapi_context2thread(kproc->context);
    kaapi_thread_save( thread, &reginit0 );
  }
  else
    thread = 0;

  if (count == 0)
    count = kaapi_default_param.cpucount;
  if (count <=0)
  {
    errno = EINVAL;
    return 0;
  }

  /* cache only the main team if KAAPI_TEAM_NOCACHE not defined */
  if (   ((flag & KAAPI_TEAM_NOCACHE)==0)
      && (kaapi_cached_global_team !=0)
      && (kaapi_cached_global_team->count == count)
      && ((kproc ==0) || (kproc->team==0))
  )
  {
    team = kaapi_cached_global_team;
    /* reset some team fields even if reused */
    _kaapi_team_init_ds(team, count);
  }
  else
  {
    int do_dynamic;
    char* data;

    /* if can be cached, then do dynamic allocation */
    if ((thread !=0) && !(((kaapi_cached_global_team ==0) && ((flag & KAAPI_TEAM_NOCACHE)==0))))
    {
      team = (kaapi_team_t*)kaapi_data_push_align( thread, sizeof(kaapi_team_t), KAAPI_CACHE_LINE_SIZE);
      team->tid            = (kaapi_thread_id_t)kproc->kid;
      team->parent_team    = kproc->team;
      do_dynamic = 0;
    }
    else
    {
      do_dynamic = 1;
      if (0 != posix_memalign( (void**)&team, KAAPI_CACHE_LINE_SIZE, sizeof(kaapi_team_t)) )
        return 0;
#if defined(KAAPI_DEBUG)
      memset(team, 0, sizeof(kaapi_team_t) );
#endif
      team->tid            = 0;
      team->parent_team    = 0;
    }

    /* clear team fields */
    _kaapi_team_init_ds(team, count);
    KAAPI_ATOMIC_WRITE(&team->started,0);
    team->thread_started   = 0;
    team->count            = 0;
    kaapi_atomic_initlock(&team->lock );
    kaapi_atomic_initlock(&team->ws.lock );
    kaapi_barrier_init(&team->barrier, count);

    unsigned int size_data = count* (sizeof(kaapi_processor_t*)+sizeof(kaapi_procinfo_t)
#if defined(KAAPI_USE_PERFCOUNTER)
    + sizeof(kaapi_threadperf_counter_t)
#endif
    );

    if (flag & KAAPI_TEAM_LDSET)
    { /* machine level (all threads=1 domain) + numa level. Other levels are 0.
      */
      size_data += sizeof(kaapi_localitydomain_t) /* machine level locality domain */
                 + sizeof(kaapi_localitydomain_t*) * /* machine level + numa level + thread level*/
                   (1 + kaapi_default_param.affinity_set[KAAPI_HWS_LEVELID_NUMA].count + count);
    }

    if (!do_dynamic)
      data = kaapi_data_push_align( thread, size_data, 8);
    else {
      if (0 != posix_memalign( (void**)&data, KAAPI_CACHE_LINE_SIZE, size_data) )
        data = 0; // error
    }
    if (data ==0)
    {
      errno = ENOMEM;
      return 0;
    }
    team->all_kprocessors = (kaapi_processor_t**)data;
    data += count*sizeof(kaapi_processor_t*);

    team->all_descriptors = (kaapi_procinfo_t*)data;
    data += count*sizeof(kaapi_procinfo_t);

#if defined(KAAPI_USE_PERFCOUNTER)
    team->perf_regs = (kaapi_threadperf_counter_t*)data;
    memset( team->perf_regs, 0, count*sizeof(kaapi_threadperf_counter_t) );
    data += count*sizeof(kaapi_threadperf_counter_t);
#endif

#if defined(KAAPI_USE_SCHED_AFFINITY)
    kaapi_assert_debug( (places ==0) || (kproc ==0) || (kproc->place_part.len == 0) || (kproc->place_part.len == places->len) );
    kaapi_assert_debug( (places ==0) || (kproc ==0) || (kproc->place_part.len ==0) ||
      ((kproc->place_part.first == places->first) && (kproc->place_off < places->len)) );
#endif
    /* compute team hierarchy */
    if (flag & KAAPI_TEAM_LDSET)
    {
      team->flag = KAAPI_TEAM_LDSET;
      for (int i=0; i<KAAPI_HWS_LEVELID_MAX; ++i)
      {
        team->localitydomains[i].bitmap = 0;
        team->localitydomains[i].count  = 0;
        team->localitydomains[i].lds    = 0;
      }
      kaapi_localitydomainset_t* ldset = &team->localitydomains[KAAPI_HWS_LEVELID_MACHINE];
      ldset->count = 1;
      ldset->lds = (kaapi_localitydomain_t**)data;
      data += sizeof(kaapi_localitydomain_t*);
      ldset->lds[0] = (kaapi_localitydomain_t*)data;
      ldset->bitmap = 0x1;
      data += sizeof(kaapi_localitydomain_t);

      _kaapi_init_ld(ldset->lds[0], 0, 8192, 0);
      /* kids should never be reallocated */
      ldset->lds[0]->count = count;
      ldset->lds[0]->kids = team->all_kprocessors;

      ldset = &team->localitydomains[KAAPI_HWS_LEVELID_NUMA];
      ldset->count = kaapi_default_param.affinity_set[KAAPI_HWS_LEVELID_NUMA].count;
      ldset->lds = (kaapi_localitydomain_t**)data;
      data += ldset->count * sizeof(kaapi_localitydomain_t*);
      for (int i=0; i<ldset->count; ++i)
        ldset->lds[i] = 0; /* allocated by each thread on its NUMA node */

      /* ldset->bitmap is computed once numa level is created by threads */

      ldset = &team->localitydomains[KAAPI_HWS_LEVELID_THREAD];
      ldset->count = count;
      ldset->lds = (kaapi_localitydomain_t**)data;
      data += count * sizeof(kaapi_localitydomain_t*);
      for (int i=0; i<ldset->count; ++i)
        ldset->lds[i] = 0; /* allocated and set by each kproc == kproc->lds[0] */
      /* ldset->bitmap is computed once numa level is created by threads */
    }
    else
      team->flag = 0;

    /* compute the team partition */
    long P = places == 0 ? 0 : places->len;
    long T = count;
    long s,rp,k,t,rt;
    long omaster; /* thread/master offsets place in places */

    if (procbind == KAAPI_PROCBIND_DEFAULT)
      procbind = kaapi_default_param.use_affinity ? KAAPI_PROCBIND_SPREAD : KAAPI_PROCBIND_FALSE;

    if (procbind != KAAPI_PROCBIND_FALSE)
    {
      omaster = (kproc ==0 ? 0 : kproc->context->place_off);
      if (procbind == KAAPI_PROCBIND_TRUE)
        procbind = KAAPI_PROCBIND_SPREAD;
      switch (procbind)
      {
        case KAAPI_PROCBIND_TRUE:
        case KAAPI_PROCBIND_MASTER:
          break;
        case KAAPI_PROCBIND_CLOSE:
          break;
        case KAAPI_PROCBIND_SPREAD:
          k = (T<=P ? T : P);   /* k = number of partitions */
          s = P/k;              /* s = number of places per partition or s+1 */
          rp = P-s*k;           /* r = number of partitions with s+1 places */
          t = T/k;              /* t = number of threads per partition */
          rt = T-k*t;           /* t = number of partitions with t+1 threads */
          break;
        default:
          kaapi_assert(0);
      }
    }

    /* It should be the only location where request are initialized */
    memset( (void*)&team->all_kprocessors[0], 0, sizeof(team->all_kprocessors[0])*count);

    for (i = 0; i<count; ++i)
    {
      team->all_descriptors[i].team                = team;
      team->all_descriptors[i].proc_index          = i;
      switch (procbind)
      {
        case KAAPI_PROCBIND_FALSE:
          team->all_descriptors[i].place_off        = 0;
          team->all_descriptors[i].place_part.first = 0;
          team->all_descriptors[i].place_part.len   = 0;
          break;

        case KAAPI_PROCBIND_TRUE:
        case KAAPI_PROCBIND_MASTER:
          team->all_descriptors[i].place_part= *places;
          team->all_descriptors[i].place_off = omaster;
          break;
        case KAAPI_PROCBIND_CLOSE:
          team->all_descriptors[i].place_part = *places;
          team->all_descriptors[i].place_off  = (omaster+i) % P;
          break;
        case KAAPI_PROCBIND_SPREAD:
        {
          long place_first, part_len;
          long part =(i < rt*(t+1)) ? i / (t+1) : rt+(i-rt*(t+1))/t;
          if (part < rp )
          {
            place_first = part*(s+1);
            part_len = s+1;
          }
          else
          {
            place_first =(part-rp)*s + rp*(s+1);
            part_len = s;
          }
          team->all_descriptors[i].place_part.first  = places->first+(place_first+omaster) % P;
          team->all_descriptors[i].place_part.len    = part_len;
          team->all_descriptors[i].place_off         = 0;
        } break;
        default:
          kaapi_assert(0);
      }
      team->all_descriptors[i].proc_type           = KAAPI_PROC_TYPE_HOST;
      team->all_descriptors[i].arg_runprocessor[0] = 0;
      team->all_descriptors[i].arg_runprocessor[1] = 0;
    }

    /* default processor number */
    team->count = count;
    if ((kaapi_cached_global_team ==0) && ((flag & KAAPI_TEAM_NOCACHE)==0))
      kaapi_cached_global_team = team;
  }

  team->save_master = kproc;
  if (thread !=0)
  {
    team->reg0           = kaapi_thread_pushdata( thread, sizeof(kaapi_thread_register_t) );
    *team->reg0          = reginit0;
  }
  team->teamid = KAAPI_ATOMIC_INCR(&count_teamid);

  return team;
}


/*
*/
kaapi_thread_t* kaapi_team_finalize( kaapi_team_t* team )
{
  kaapi_assert_debug( kaapi_self_processor() == team->all_kprocessors[0] );
  kaapi_processor_t* kproc0 = team->all_kprocessors[0];
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc0->perf !=0)
    kaapi_perf_thread_stopswapstart(kproc0->perf, KAAPI_PERF_SCHEDULE_STATE );

  /* new barrier at the end of parallel region to schedule tasks and returns */
  kaapi_team_barrier_wait( team, kproc0, KAAPI_BARRIER_FLAG_DEFAULT );

  if (kproc0->perf !=0)
  {
    kaapi_perf_thread_stopswapstart(kproc0->perf, KAAPI_PERF_USER_STATE );
    kaapi_perf_thread_stop(kproc0->perf);
    KAAPI_EVENT_PUSH0(kproc0->perf, 0, KAAPI_EVT_KPROC_STOP );

    /* recopy in the team data structure */
    for (int s=KAAPI_PERF_USER_STATE; s<= KAAPI_PERF_SCHEDULE_STATE; ++s)
      for (int id=0; id<KAAPI_PERF_ID_MAX; ++id)
      {
        team->perf_regs[0][s][id] += kproc0->perf->perf_regs[s][id];
        kproc0->perf->perf_regs[s][id] = 0;
      }
  }
  kaapi_collect_trace(team);
#endif

  /* normal implicit barrier at the end of parallel region */
  kaapi_team_barrier_wait( team, kproc0, KAAPI_BARRIER_FLAG_DEFAULT );

#if 0
{
  int i;
  uintptr_t p = team->all_kprocessors[0]->fnc_selecarg[1];
  uintptr_t u = team->all_kprocessors[0]->fnc_selecarg[1];
  printf("0 : %lu\n", p);
  for (i=1; i<team->count; ++i)
  {
    p &= team->all_kprocessors[i]->fnc_selecarg[1];
    u |= team->all_kprocessors[i]->fnc_selecarg[1];
    printf("%i :: intersec: %lu  union: %lu\n", i, p, u);
  }
}
#endif

  KAAPI_ATOMIC_SUB(&kaapi_running_threads, team->count-1);

  kaapi_assert_debug( kaapi_self_kid() == 0 );
  kaapi_context_clear( team->all_kprocessors[0]->context );
  kaapi_set_self_processor( team->save_master );
  if (team->save_master ==0)
    return 0;

  kaapi_assert_debug( team->reg0 != 0);
  kaapi_thread_restore( kaapi_context2thread(team->save_master->context), team->reg0 );

  return kaapi_context2thread(team->save_master->context);
}


/*
*/
int kaapi_team_destroy( kaapi_team_t* team, int force)
{
  /* only destroy non initial team */
  if ((team != kaapi_cached_global_team) || force)
  {
    team->isterm = -1;
    kaapi_writemem_barrier();
    kaapi_team_barrier_wait_signal( team, team->all_kprocessors[0] );
    kaapi_team_barrier_wait_term( team, team->all_kprocessors[0] );

    for (int i=0; i<team->count; ++i)
    {
      kaapi_processor_destroy(team->all_kprocessors[i]);
      kaapi_processor_free(team->all_kprocessors[i]);
    }

    /* */
    kaapi_barrier_destroy( &team->barrier);
    kaapi_atomic_destroylock(&team->lock );
    kaapi_atomic_destroylock(&team->ws.lock );
    kaapi_localitydomainset_t* ldset = &team->localitydomains[KAAPI_HWS_LEVELID_MACHINE];
    _kaapi_destroy_ld( ldset->lds[0] );
    ldset->lds[0] = 0;

    KAAPI_DEBUG_INST(
      team->count          = 0;
      team->fnc_select     = 0;
      team->fnc_localsteal = 0;
      team->emitsteal      = 0;
    );

    if (team->save_master == 0)
    {
      free( team->all_kprocessors );
      KAAPI_DEBUG_INST(team->all_kprocessors = 0;)
      free( team );
    }
    else if (team != kaapi_cached_global_team) 
    {
      kaapi_setcontext( team->save_master, team->save_master->context );
      kaapi_set_self_processor( team->save_master );
    }
  }
  return 0;
}


/* Lock of a team
   if schedule!=0 then schedule tasks among the team
*/
int kaapi_team_lock (struct kaapi_team_t *team, int schedule)
{
  if (team ==0) return 0;
  if (team->count ==0) return 0;
  return kaapi_atomic_lock( &team->ws.lock);
}

/* Unlock of a team
   if schedule!=0 then schedule tasks among the team
*/
int kaapi_team_unlock (struct kaapi_team_t *team, int schedule)
{
  if (team ==0) return 0;
  if (team->count ==0) return 0;
  return kaapi_atomic_unlock( &team->ws.lock);
}


/* Similar to omp single nowait.
*/
int kaapi_team_single ( kaapi_team_t* team, kaapi_processor_t* kproc )
{
  if (team ==0) return 1;
  if (kproc == 0) kproc = kaapi_self_processor();
  kaapi_context_t *ctxt = kproc->context;
  int oldvalue = ctxt->ws.count_single;
  ++ctxt->ws.count_single;
  int retval = KAAPI_ATOMIC_CAS(&team->ws.single, oldvalue, ctxt->ws.count_single);
  return retval;
}


/* Similar to omp master nowait.
*/
int kaapi_team_master ( kaapi_team_t* team, kaapi_processor_t* kproc )
{
  if (team ==0) return 1;
  if (kproc == 0) kproc = kaapi_self_processor();
  if (kproc->kid == 0) return 1;
  return 0;
}


/*
*/
int kaapi_team_start(
  kaapi_team_t* team
)
{
  KAAPI_ATOMIC_ADD(&kaapi_running_threads, team->count-1);

  kaapi_processor_t* kproc0 = team->all_kprocessors[0];

  if ((team == kaapi_cached_global_team) && (team->thread_started !=0))
  {
    /* reset kproc 0, the mast */
    kproc0->team         = team;
    kaapi_setcontext(kproc0, kproc0->context);
    kaapi_set_self_processor(kproc0);

    kaapi_team_barrier_wait_signal( team, kproc0 );

#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc0->perf !=0)
      kaapi_perf_thread_init( kproc0->kid, kproc0->perf, KAAPI_PERF_USER_STATE);
#endif
  }
  else
  {
    int alreadystarted = KAAPI_ATOMIC_READ(&team->started);
    if (0 != kaapi_team_creates_threads(
          team,
          &team->all_descriptors[0],
          &team->all_descriptors[team->count]
      ))
        return EINVAL;

    if (1)
    {
      while (alreadystarted+team->count != KAAPI_ATOMIC_READ(&team->started))
      {
        kaapi_slowdown_cpu();
        sched_yield();
      }
    }
    team->thread_started = 1;
  }

  /* master thread can be stolen */
  if (kproc0) kproc0->hasnowork = 0;

#if defined(KAAPI_USE_PERFCOUNTER)
  if (team->all_kprocessors[0]->perf)
  {
    kaapi_processor_t* kproc = team->all_kprocessors[0];
    KAAPI_EVENT_PUSH1(kproc->perf, 0, KAAPI_EVT_KPROC_START, kproc->proc_type );
    kaapi_perf_thread_start(kproc->perf);
  }
#endif

  return 0;
}


typedef struct {
  kaapi_team_t* team;
  kaapi_procinfo_t* kpi_start;
  kaapi_procinfo_t* kpi_end;
} kaapi_thread_create_range_t;

/*
*/
static void* helper_kaapi_team_creates_threads( void* arg )
{
  kaapi_thread_create_range_t* range = (kaapi_thread_create_range_t*)arg;
  kaapi_team_creates_threads( range->team, range->kpi_start, range->kpi_end );
  free(range);
  return 0;
}

/*
*/
static int kaapi_team_creates_threads(
    kaapi_team_t* team,
    kaapi_procinfo_t* kpi_start,
    kaapi_procinfo_t* kpi_end   /* past the last */
)
{
  size_t count = kpi_end - kpi_start;
  if (count <= 24)
  {
    int err;
    for (; kpi_start < kpi_end; ++kpi_start)
    {
      if (0 != (err = kaapi_team_addthread(team, kpi_start)))
        return err;
    }
  }
  else
  {
    pthread_t tid;
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED);

    kaapi_thread_create_range_t* arg = malloc(sizeof(kaapi_thread_create_range_t));
    arg->team = team;
    arg->kpi_start = kpi_start+count/2;
    arg->kpi_end = kpi_end;
    kpi_end = arg->kpi_start;

    if (0 != pthread_create(&tid, &attr, helper_kaapi_team_creates_threads, arg))
      return EAGAIN;
    kaapi_team_creates_threads( team, kpi_start, kpi_end );
  }

  return 0;
}


/* Creates thread on the specified cpuset.
   Allocate the stack (C stack) of the thread on the right numa node.
*/
static void* _kaapi_team_addthread( kaapi_procinfo_t* kpi )
{
  pthread_t tid;
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED);
  size_t stacksize;
  void* stackaddr;

  /* reschedule the thread on the local cpu */
  sched_yield();

  if (kaapi_default_param.threadstacksize !=0)
    stacksize = kaapi_default_param.threadstacksize;
  else
    kaapi_assert( 0 == pthread_attr_getstacksize( &attr, &stacksize));

  /* roundup to pagesize */
  stacksize = (stacksize + getpagesize()-1) & ~(getpagesize()-1);
  //printf("stack size:%lu\n", stacksize);
  stackaddr = mmap(0, stacksize,PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON, -1, (off_t)0);
  kaapi_assert(stackaddr != (void*)-1);

#if defined(KAAPI_USE_SCHED_AFFINITY) && defined(KAAPI_USE_NUMA)
  if (kpi->place_part.len >0)
  {
    size_t cnt = 0;
    kaapi_assert_debug( kpi->place_part.first !=  0);
    kaapi_assert_debug( kpi->place_off < kpi->place_part.len );
    kaapi_place_t* place = kpi->place_part.first + kpi->place_off;
    if ((cnt = CPU_COUNT(place)) != 0)
    {
      kaapi_assert((!pthread_attr_setaffinity_np(&attr, sizeof(cpu_set_t), place)));
      int r = 1+(int)rand() % cnt;
      int cpu = 0;
      for (cpu=0; cpu<sizeof(cpu_set_t)*8; ++cpu)
        if (CPU_ISSET(cpu, place))
          if (--r ==0)
            break;
      kaapi_assert_debug( cpu != 8*sizeof(cpu_set_t) );
      int node = numa_node_of_cpu(cpu);
      int maxnode = numa_num_configured_cpus();
      unsigned long nodemask
          [(maxnode + (8 * sizeof(unsigned long) -1))/ (8 * sizeof(unsigned long))];
      memset(nodemask, 0, sizeof(nodemask));
      nodemask[node / (8 * sizeof(unsigned long))] |= 1UL << (node % (8 * sizeof(unsigned long)));
      kaapi_assert (mbind((void*)stackaddr, stacksize, MPOL_BIND, nodemask, maxnode, MPOL_MF_STRICT | MPOL_MF_MOVE) == 0);
    }
  }
#endif /* KAAPI_USE_SCHED_AFFINITY */

#if defined(__linux__)
  kaapi_assert( !pthread_attr_setstack(&attr, stackaddr, stacksize));
#else
  int err;
  kaapi_assert( !(err = pthread_attr_setstacksize(&attr, stacksize)));
  kaapi_assert( !(err = pthread_attr_setstackaddr(&attr, stackaddr)));
#endif
  if (EAGAIN == pthread_create(&tid, &attr,kaapi_default_sched_run_processor, kpi))
  {
    pthread_attr_destroy(&attr);
    return 0;
  }
  pthread_attr_destroy(&attr);
  return 0;
}


/*
*/
static int kaapi_team_addthread( kaapi_team_t* team, kaapi_procinfo_t* kpi )
{
  kaapi_processor_id_t kid = kpi->proc_index;
  kpi->team = team;

  if (kid >= (uint64_t)team->count)
    return EINVAL;

  if (kid !=0)
    _kaapi_team_addthread(kpi);
  else /* if (kid == 0) */
  {
    kaapi_processor_t* kproc;

#if defined(KAAPI_USE_SCHED_AFFINITY)
    if (kpi->place_part.len >0)
    {
      kaapi_assert_debug( kpi->place_part.first !=  0);
      kaapi_assert_debug( kpi->place_off < kpi->place_part.len );
      kaapi_place_t* place = kpi->place_part.first + kpi->place_off;
      if (CPU_COUNT(place) != 0)
        kaapi_assert(!pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), place));
    }
#endif /* KAAPI_USE_SCHED_AFFINITY */

    kproc = kaapi_processor_allocate();
    if (kproc == 0)
      return ENOMEM;

    kaapi_assert(0 == kaapi_processor_init( kproc, team, kpi ) );
    kaapi_set_self_processor(kproc);
    kproc->hasnowork         = 0;
    team->all_kprocessors[0] = kproc;
    KAAPI_ATOMIC_INCR(&team->started);
#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
    {
      /*  */
      kaapi_perf_thread_init(kproc->kid, kproc->perf, KAAPI_PERF_USER_STATE);
    }
#endif
  }

  return 0;
}


/* Create orphan kproc with no team
*/
kaapi_processor_t* kaapi_create_orphan_kprocessor(void)
{
  kaapi_processor_t* kproc = kaapi_processor_allocate();
  if (kproc ==0) return 0;
  kaapi_procinfo_t kpi = {
    .place_part = kaapi_default_param.places,
    .place_off = 0,
    .proc_type = KAAPI_PROC_TYPE_CPU,
    .proc_index = 0
  };
  kaapi_assert(0 == kaapi_processor_init(kproc, 0, &kpi));

  kaapi_set_self_processor(kproc);
  return kproc;
}


/**
*/
void* kaapi_default_sched_run_processor( void* arg )
{
  kaapi_procinfo_t*  kpi = (kaapi_procinfo_t*)arg;
  kaapi_processor_t* kproc = 0;
  kaapi_processor_id_t kid;
  kaapi_team_t* team = kpi->team;
  kaapi_team_t* oldteam = 0;
  void (*body)(void*) = (void (*)(void*))kpi->arg_runprocessor[0];

  /* reschedule the thread on the local cpu */
  sched_yield();

#if defined(KAAPI_USE_SCHED_AFFINITY)
  if (kpi->place_part.len >0)
  {
    kaapi_assert_debug( kpi->place_part.first !=  0);
    kaapi_assert_debug( kpi->place_off < kpi->place_part.len );
    kaapi_place_t* place = kpi->place_part.first + kpi->place_off;
    if (CPU_COUNT(place) != 0)
      sched_setaffinity( 0, sizeof(cpu_set_t), place);
  }
#endif /* KAAPI_USE_SCHED_AFFINITY */

  /* processor initialization */
  kid = kpi->proc_index;
  team = kpi->team;
  kproc = kaapi_processor_allocate();
  kaapi_assert( 0 == kaapi_processor_init( kproc, team, kpi ));
  kaapi_assert( kproc->context != 0 );

  kaapi_set_self_processor(kproc);

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
    kaapi_perf_thread_init(kproc->kid, kproc->perf, KAAPI_PERF_SCHEDULE_STATE);
#endif

  KAAPI_ATOMIC_INCR(&team->started);

  /* infinite loop 
  */
  do {
    //sched_yield();
#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf !=0)
    {
      KAAPI_EVENT_PUSH1(kproc->perf, 0, KAAPI_EVT_KPROC_START, kpi->proc_type );
      kaapi_perf_thread_start(kproc->perf);
    }
#endif

    if (oldteam != team)
    {
      team->all_kprocessors[kid] = kproc;
      if (oldteam !=0) _kaapi_processor_register_hierarchy(kproc);
    }
    kaapi_assert( kproc->context != 0 );
    kaapi_setcontext(kproc, kproc->context);
    kaapi_set_self_processor(kproc);

    /* main work stealing loop or user code */
    if (body !=0)
    {
#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf)
      kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_USER_STATE );
#endif

      kproc->hasnowork = 0;
      body( kpi->arg_runprocessor[1] );
      kaapi_sched_sync( kaapi_context2thread(kproc->context) );

#if defined(KAAPI_USE_PERFCOUNTER)
    if (kproc->perf)
      kaapi_perf_thread_stopswapstart(kproc->perf, KAAPI_PERF_SCHEDULE_STATE );
#endif
    }
    kproc->hasnowork = 1;

#if defined(KAAPI_USE_PERFCOUNTER)
    /* new barrier at the end of parallel region to schedule tasks and returns */
    kaapi_team_barrier_wait( team, kproc, KAAPI_BARRIER_FLAG_DEFAULT );

    if (kproc->perf !=0)
    {
      /*  */
      kaapi_perf_thread_stop(kproc->perf);
      KAAPI_EVENT_PUSH0(kproc->perf, 0, KAAPI_EVT_KPROC_STOP );

      /* recopy in the team data structure */
      for (int s=KAAPI_PERF_USER_STATE; s<=KAAPI_PERF_SCHEDULE_STATE; ++s)
        for (int i=0; i<KAAPI_PERF_ID_MAX; ++i)
        {
          team->perf_regs[kid][s][i] += kproc->perf->perf_regs[s][i];
          kproc->perf->perf_regs[s][i]= 0;
        }
      kaapi_perf_thread_start(kproc->perf);
    }
#endif

    /* implicit barrier at the end of parallel region */
    kaapi_team_barrier_wait( team, kproc, KAAPI_BARRIER_FLAG_WAITEXIT );

    kaapi_context_clear( kproc->context );

  if (team->isterm == -1)
    break;

#if 0
    if (KAAPI_ATOMIC_DECR(&kaapi_running_threads) <0)
      abort();
#endif

    oldteam = team;
    kid  = kpi->proc_index;
    team = kpi->team;
    kaapi_assert_debug(kproc->team == team); /* may have changed ? */
    body = (void (*)(void*))kpi->arg_runprocessor[0];

  } while (1);

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
    kaapi_mt_perf_thread_fini( kproc->perf );
#endif

  kaapi_team_barrier_wait_signal_term(team, kproc );
  //printf("Delay wait: thread=%i, wait1=%e, waverage/call=%e\n", kid, delay_wait1, delay_wait1/count);
  
  kaapi_assert_debug( kaapi_self_processor() == kproc );

  return 0;
}


#if defined(KAAPI_USE_SCHED_AFFINITY)
static int get_cpuid(kaapi_processor_t* kproc)
{
  int cpu = -1; //kproc->cpuid;
#if defined(__MIC__) || defined(__linux__)
  if (cpu == -1)
    cpu = sched_getcpu();
#elif defined(__APPLE__)
  cpu = (int)kproc->kid;
#else
#error "Undefined OS"
#endif
  return cpu;
}


static int get_physicalproc(kaapi_processor_t* kproc)
{
  int processor;
  int cpu = get_cpuid(kproc);
#if defined(__MIC__)
  processor = cpu -1;
  if (processor <0) processor += kaapi_default_param.syscpucount;
  processor /= 4; /* for series 5100, 3100, 7100 */
#elif defined(__linux__)
  char filename[256];
  FILE* file;
  int err;
  sprintf(filename, "/sys/devices/system/cpu/cpu%i/topology/core_id", cpu);
  file = fopen(filename, "rt");
  if (file ==0)
    err = -EINVAL;
  else
    err = fscanf(file,"%d",
                 &processor
    );

  /* test if processor has been read */
  if (err < 1)
    return -1;
#elif defined(__APPLE__)
  processor = cpu;
#else
#error "Undefined OS"
#endif
  return processor;
}


/*
*/
int _kaapi_processor_getcpuid(
    kaapi_processor_t*   kproc
)
{
  exit(1);
  return -1;
}
#endif


/*
*/
int kaapi_processor_getpu( kaapi_team_t* team, kaapi_processor_t* kproc,  int n, int* kids )
{
  if (n<0) return -1;
#if defined(__MIC__)
  int* kids0 = kids;
  int cpu = get_cpuid(kproc);
  int processor = cpu -1;
  if (processor <0) processor += kaapi_default_param.syscpucount;
  processor /= 4; /* for series 5100, 3100, 7100 */
  if (n >4) n = 4; /* 4 Pus per core */
  for (int i=0; i<n; ++i)
  {
    int cpuid = processor*4+1+i;   /** with respect to the team-> core/pu mapping */
    if (cpuid >= kaapi_default_param.syscpucount)
      cpuid -= kaapi_default_param.syscpucount;
    if (*kids != -1) ++kids;
  }
  return kids-kids0;
#else
  return -1;
#endif
}

