/*
** xkaapi
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
**
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/

#include "kaapi_impl.h"


/* ========================================================================= */
/* Functions to manipulate list of requests                                  */
/* ========================================================================= */

/** This is the new implementation of the combining steal request as proposed by Fatourou, CC_sync
    algorithm. The tail is stored in the team structure.
*/
typedef struct kaapi_request_node_t {
 kaapi_request_t              req;
 struct kaapi_request_node_t* volatile next;
 int volatile                 wait;
 int volatile                 completed;
} kaapi_request_node_t;


/* structure to retreive request object */
typedef struct kaapi_listrequest_t {
  kaapi_team_t*                team;
  struct kaapi_request_node_t* tail;
  kaapi_request_node_t*        node;       /* to be reused */
  kaapi_request_node_t*        currNode;   /* to init the iterator */
} kaapi_listrequest_t __attribute__((aligned (KAAPI_CACHE_LINE)));


/* The most important structure to iterate over the list of requests
   - only serve at most numserve request before stopping aggregation
*/
typedef struct kaapi_listrequest_iterator_ccsync_t {
  struct kaapi_listrequest_iterator_t lri;
  int                                 numserve;
  kaapi_request_node_t*               curr;
  kaapi_request_node_t*               next;
} kaapi_listrequest_iterator_ccsync_t;


/*
*/
static int kaapi_request_replytask_ccsync
( 
  kaapi_request_t*        request,
  int                     value
)
{
#if defined(KAAPI_DEBUG)
  int oldstatus = KAAPI_ATOMIC_READ(&request->status);
  kaapi_assert_debug( oldstatus == KAAPI_REQUEST_S_POSTED );
#endif

  if (value == KAAPI_REQUEST_S_OK)
    KAAPI_ATOMIC_WRITE(&request->status, KAAPI_REQUEST_S_OK);
  else
    KAAPI_ATOMIC_WRITE(&request->status, KAAPI_REQUEST_S_NOK);

  struct kaapi_request_node_t* currNode = (kaapi_request_node_t*)request;
  currNode->completed = 1;
  kaapi_writemem_barrier();
  currNode->wait      = 0;
  return 0;
}



/*
*/
static void kaapi_listrequest_iterator_destroy(
  kaapi_listrequest_iterator_ccsync_t* lrrange
)
{
  lrrange->curr->wait = 0;
  kaapi_writemem_barrier();
}

static inline void kaapi_listrequest_prepare_post(
  kaapi_team_t* team,
  kaapi_listrequest_t* lrequests
)
{ 
  lrequests->team = team;
}


/* return !=0 iff the range is empty
*/
static int kaapi_listrequest_ccsync_iterator_empty(
  struct kaapi_listrequest_iterator_t* lri
)
{ 
  kaapi_listrequest_iterator_ccsync_t* lrrange = (kaapi_listrequest_iterator_ccsync_t*)lri;
  return (lrrange->numserve <=0) || (lrrange->next ==0);
}


/* get the first request of the range. range iterator should have been initialized by kaapi_listrequest_iterator_init 
*/
static kaapi_request_t* kaapi_listrequest_ccsync_iterator_get(
  struct kaapi_listrequest_iterator_t* lri
)
{ 
  kaapi_listrequest_iterator_ccsync_t* lrrange = (kaapi_listrequest_iterator_ccsync_t*)lri;
  return &lrrange->curr->req;
}


/* return the next entry in the request. return 0 if the range is empty.
*/
static kaapi_request_t* kaapi_listrequest_ccsync_iterator_next(
  struct kaapi_listrequest_iterator_t* lri
)
{
  kaapi_listrequest_iterator_ccsync_t* lrrange = (kaapi_listrequest_iterator_ccsync_t*)lri;
  --lrrange->numserve;
  kaapi_request_node_t* nextNode =lrrange->next;
  lrrange->curr = nextNode;
  lrrange->next = nextNode->next;
  if (kaapi_listrequest_iterator_empty(lri))
    return 0;
  return kaapi_listrequest_iterator_get(lri);
}


/*
*/
static int kaapi_listrequest_ccsync_iterator_count(
  struct kaapi_listrequest_iterator_t* lri
)
{
  kaapi_listrequest_iterator_ccsync_t* lrrange = (kaapi_listrequest_iterator_ccsync_t*)lri;
  kaapi_request_node_t* curr = lrrange->curr;
  int cnt = 0;
  while (curr != 0)
  {
    ++cnt;
    curr = curr->next;
  }
  return cnt-1;
}


/*
*/
static void kaapi_listrequest_iterator_init(
  kaapi_listrequest_t*   own_lrequests,
  kaapi_listrequest_t*   victim_lrequests,
  kaapi_listrequest_iterator_ccsync_t* lrrange
)
{ 
  lrrange->lri.lri_empty = kaapi_listrequest_ccsync_iterator_empty;
  lrrange->lri.lri_get   = kaapi_listrequest_ccsync_iterator_get;
  lrrange->lri.lri_next  = kaapi_listrequest_ccsync_iterator_next;
  lrrange->lri.lri_count = kaapi_listrequest_ccsync_iterator_count;
  lrrange->curr     = own_lrequests->currNode;
  if (lrrange->curr->completed == 1)
  { /* make empty range */
    lrrange->numserve = 0;
    lrrange->next     = 0;
  }
  else {
    lrrange->numserve = 65536;
    lrrange->next     = lrrange->curr->next;
  }
}



/*
*/
static int kaapi_listrequest_init( kaapi_listrequest_t* pklr, kaapi_processor_t* kproc )
{
  kaapi_request_node_t* node;
  pklr->team = 0;

  /* allocate the tail of the list */
// Else, bug at termination a thread finish and free its kproc entry, but note is referenced by
  node = kaapi_processor_alloc_data(kproc, sizeof(kaapi_request_node_t) );
  //node = malloc(sizeof(kaapi_request_node_t) );
  node->req.reply          = 0;
  node->req.ident          = 0x12345678910;
  node->req.thread.sp      = 0;
  node->req.thread.sp_data = 0;
  KAAPI_ATOMIC_WRITE(&node->req.status, KAAPI_REQUEST_S_INIT);
  node->wait               = 0;
  node->completed          = 0;
  node->next               = 0;

  pklr->tail = node;

  /* allocate the next node to use */
  node = kaapi_processor_alloc_data(kproc, sizeof(kaapi_request_node_t) );
  //node = malloc( sizeof(kaapi_request_node_t) );
  node->req.reply = 0;
  node->req.ident = -1;
  node->req.ident  = (uintptr_t)-1;
  KAAPI_ATOMIC_WRITE(&node->req.status, KAAPI_REQUEST_S_INIT);
  node->wait = 0;
  node->completed = 0;
  node->next = 0;
  pklr->currNode = 0;
  kaapi_writemem_barrier();

  pklr->node = node;

  return 0;
}


/*
*/
static void kaapi_listrequest_destroy(
  kaapi_listrequest_t* pklr
)
{
#if 0
  kaapi_request_node_t* curr = pklr->currNode;
  while (curr !=0)
  {
    kaapi_request_node_t* nextnode = curr->next;
    free( curr );
    curr = nextnode;
  }
#endif
}


/*
*/
static kaapi_request_t* kaapi_request_post(
  kaapi_listrequest_t*   own_lrequests,
  kaapi_listrequest_t*   victim_lrequests,
  kaapi_processor_id_t   selfid,
  const kaapi_thread_t*  thread
)
{
  kaapi_request_node_t* currNode;
  kaapi_request_node_t* nextNode = own_lrequests->node;
  nextNode->next = 0;
  nextNode->wait = 1;
  nextNode->completed = 0;
  nextNode->req.ident  = (uintptr_t)-1;
  KAAPI_ATOMIC_WRITE(&nextNode->req.status, KAAPI_REQUEST_S_INIT);
  currNode = nextNode;

  kaapi_assert_debug( own_lrequests->node == nextNode );
  kaapi_assert_debug( currNode == nextNode );
  
  /* on return currNode contains the old value of tail */
  KAAPI_ATOMIC_EXCHANGE(currNode, victim_lrequests->tail);
  kaapi_assert_debug( currNode != nextNode );
  
  kaapi_assert_debug( currNode->next == 0 );

  KAAPI_ATOMIC_WRITE(&currNode->req.status, KAAPI_REQUEST_S_POSTED);
  currNode->req.ident  = (uintptr_t)selfid;
  currNode->req.thread = *thread;

  kaapi_writemem_barrier();
  currNode->next = nextNode;
  own_lrequests->node = currNode; /* next node used by the thread */

  own_lrequests->currNode = currNode;
  currNode->req.reply = kaapi_request_replytask_ccsync;
  return &currNode->req;
}



#define NANO_SLEEP 1
#if NANO_SLEEP
#if !defined(_XOPEN_SOURCE)
#  define _XOPEN_SOURCE 600
#endif
#include <time.h>
#endif


typedef struct kaapi_ccsync_steal_context {
  kaapi_listrequest_t lr;   /* see listrequest.h */
} kaapi_ccsync_steal_context;

/* no concurrency here: always called before starting threads */
int kaapi_sched_ccsync_emitsteal_init(kaapi_processor_t* kproc)
{
  kaapi_ccsync_steal_context* ctxt;
  kproc->emitsteal_ctxt = ctxt = kaapi_processor_alloc_data(kproc, sizeof(kaapi_ccsync_steal_context));
  if (kproc->emitsteal_ctxt ==0) return ENOMEM;
  kaapi_listrequest_init( &ctxt->lr, kproc );
  return 0;
}

/* no concurrency here: always called before starting threads */
int kaapi_sched_ccsync_emitsteal_dstor(kaapi_processor_t* kproc)
{
  kaapi_ccsync_steal_context* ctxt = kproc->emitsteal_ctxt;
  kproc->emitsteal_ctxt =0;
  kaapi_listrequest_destroy( &ctxt->lr );
  return 0;
}


/*
*/
kaapi_request_status_t kaapi_sched_ccsync_emitsteal ( kaapi_processor_t* kproc )
{
  kaapi_team_t*                team;
  kaapi_victim_t               victim;
  kaapi_request_t*             self_request;
  kaapi_request_t*             request;
  int                          err;
  kaapi_listrequest_iterator_ccsync_t lri;

#if defined(KAAPI_USE_PERFCOUNTER)
  uintptr_t serial;
  kaapi_perf_counter_t date0 = 123124;
  kaapi_perf_counter_t date1 = 123124;
#endif

  team = kproc->team;

  if (team->count <2) 
    return KAAPI_REQUEST_S_NOK;

redo_select:
  /* select the victim processor */
  err = (*team->fnc_select)( kproc, &victim, KAAPI_SELECT_VICTIM );
  if (unlikely(err !=0)) 
    goto redo_select;

  if (victim.kproc ==0) goto redo_select;
  kaapi_assert_debug( (victim.kproc->kid >=0) && (victim.kproc->kid < (uintptr_t)team->count));

  /* quick test to detect if thread has no work */
  if (kaapi_processor_has_nowork(victim.kproc))
  {
    (*team->fnc_select)( kproc, &victim, KAAPI_STEAL_FAILED );
    return KAAPI_REQUEST_S_NOK;
  }

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
    ++kproc->perf->serial;
#endif

  /* (1) 
     Fill & Post the request to the victim processor 
  */
  kaapi_ccsync_steal_context* own_stealctxt
    = (kaapi_ccsync_steal_context*)kproc->emitsteal_ctxt;
  kaapi_ccsync_steal_context* victim_stealctxt 
    = (kaapi_ccsync_steal_context*)victim.kproc->emitsteal_ctxt;

  kaapi_listrequest_prepare_post(team, &victim_stealctxt->lr);
  self_request = kaapi_request_post(
    &own_stealctxt->lr,
    &victim_stealctxt->lr,
    kproc->kid,
    kaapi_context2thread(kproc->context)
  );

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    kaapi_perfproc_t* perf = kproc->perf;
    KAAPI_IFUSE_TRACE(perf,
      self_request->victim = (uintptr_t)victim.kproc->kid;
      serial = ++kproc->perf->serial;
      self_request->serial = serial;
      KAAPI_EVENT_PUSH2(perf, 0, KAAPI_EVT_STEAL_OP,
          (uintptr_t)victim.kproc->kid, 
          self_request->serial
      );
    );
    ++KAAPI_PERF_REG(kproc->perf, KAAPI_PERF_ID_STEALREQ);

    kaapi_assert_debug( KAAPI_PERF_SCHEDULE_STATE == KAAPI_GET_THREAD_STATE(kproc->perf) );
    date0 = kaapi_get_elapsedns();
  }
#endif

  /* (2) In case of aggregation, lock after thief has posted its request.
     lock and re-test if they are yet posted requests on victim or not 
     if during tentaive of locking, a reply occurs, then return with reply
  */
  while ( ((kaapi_request_node_t*)self_request)->wait == 1)
  {
#  if defined(KAAPI_USE_NETWORK)
    kaapi_network_poll();
#  endif
#if defined(NANOSLEEP)
    struct timespec ts = { 0, 128 }; 
    nanosleep( &ts, 0 ); 
#else
    kaapi_slowdown_cpu();
#endif
  }

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    date1 = kaapi_get_elapsedns();
    KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_TSYNC) += date1-date0;
    KAAPI_EVENT_PUSH2(kproc->perf, 0, KAAPI_EVT_REQUESTS_BEG,
                      (uintptr_t)victim.kproc->kid, serial );
  }
#endif

  if (kaapi_request_test( self_request ))
    return kaapi_sched_emitsteal_return( kproc, &victim, self_request );

  /* here becomes an aggregator... the trylock has synchronized memory */
  kaapi_listrequest_iterator_init(
          &own_stealctxt->lr,
          &victim_stealctxt->lr,
          &lri
  );
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
    ++KAAPI_PERF_REG(kproc->perf, KAAPI_PERF_ID_STEALOP);
#endif


  /* (3)
     process all requests on the victim kprocessor and reply failed to remaining requests
     Warning: In this version the aggregator has a lock on the victim processor 
     steal context (i.e. the list of requests).
  */
  kaapi_assert_debug(!kaapi_listrequest_iterator_empty((kaapi_listrequest_iterator_t*)&lri) );

  if (!kaapi_processor_has_nowork(victim.kproc))
  {
    /* it is a combinator */
//    KAAPI_ATOMIC_INCR64( &victim.kproc->stealversion );
    kaapi_sched_stealprocessor( victim.kproc, (kaapi_listrequest_iterator_t*)&lri );
    /* unlock the victim kproc after processing the steal operation */
  }

  /* reply failed for all others requests */
  while (!kaapi_listrequest_iterator_empty((kaapi_listrequest_iterator_t*)&lri))
  {
    request = kaapi_listrequest_iterator_get( (kaapi_listrequest_iterator_t*)&lri );
    kaapi_request_replytask_ccsync(request, KAAPI_REQUEST_S_NOK);
    kaapi_listrequest_iterator_next( (kaapi_listrequest_iterator_t*)&lri );
  }

  kaapi_listrequest_iterator_destroy( &lri );
  

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
    KAAPI_EVENT_PUSH2(kproc->perf, 0, KAAPI_EVT_REQUESTS_END,
              (uintptr_t)victim.kproc->kid, serial
    );
#endif

  kaapi_assert_debug( ((kaapi_request_node_t*)self_request)->completed == 1);
  if (kaapi_request_test( self_request ))
    return kaapi_sched_emitsteal_return( kproc, &victim, self_request );

  return KAAPI_REQUEST_S_NOK;
}

