/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "kaapi_impl.h"
#include <errno.h>
#include <stddef.h> // offsetof
#include <stdio.h>
#if defined(KAAPI_USE_PERFCOUNTER)
#include "kaapi_dbg.h"
#endif

//#define LOG

static int __kaapi_sched_sync_rd_list(
  kaapi_stack_t*  stack,
  kaapi_frame_t*  unlink
);

static kaapi_task_t* _kaapi_sched_activate_task(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd,
    kaapi_task_t* task
);

struct retrydata_t;
static kaapi_task_t* _kaapi_sched_steal_activated_task(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd,
    struct retrydata_t* r
);

static void _kaapi_sched_push_init_tasks(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd
);

/*
*/
static inline //__attribute__((always_inline))
kaapi_frame_t* _kaapi_stack_save_and_push_frame(
    kaapi_stack_t* stack,
    kaapi_frame_t* frame,
    void*          sp_data
)
{
  kaapi_assert_debug( kaapi_self_thread() == kaapi_stack2thread(stack) );
  kaapi_frame_t* curr_frame = stack->unlink;

  /* save the stack.thread into its current_frame */
  curr_frame->save_pc             = stack->pc;
  curr_frame->save_thread.sp      = stack->thread.sp;
  curr_frame->save_thread.sp_data = sp_data;

  /* init the new current frame: next== 0 means save_XX undefined values */
  _kaapi_frame_init( frame );

  /* update next pc to the new empty current frame */
  stack->pc                       = stack->thread.sp;

  /* update link */
  curr_frame->next                = frame;
  stack->unlink                   = frame;
  kaapi_writemem_barrier();
  ++stack->popfp;
  return curr_frame;
}


/*
*/
kaapi_frame_t* kaapi_stack_push_frame( kaapi_stack_t* stack )
{
  kaapi_assert_debug( kaapi_self_thread() == kaapi_stack2thread(stack) );
  kaapi_frame_t* frame;

  /* save stack pointer to restor on pop */
  char* sp_data            = stack->thread.sp_data;
  if (_kaapi_has_enough_dataspace(&stack->thread, sizeof(kaapi_frame_t)) )
  {
    frame = (kaapi_frame_t*)stack->thread.sp_data;
    stack->thread.sp_data += sizeof(kaapi_frame_t);
  }
  else
    frame = (kaapi_frame_t*)kaapi_thread_slow_push_data(&stack->thread, sizeof(kaapi_frame_t));

  return _kaapi_stack_save_and_push_frame( stack, frame, sp_data);
}


/* 
*/
static 
void _kaapi_stack_pop_and_restore_frame(
    kaapi_stack_t* stack,
    kaapi_frame_t* unlink
)
{
#if defined(KAAPI_DEBUG)
  kaapi_thread_t* th = kaapi_self_thread();
  kaapi_assert_debug( th == kaapi_stack2thread(stack) );
#endif
  kaapi_assert_debug( unlink !=0 ); /* else more pop than push */
  kaapi_assert_debug( _kaapi_start_blocaddr(stack->thread.sp_data, kaapi_stack_bloc_t*) == stack->bloc_data );
  kaapi_assert_debug( _kaapi_start_blocaddr(stack->thread.sp, kaapi_stack_bloc_t*) == stack->bloc_task );

  /* restore fields */
  stack->thread.sp_data = unlink->save_thread.sp_data;
  stack->thread.sp      = unlink->save_thread.sp;
  stack->pc             = unlink->save_pc;

  /* suppress all blocs from next bloc that contains stack->thread.sp */
  kaapi_stack_bloc_t* bloc = _kaapi_start_blocaddr(stack->thread.sp, kaapi_stack_bloc_t*);
  if (bloc != stack->bloc_task)
  {
    kaapi_processor_t* kproc = kaapi_stack2context(stack)->proc;
    kaapi_atomic_lock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);

    stack->bloc_task->next = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
    kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = bloc->next;
    bloc->next = 0;
    stack->bloc_task = bloc;

    kaapi_atomic_unlock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);
  }

  /* suppress all blocs from next bloc that contains stack->thread.sp_data */
  bloc = _kaapi_start_blocaddr(stack->thread.sp_data, kaapi_stack_bloc_t*);
  if (bloc != stack->bloc_data)
  {
    kaapi_processor_t* kproc = kaapi_stack2context(stack)->proc;
    kaapi_atomic_lock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);

    stack->bloc_data->next = kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc;
    kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->freebloc = bloc->next;
    bloc->next = 0;
    stack->bloc_data = bloc;

    kaapi_atomic_unlock( &kproc->nodeinfos[KAAPI_HWS_LEVELID_THREAD]->lock);
  }

  kaapi_assert_debug(
    _kaapi_start_blocaddr(stack->thread.sp,char*) >= (char*)stack->bloc_task );
  kaapi_assert_debug(
    _kaapi_start_blocaddr(stack->thread.sp, char*) <= kaapi_lastin_stack_bloc(stack->bloc_task,char) );
  kaapi_assert_debug(
    _kaapi_start_blocaddr(stack->thread.sp_data,char*) >= (char*)stack->bloc_data );
  kaapi_assert_debug(
    _kaapi_start_blocaddr(stack->thread.sp_data, char*) <= kaapi_lastin_stack_bloc(stack->bloc_data,char) );

  /* when next ==0, thief may access to stack-> fields */
  kaapi_writemem_barrier();

  /* unlink */
  unlink->next = 0;

  /* synchronize with thieves: no more thief could access to ????
   */
  unlink->flag |= KAAPI_FRAME_FLAG_NOTSTEAL;
  unlink->flag &= ~KAAPI_FRAME_FLAG_DFG_OK;
  long popfp = --stack->popfp;
  //kaapi_writemem_barrier();
  kaapi_mem_barrier();
#if defined(KAAPI_USE_PERFCOUNTER)
  if (stack->thieffp >= popfp)
#endif
  {
    while (stack->thieffp >= popfp)
      kaapi_slowdown_cpu();
#if defined(KAAPI_USE_PERFCOUNTER)
    {
      kaapi_processor_t* kproc =kaapi_stack2context(stack)->proc;
      if (kproc->perf !=0)
        ++KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_CONFLICTPOP);
    }
#endif
  }

  stack->unlink    = unlink;
  if (unlink->rdlist !=0)
  {
    if ((unlink->flag & KAAPI_FRAME_FLAG_STACKALLOC) ==0)
      free(unlink->rdlist);
    unlink->rdlist = 0;
  }
}


/*
*/
void kaapi_stack_pop_frame( kaapi_stack_t* stack, kaapi_frame_t* frame )
{
  _kaapi_stack_pop_and_restore_frame(stack, frame);
}



/* To wakup thread when executing a stolen task (slowexec body)
*/
static int _kaapi_condition_task_isready(void* thearg)
{
  kaapi_task_t* task = (kaapi_task_t*)thearg;
  kaapi_task_body_t body = task->state.steal;
  if ((body == kaapi_nop_body) || (body == kaapi_aftersteal_body))
    return 1;
  return 0;
}


/* Execute when the owner execute a stolen task
*/
void kaapi_slowexec_body( 
  kaapi_task_t*   task,
  kaapi_thread_t* th
)
{
  kaapi_processor_t* kproc = kaapi_self_processor();
#if defined(KAAPI_USE_PERFCOUNTER)
  if (kproc->perf !=0)
  {
    kaapi_perfproc_t* perf = kproc->perf;
    kaapi_perf_thread_stopswapstart(perf, KAAPI_PERF_SCHEDULE_STATE );
    ++KAAPI_PERF_REG(perf, KAAPI_PERF_ID_SUSPEND);
    KAAPI_EVENT_PUSH0( perf, 0, KAAPI_EVT_SCHED_IDLE_BEG );
    ++KAAPI_PERF_KPROC_REG_SYS(kproc, KAAPI_PERF_ID_TASKS);
  }
#endif
  kaapi_sched_suspend(kproc, _kaapi_condition_task_isready, task);
  kaapi_assert_debug((task->state.steal == kaapi_nop_body) || (task->state.steal == kaapi_aftersteal_body));
  if (task->state.steal == kaapi_aftersteal_body)
    kaapi_aftersteal_body(task, th);
}



/* Precond: thread->pc != thread->sp (not empty)
   Post cond: all tasks between [pc,sp) are completed.
*/
static int __kaapi_sched_sync(
  kaapi_stack_t*  stack,
  kaapi_frame_t*  unlink
)
{
  unlink->flag &= ~KAAPI_FRAME_FLAG_NOTSTEAL;
  kaapi_task_t*  pc = stack->pc;
  kaapi_task_t*  sp = stack->thread.sp;

  kaapi_assert_debug(
    stack->bloc_task == _kaapi_start_blocaddr(stack->thread.sp,kaapi_stack_bloc_t*)
  );
  kaapi_assert_debug(
    stack->bloc_data == _kaapi_start_blocaddr(stack->thread.sp_data,kaapi_stack_bloc_t*)
  );
  kaapi_assert_debug((stack->bloc_task->next == 0) && (stack->bloc_data->next == 0));

  kaapi_stack_bloc_t* bloc = _kaapi_start_blocaddr(pc, kaapi_stack_bloc_t*);
  /* -1 because the last entry is lost in the bloc in order to detect bloc change when pushing */
  kaapi_task_t* sp_bloc = stack->bloc_task == bloc ?
                  sp :
                  _kaapi_end_blocaddr(pc,kaapi_task_t*)-1;
  kaapi_task_body_t body;
  kaapi_frame_t  frame;

#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_processor_t* kproc = kaapi_stack2context(stack)->proc;
  uint64_t cnt_task = 0;
#endif
  kaapi_assert_debug( stack->unlink == unlink );
  kaapi_assert_debug( stack->unlink->next == 0 );

  /* save thread state and push frame, frame automatic variable */
  _kaapi_stack_save_and_push_frame( stack, &frame, stack->thread.sp_data );
  
redo:
  kaapi_assert_debug( sizeof(kaapi_task_t)*(sp_bloc - pc) <= KAAPI_STACKBLOCSIZE);
  __builtin_prefetch( pc );
  while (pc != sp_bloc)
  {
    __builtin_prefetch( pc+1 );
    body = 0;
    KAAPI_ATOMIC_EXCHANGE(body, pc->state.steal );

    kaapi_assert_debug( body != 0 );
    __builtin_prefetch( pc->arg );
    body(pc, kaapi_stack2thread(stack));

    if (__builtin_expect(stack->pc != stack->thread.sp, 0))
    {
      if (frame.rdlist)// && (frame.rdlist->cnt_task !=0))
        __kaapi_sched_sync_rd_list( stack, &frame );
      else
        __kaapi_sched_sync( stack, &frame );
      kaapi_assert_debug(stack->pc == stack->thread.sp);
    }

#if defined(KAAPI_USE_PERFCOUNTER)
    ++cnt_task;
#endif

    ++pc;
    /* next line may be important to let thief go directly on the next running task
       this is particulary important for frame with a lot of task
    */
    unlink->save_pc = pc;
  }

  /* here, may be some tasks stored on the next bloc */
  if (pc != sp)
  {
    kaapi_stack_bloc_t* nextbloc = bloc->next;
    pc = kaapi_firstin_stack_bloc(nextbloc,kaapi_task_t);
    bloc = nextbloc;
    sp_bloc = _kaapi_start_blocaddr(sp, kaapi_stack_bloc_t*) == bloc ?
                    sp :
                    _kaapi_end_blocaddr(pc,kaapi_task_t*)-1;
    kaapi_assert_debug( pc != sp );
    stack->cnt_task -= sp_bloc - pc;
    goto redo;
  }

  /* pop the frame and return */
  _kaapi_stack_pop_and_restore_frame( stack, unlink );
  kaapi_assert_debug(stack->pc == stack->thread.sp);
  kaapi_assert_debug( stack->bloc_task == _kaapi_start_blocaddr(stack->thread.sp, kaapi_stack_bloc_t*));
  kaapi_assert_debug( stack->bloc_data == _kaapi_start_blocaddr(stack->thread.sp_data, kaapi_stack_bloc_t*));

#if defined(KAAPI_USE_PERFCOUNTER)
  //printf("#task: %llu\n", cnt_task );
  if (kproc->perf !=0)
    KAAPI_PERF_KPROC_REG_USR(kproc, KAAPI_PERF_ID_TASKS) += cnt_task;
#endif

  return 0;
}


  struct retrydata_t {
    int count_retry;
    int count;
    int count_numa;
    int delay;
  };

/* Execute task using a ready list of tasks.
   Assumption : all the tasks of the frame have processed to compute successors,
   i.e. the list of tasks accessible through readylist == the list of tasks in the stack.
 */
static int __kaapi_sched_sync_rd_list(
  kaapi_stack_t*  stack,
  kaapi_frame_t*  unlink
)
{
  kaapi_task_t*  pc;
  kaapi_task_body_t body;
  kaapi_frame_t  frame;
  kaapi_frame_t* curr_frame = unlink;
  kaapi_frame_wrdlist_t* frame_rd = curr_frame->rdlist;
  kaapi_frame_wrdlist_t* frame_master = frame_rd->master == 0 ? frame_rd : frame_rd->master;
  kaapi_processor_t* kproc = kaapi_stack2context(stack)->proc;

#if defined(KAAPI_USE_PERFCOUNTER)
  uint64_t local_exec_task = 0; /* local executed task: cumulated for perf counter */
#endif

  uint64_t local_frame_task = 0;
  int register_tomaster_flag = 0;
  struct retrydata_t retrydata;

#if 0
#define KAAPI_STEAL_FAILED_INST( arg )\
   { ++(arg)->count;  \
     kaapi_slowdown_cpu(); \
     kaapi_slowdown_cpu(); \
   }
#else
#define KAAPI_STEAL_FAILED_INST( arg )\
   { ++(arg)->count; \
     if ((arg)->count > 16) \
     { \
       (arg)->delay *= 2; \
       struct timespec ts = { 0, (arg)->delay }; \
       nanosleep( &ts, 0 ); \
       if ((arg)->delay > 64) (arg)->delay = 64; \
     } \
     else kaapi_slowdown_cpu(); \
   }
#endif

//    {(arg)->count_retry = 8*kaapi_get_concurrency(); 
#define KAAPI_STEAL_SUCCESS_INST( arg )\
   {(arg)->count_retry = 8;\
    (arg)->count = 0;\
    (arg)->count_numa = 0;\
    (arg)->delay = 1; }\

  kaapi_assert_debug( unlink->flag & KAAPI_FRAME_FLAG_NOTSTEAL);

  /* distribute initial ready tasks */
  if (!kaapi_list_empty(&frame_rd->rd))
    _kaapi_sched_push_init_tasks( kproc, frame_rd );

  unlink->flag &= ~KAAPI_FRAME_FLAG_NOTSTEAL;

#if defined(KAAPI_USE_PERFCOUNTER)
  /* here sfp is initialized, dump graph if required */
  if (getenv("KAAPI_DUMP_GRAPH") !=0)
  {
    static uint32_t counter = 0;
    char filename[128]; 
    if (getenv("USER") !=0)
      sprintf(filename,"/tmp/graph_sync_rdlist.%s.%i.dot", getenv("USER"), counter++ );
    else
      sprintf(filename,"/tmp/graph_sync_rdlist.%i.dot", counter++);
    FILE* filedot = fopen(filename, "w");
    kaapi_frame_print_dot( filedot, stack, unlink );
    fclose(filedot);
  }
#endif

  KAAPI_STEAL_SUCCESS_INST(&retrydata);
  
  kaapi_assert_debug( stack->unlink == unlink );
  kaapi_assert_debug( stack->unlink->next == 0 );

  /* save thread state and push newframe, frame automatic variable */
  _kaapi_stack_save_and_push_frame( stack, &frame, stack->thread.sp_data );

redo:
  pc = _kaapi_sched_steal_activated_task(kproc, frame_rd, &retrydata );
  while ( pc != 0 )
  {
    kaapi_assert_debug( pc->u.s.flag & KAAPI_TASK_FLAG_DFGOK);

    /* task is pop from queue, do not need atomic exchange to mark it */
    body = pc->body;
    /* not possible here : except if task was poped several times */
    kaapi_assert_debug(body != kaapi_slowexec_body);

    __builtin_prefetch( pc->arg );
    body(pc, (kaapi_thread_t*)stack);

    /* If task creates new task : always execute them first before activating sucessor of pc
    */
    if (__builtin_expect(stack->thread.sp != stack->pc, 0))
      __kaapi_sched_sync( stack, &frame );

#if defined(KAAPI_USE_PERFCOUNTER)
    ++local_exec_task;
#endif

    /* if steal task from frame != frame_master, then signal original master frame
       else cache data count locally
    */
    if (pc->state.frame != frame_master)
    {
      kaapi_writemem_barrier();
      KAAPI_ATOMIC_INCR(&pc->state.frame->cnt_exec);
    }
    else
      ++local_frame_task;

    /* increment here to ensure that child tasks are executed */
    pc = _kaapi_sched_activate_task( kproc, frame_rd, pc );
  }

  kaapi_writemem_barrier();
  if (frame_rd->master ==0)
  {
    if (local_frame_task)
    {
      kaapi_writemem_barrier();
      KAAPI_ATOMIC_ADD(&frame_rd->cnt_exec, local_frame_task);
      local_frame_task = 0;
      KAAPI_STEAL_SUCCESS_INST(&retrydata);
    }
    else 
    {
      //SEEMS to not be good 
      KAAPI_STEAL_FAILED_INST(&retrydata);
    }

    if (KAAPI_ATOMIC_READ(&frame_rd->cnt_exec) != frame_rd->cnt_task)
      goto redo;

    /* here: cnt_exec == cnt_task, wait worker to finish */
    while (KAAPI_ATOMIC_READ(&frame_rd->cnt_worker) !=0)
    {
      kaapi_slowdown_cpu();
    }
  }
  else
  {
    /* no need to synchronize because worker steal at least one task, thus master always has
       its frame valid.
    */
    if (!register_tomaster_flag)
    {
       register_tomaster_flag = 1;
       KAAPI_ATOMIC_INCR(&frame_master->cnt_worker);
    }

    if (local_frame_task ==0)
    {
      if ((KAAPI_ATOMIC_READ(&frame_master->cnt_exec) != frame_master->cnt_task)
       && (--retrydata.count_retry !=0))
      {
        KAAPI_STEAL_FAILED_INST(&retrydata);
        goto redo;
      }
    }
    else 
    { /* signal master frame and returns if completion */
      kaapi_writemem_barrier();
      if (KAAPI_ATOMIC_ADD(&frame_master->cnt_exec, local_frame_task ) != frame_master->cnt_task)
      {
        local_frame_task = 0;
        KAAPI_STEAL_SUCCESS_INST(&retrydata);
        goto redo;
      }
    }
    /* do not continue to steal directly from frame, update master worker counter */
    KAAPI_ATOMIC_DECR(&frame_master->cnt_worker);
  }

  curr_frame->save_pc = curr_frame->save_thread.sp;

  /* pop the frame and return */
  _kaapi_stack_pop_and_restore_frame( stack, unlink );

#if defined(KAAPI_USE_PERFCOUNTER)
  //printf("Local exec task: %llu\n", local_cnt_task);
  if (kproc->perf !=0)
    KAAPI_PERF_KPROC_REG_USR(kproc, KAAPI_PERF_ID_TASKS) += local_exec_task;
#endif
  return 0;
}


/* Threshold to switch from stack of tasks to ready list management
   Should be computed by stealer to avoid the necessity to use threshold.
*/
//#define KAAPI_TASK_THRESHOLD 256
#define KAAPI_TASK_THRESHOLD 8

/* Public entry point
*/
int kaapi_sched_sync( kaapi_thread_t* thread )
{
  /* return if empty task to perform */
  kaapi_stack_t* stack = kaapi_thread2stack(thread);
  if (stack->pc == thread->sp)
    return 0;

  kaapi_frame_t* frame = stack->unlink;
  if ((frame->flag & KAAPI_FRAME_FLAG_DFG_OK) ==0)
  {
    if (kaapi_frame_count_task(stack, frame) <= KAAPI_TASK_THRESHOLD)
    {
      __kaapi_sched_sync(stack, stack->unlink);
      kaapi_assert_debug( stack->pc == thread->sp );
      return 0;
    }
    kaapi_assert_debug( frame->rdlist == 0 );
    kaapi_frame_computereadylist(stack,frame);
    __kaapi_sched_sync_rd_list(stack, frame);

    //kaapi_assert_debug( kaapi_queue_empty(&frame->rdlist->rd) );
    return 0;
  }

  if ((frame->rdlist->starttask !=0) || !kaapi_list_empty(&frame->rdlist->rd))
    __kaapi_sched_sync_rd_list(stack, frame);
  //kaapi_assert_debug( kaapi_queue_empty(&frame->rdlist->rd) );
  return 0;
}


/* */
static kaapi_task_t* _kaapi_sched_steal_ld_task( kaapi_localitydomain_t* ld )
{
  kaapi_task_t* task = 0;
  if ((ld !=0) && !kaapi_queue_empty(&ld->queue))
  {
    kaapi_atomic_lock( &ld->lock_r );
    task = kaapi_queue_steal(&ld->queue);
    kaapi_atomic_unlock( &ld->lock_r );
  }
  return task;
}

/* */
static kaapi_task_t* _kaapi_sched_pop_ld_task( kaapi_localitydomain_t* ld )
{
  kaapi_task_t* task = 0;
  if ((ld !=0) && (!kaapi_queue_empty(&ld->queue) || !kaapi_queue_empty(&ld->private_queue)))
  {
    kaapi_atomic_lock( &ld->lock_r );
    task = kaapi_queue_pop(&ld->private_queue);
    if (task ==0)
      task = kaapi_queue_pop(&ld->queue);
    kaapi_atomic_unlock( &ld->lock_r );
  }
  return task;
}


/*
*/
static kaapi_task_t* _kaapi_sched_steal_activated_task(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd,
    struct retrydata_t* r
)
{
  kaapi_task_t* task;
  kaapi_localitydomain_t*ld;

  /* initial task - kind of private queue */
  task = frame_rd->starttask;
  if (task)
  {
    frame_rd->starttask = 0;
    return task;
  }

  /* local queue */
  ld = kproc->lds[KAAPI_HWS_LEVELID_THREAD];
  task = _kaapi_sched_pop_ld_task( ld  );
  if (task !=0) return task;

  /* local NUMA queue */
  ld = kproc->lds[KAAPI_HWS_LEVELID_NUMA];
  task = _kaapi_sched_pop_ld_task( ld );
  if (task !=0) return task;

#if 1
  /* other kproc in the same NUMA node */
  if (kproc->lds[KAAPI_HWS_LEVELID_NUMA] !=0)
  {
    int rn = rand_r( (unsigned int*)&kproc->seed ) % kproc->lds[KAAPI_HWS_LEVELID_NUMA]->count;
    ld = kproc->lds[KAAPI_HWS_LEVELID_NUMA]->kids[rn]->lds[KAAPI_HWS_LEVELID_THREAD];
    task = _kaapi_sched_steal_ld_task( ld );
    if (task !=0) return task;
  }
#endif

#if 0
  if (kaapi_default_param.wsselect == kaapi_sched_select_victim_numa)
  {
    kaapi_team_t* team = kproc->team;
    if (++r->count_numa < 4*team->localitydomains[KAAPI_HWS_LEVELID_NUMA].count) return 0;
    r->count_numa = 0;
  }

#if defined(KAAPI_USE_NUMA)
{
  kaapi_team_t* team = kproc->team;
  /* random on all NUMA nodes */
  int rn = rand_r( (unsigned int*)&kproc->seed ) % team->localitydomains[KAAPI_HWS_LEVELID_NUMA].count;
  ld = team->localitydomains[KAAPI_HWS_LEVELID_NUMA].lds[rn];
  task = _kaapi_sched_steal_ld_task( ld );
  if (task !=0) return task;
}
#endif

#if 1
{
  kaapi_team_t* team = kproc->team;
  /* random on all THREADs */
  int rn = rand_r( (unsigned int*)&kproc->seed ) % team->localitydomains[KAAPI_HWS_LEVELID_THREAD].count;
  ld = team->localitydomains[KAAPI_HWS_LEVELID_THREAD].lds[rn];
  task = _kaapi_sched_steal_ld_task( ld );
  if (task !=0) return task;
}
#endif
#endif


  return 0;
}


/*
*/
static void _kaapi_sched_push_init_tasks(
    kaapi_processor_t* kproc,
    kaapi_frame_wrdlist_t* frame_rd
)
{
  kaapi_task_t* task;
  uint64_t      cnt_task = 0;

  //kaapi_atomic_lock(&kproc->lds[KAAPI_HWS_LEVELID_THREAD]->lock_r);
  while ((task = kaapi_list_pop_head(&frame_rd->rd)) !=0)
  {
    /* */
    ++cnt_task;
    kaapi_assert( 0 == kaapi_default_param.wspush( kproc,
                                                   frame_rd,
                                                   task,
                                                   &kproc->lds[KAAPI_HWS_LEVELID_THREAD]->queue)
    );
  }
  //kaapi_atomic_unlock(&kproc->lds[KAAPI_HWS_LEVELID_THREAD]->lock_r);
}


/*
*/
static kaapi_task_t* _kaapi_sched_activate_task(
    kaapi_processor_t*     kproc,
    kaapi_frame_wrdlist_t* frame_rd,
    kaapi_task_t*          task
)
{
  kaapi_access_mode_t   mode, mode_param;
  kaapi_access_mode_t   first_next_mode_param, next_mode_param;
  kaapi_access_t*       access;
  const kaapi_format_t* fmt;
  void*                 access_data;

  unsigned int          i;
  size_t                count_params;
  kaapi_task_t*         retval = 0;
  kaapi_queue_t*        local_queue;
  kaapi_queue_t         list;

  kaapi_queue_init(&list);

  kaapi_assert_debug( task->u.s.flag & KAAPI_TASK_FLAG_DFGOK);
  if (task->u.s.flag & KAAPI_TASK_FLAG_INDPENDENT)
  {
    kaapi_assert_debug( task->next == 0);
    goto return_value;
  }

  fmt         = task->fmt;
  kaapi_assert_debug( fmt != 0)
  count_params = kaapi_format_get_count_params( fmt, task->arg );

  for (i=0; i<count_params; ++i)
  {
    mode            = kaapi_format_get_mode_param(fmt, i, task->arg);
    mode_param      = KAAPI_ACCESS_GET_MODE( mode ); 
    if (mode_param & KAAPI_ACCESS_MODE_V)
      continue;

    access      = kaapi_format_get_access_param(fmt, i, task->arg);

#if 0
    if (access->version !=0)
    {
      kaapi_abort();
      /* case of WAR: version was reallocated with view of task as original view
         Here we assume no variable renaming for detecting dependency.
         Else...to do
      */
      kaapi_memory_view_t view_dest;
      kaapi_memory_view_t view_src;
      access_data = access->version;
      kaapi_format_get_view_param(fmt, i, task->arg, &view_dest);
      view_src = view_dest;
      kaapi_memory_view_reallocated( &view_src );
    }
    else
#endif
      access_data = access->data;

    /* activate all the nexts concurrent with the task */
    access = access->next;
    if (access !=0)
    {
      next_mode_param = KAAPI_ACCESS_GET_MODE( access->mode );
      first_next_mode_param = next_mode_param;
      do {
        if (access->data != access_data)
          kaapi_abort(__LINE__,__FILE__,"*** internal error: invalid case\n");

        kaapi_task_t* next_task = access->task;
//          printf("Task :%p decr counter for task: %p\n", (void*)task, (void*)next_task);
        int wc = KAAPI_ATOMIC_DECR(&next_task->wc);
        kaapi_assert_debug( wc >= 0 );
        if (wc == 0)
        {
          kaapi_assert_debug( next_task->u.s.flag & KAAPI_TASK_FLAG_DFGOK );

/* Ici virtualisation du push
   - prendre la list des tâches dans lists dans son ensemble (ci dessous) ?
*/
          /* */
          kaapi_assert( 0 == kaapi_default_param.wspush(kproc, frame_rd, next_task, &list) );

/* Ici fin virtualization du push
*/
        }
        kaapi_access_t* prev = access;
        access = access->next;
        if (access ==0) break;
        next_mode_param = KAAPI_ACCESS_GET_MODE( access->mode );
        if (!KAAPI_ACCESS_IS_CONCURRENT( first_next_mode_param, next_mode_param))
          break;
        /* do not forget to clear the link ! */
        prev->next = 0;
      } while (1);
    }
  }

return_value:
  local_queue = &kproc->lds[KAAPI_HWS_LEVELID_THREAD]->queue;
  kaapi_atomic_lock(&kproc->lds[KAAPI_HWS_LEVELID_THREAD]->lock_r);
  if (!kaapi_queue_empty(&list))
    kaapi_queue_append_head( local_queue, &list );
  if (retval == 0)
    retval = kaapi_queue_pop( local_queue );
  kaapi_atomic_unlock(&kproc->lds[KAAPI_HWS_LEVELID_THREAD]->lock_r);

  return retval;
}


/*
*/
int kaapi_thread_save(kaapi_thread_t* thread, kaapi_thread_register_t* frame)
{
  kaapi_stack_t* stack = kaapi_thread2stack(thread);
  frame->pc = stack->pc;
  frame->sp = thread->sp;
  frame->sp_data = thread->sp_data;
  frame->reserved1 = stack->unlink;
  frame->reserved2 = stack->bloc_task;
  frame->reserved3 = stack->bloc_data;
  frame->reserved4 = 0;
  return 0;
}


/* 
*/
int kaapi_thread_restore(kaapi_thread_t* thread, const kaapi_thread_register_t* frame)
{
  kaapi_stack_t* stack = kaapi_thread2stack(thread);
  kaapi_assert_debug( (void*)stack->bloc_task == frame->reserved2);
  kaapi_assert_debug( (void*)stack->bloc_data == frame->reserved3);
  kaapi_assert_debug( 0 == frame->reserved4);

  stack->pc = 0;
  kaapi_writemem_barrier();
  thread->sp = frame->sp;
  kaapi_writemem_barrier();
  stack->pc = frame->pc;
  thread->sp_data = frame->sp_data;
  stack->unlink = frame->reserved1;
  return 0;
}


static void _kaapi_resize_stack( kaapi_stack_t* stack, size_t data_size, size_t task_size )
{
  kaapi_stack_bloc_t** bloc= 0;
  if( data_size ==0)
    bloc = &stack->bloc_task;
  else if (task_size ==0)
    bloc = &stack->bloc_data;
  else
    kaapi_abort(__LINE__, __FILE__, "Invalid case");

  if ((data_size+task_size) >= kaapi_default_param.stackblocsize)
  {
    fprintf(stderr,"*** stack bloc overflow: data too big for one allocation,"
         " please extend your KAAPI_STACKBLOCSIZE and recompile the library."
         " KAAPI_STACKBLOCSIZE is actually set to %li bytes / %.2f MBytes",
      kaapi_default_param.stackblocsize, kaapi_default_param.stackblocsize/1024.0/1024.0
    );
    kaapi_abort(__LINE__,__FILE__, "Invalid value");
  }

  kaapi_processor_t* kproc = kaapi_stack2context(stack)->proc;
  kaapi_stack_bloc_t* newbloc = kaapi_processor_alloc_stackbloc(kproc);
  kaapi_assert_debug(newbloc != *bloc);
  kaapi_assert_debug(newbloc->next == 0);

  if (newbloc ==0)
    kaapi_abort(__LINE__,__FILE__, "No enough memory");

  /* link new bloc */
  newbloc->save_ptr = (data_size == 0 ? (void*)stack->thread.sp : (void*)stack->thread.sp_data);
  kaapi_assert_debug( _kaapi_start_blocaddr(newbloc->save_ptr,void*) >= (void*)*bloc );
  kaapi_assert_debug( _kaapi_start_blocaddr(newbloc->save_ptr, void*) <= (void*)*bloc );

  (*bloc)->next = newbloc;
  (*bloc) = newbloc;

  if (data_size ==0)
  { /* save the number of task in the bloc : sp - pc if pc in the bloc or sp-firstin_stack_bloc(sp)*/
    if (!_kaapi_addr_inbloc( *bloc, stack->pc))
      stack->cnt_task += ((*bloc)->save_cnt_task = stack->thread.sp-kaapi_firstin_stack_bloc(*bloc,kaapi_task_t));
    stack->thread.sp = kaapi_firstin_stack_bloc( newbloc, kaapi_task_t);
  }
  else
    stack->thread.sp_data = kaapi_firstin_stack_bloc( newbloc, char);
}


/*
*/
void* kaapi_thread_slow_push_data( kaapi_thread_t* thread, unsigned long requested_size )
{
  kaapi_stack_t* stack = kaapi_thread2stack(thread);
  _kaapi_resize_stack( stack, requested_size, 0 );
  char* retval = stack->thread.sp_data;
  stack->thread.sp_data += requested_size;
  return retval;
}

/*
*/
kaapi_task_t* kaapi_thread_slow_pushtask(kaapi_thread_t* thread)
{
  kaapi_stack_t* stack = kaapi_thread2stack(thread);
  _kaapi_resize_stack( stack, 0, 1 );
  kaapi_task_t* retval = stack->thread.sp;
  return retval;
}

